/* chanserv.c - Channel service bot
 * Copyright 2000-2003 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  Important limitations are
 * listed in the COPYING file that accompanies this software.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, email srvx-maintainers@srvx.net.
 *
 * $Id: chanserv.c,v 1.402 2003/12/23 13:59:16 entrope Exp $
 */

#include "chanserv.h"
#include "conf.h"
#include "global.h"
#include "modcmd.h"
#include "opserv.h" /* for opserv_bad_channel() */
#include "saxdb.h"
#include "timeq.h"

#define CHANSERV_CONF_NAME	"services/chanserv"

/* ChanServ options */
#define KEY_SUPPORT_CHANNEL     "support_channel"
#define KEY_SUPPORT_CHANNEL_MODES "support_channel_modes"
#define KEY_DB_BACKUP_FREQ	"db_backup_freq"
#define KEY_INFO_DELAY		"info_delay"
#define KEY_MAX_GREETLEN       	"max_greetlen"
#define KEY_ADJUST_THRESHOLD   	"adjust_threshold"
#define KEY_ADJUST_DELAY   	"adjust_delay"
#define KEY_CHAN_EXPIRE_FREQ	"chan_expire_freq"
#define KEY_CHAN_EXPIRE_DELAY	"chan_expire_delay"
#define KEY_MAX_CHAN_USERS     	"max_chan_users"
#define KEY_MAX_CHAN_BANS	"max_chan_bans"
#define KEY_NICK		"nick"
#define KEY_TRIGGER             "trigger"
#define KEY_OLD_CHANSERV_NAME	"old_chanserv_name"
#define KEY_MAX_SWITCH_LOAD	"max_switch_load"
#define KEY_SWITCH_TIMEOUT	"switch_timeout"
#define KEY_8BALL_RESPONSES     "8ball"
#define KEY_REFRESH_PERIOD      "refresh_period"
#define KEY_CTCP_SHORT_BAN_DURATION "ctcp_short_ban_duration"
#define KEY_CTCP_LONG_BAN_DURATION  "ctcp_long_ban_duration"
#define KEY_MAX_OWNED               "max_owned"
#define KEY_IRC_OPERATOR_EPITHET    "irc_operator_epithet"
#define KEY_NETWORK_HELPER_EPITHET  "network_helper_epithet"
#define KEY_SUPPORT_HELPER_EPITHET  "support_helper_epithet"
#define KEY_NODELETE_LEVEL      "nodelete_level"

/* ChanServ database */
#define KEY_CHANNELS		"channels"
#define KEY_NOTE_TYPES          "note_types"

/* Note type parameters */
#define KEY_NOTE_OPSERV_ACCESS  "opserv_access"
#define KEY_NOTE_CHANNEL_ACCESS "channel_access"
#define KEY_NOTE_SETTER_ACCESS  "setter_access"
#define KEY_NOTE_VISIBILITY     "visibility"
#define KEY_NOTE_VIS_PRIVILEGED "privileged"
#define KEY_NOTE_VIS_CHANNEL_USERS "channel_users"
#define KEY_NOTE_VIS_ALL        "all"
#define KEY_NOTE_MAX_LENGTH     "max_length"
#define KEY_NOTE_SETTER         "setter"
#define KEY_NOTE_NOTE           "note"

/* Do-not-register channels */
#define KEY_DNR                 "dnr"
#define KEY_DNR_SETTER          "setter"
#define KEY_DNR_REASON          "reason"

/* Channel data */
#define KEY_REGISTERED		"registered"
#define KEY_REGISTRAR		"registrar"
#define KEY_SUSPENDED           "suspended"
#define KEY_PREVIOUS            "previous"
#define KEY_SUSPENDER		"suspender"
#define KEY_ISSUED              "issued"
#define KEY_REVOKED             "revoked"
#define KEY_SUSPEND_EXPIRES     "suspend_expires"
#define KEY_SUSPEND_REASON      "suspend_reason"
#define KEY_VISITED		"visited"
#define KEY_TOPIC		"topic"
#define KEY_GREETING		"greeting"
#define KEY_USER_GREETING	"user_greeting"
#define KEY_MODES		"modes"
#define KEY_FLAGS		"flags"
#define KEY_USERS		"users"
#define KEY_BANS		"bans"
#define KEY_MAX			"max"
#define KEY_NOTES               "notes"
#define KEY_TOPIC_MASK          "topic_mask"

/* User data */
#define KEY_LEVEL		"level"
#define KEY_INFO		"info"
#define KEY_SEEN		"seen"

/* Ban data */
#define KEY_OWNER		"owner"
#define KEY_REASON		"reason"
#define KEY_SET			"set"
#define KEY_DURATION		"duration"
#define KEY_EXPIRES             "expires"
#define KEY_TRIGGERED		"triggered"

/* Administrative messages */
#define CSMSG_CHANNELS_EXPIRED	"%i channels expired."

/* Channel registration */
#define CSMSG_REG_SUCCESS	"You now have ownership of $b%s$b."
#define CSMSG_PROXY_SUCCESS	"%s now has ownership of $b%s$b."
#define CSMSG_ALREADY_REGGED	"$b%s$b is registered to someone else."
#define CSMSG_MUST_BE_OPPED	"You must be a channel operator in $b%s$b to register it."
#define CSMSG_PROXY_FORBIDDEN	"You may not register a channel for someone else."
#define CSMSG_COLORS_DISALLOWED "You may not have color codes in your %s."
#define CSMSG_OWN_TOO_MANY      "%s already owns enough channels (at least %d); use FORCE to override."

/* Do-not-register channels */
#define CSMSG_NOT_DNR           "$b%s$b is not a valid channel name or *account."
#define CSMSG_DNR_INFO          "$b%s$b is do-not-register (by $b%s$b): %s"
#define CSMSG_DNR_CHANNEL       "Only network staff may register $b%s$b."
#define CSMSG_DNR_CHANNEL_MOVE  "Only network staff may move $b%s$b."
#define CSMSG_DNR_ACCOUNT       "Only network staff may register channels to $b%s$b."
#define CSMSG_NOREGISTER_CHANNEL "$b%s$b has been added to the do-not-register list."
#define CSMSG_NO_SUCH_DNR       "$b%s$b is not in the do-not-register list."
#define CSMSG_DNR_REMOVED       "$b%s$b has been removed from the do-not-register list."

/* Channel unregistration */
#define CSMSG_UNREG_SUCCESS	"$b%s$b has been unregistered."
#define CSMSG_UNREG_NODELETE	"$b%s$b is protected from unregistration."
#define CSMSG_CHAN_SUSPENDED    "$b$C$b access to $b%s$b has been temporarily suspended (%s)."
#define CSMSG_CONFIRM_UNREG	"To confirm this unregistration, you must use 'unregister %s'."

/* Channel moving */
#define CSMSG_MOVE_SUCCESS	"Channel registration has been moved to $b%s$b."
#define CSMSG_MOVE_NODELETE	"$b%s$b is protected from unregistration, and cannot be moved."

/* Channel merging */
#define CSMSG_MERGE_SUCCESS	"Channel successfully merged into $b%s$b."
#define CSMSG_MERGE_SELF	"Merging cannot be performed if the source and target channels are the same."
#define CSMSG_MERGE_NODELETE    "You may not merge a channel that is marked NoDelete."
#define CSMSG_MERGE_SUSPENDED	"Merging cannot be performed if the source or target channel is suspended."
#define CSMSG_MERGE_NOTOWNER    "You must be the owner of the target channel (or a helper) to merge into the channel."

/* Handle unregistration */
#define CSMSG_HANDLE_UNREGISTERED "As a result of your account unregistration, you have been deleted from all of your channels' userlists."

/* Error messages */
#define CSMSG_NOT_USER		"You lack access to $b%s$b."
#define CSMSG_NO_CHAN_USER	"%s lacks access to $b%s$b."
#define CSMSG_NO_ACCESS		"You lack sufficient access to use this command."
#define CSMSG_ACCESS_REQUIRED  	"You must have $b%s$b access in %s to use this command."
#define CSMSG_NOT_REGISTERED	"$b%s$b has not been registered with $b$C$b."
#define CSMSG_STRUCTURE_LIMIT	"This channel has reached the maximum %s limit of $b%d$b."
#define CSMSG_ILLEGAL_CHANNEL	"$b%s$b is an illegal channel, and cannot be registered."
#define CSMSG_GODMODE_UP        "You may not use $b%s$b to op yourself unless you are on the user list.  Use the $bop$b command instead."
#define CSMSG_ALREADY_UP	"You are already %s in $b%s$b."
#define CSMSG_ALREADY_DOWN	"You are not opped or voiced in $b%s$b."
#define CSMSG_ALREADY_OPCHANNED "There has been no net.join since the opchan in $b%s$b."
#define CSMSG_OPCHAN_DONE       "I have (re-)opped myself in $b%s$b."

/* Removing yourself from a channel. */
#define CSMSG_NO_OWNER_DELETEME "You cannot delete your owner access in $b%s$b."
#define CSMSG_CONFIRM_DELETEME  "To confirm this user removal, you must use 'deleteme %s'."
#define CSMSG_DELETED_YOU       "Your $b%s$b access has been deleted from $b%s$b."

/* User management */
#define CSMSG_ADDED_USER	"Added new $b%s$b %s to the %s user list."
#define CSMSG_DELETED_USER	"Deleted %s $b%s$b %s from the %s user list."
#define CSMSG_DELETED_USERS	"Deleted %ss matching '%s' from the %s user list."
#define CSMSG_TRIMMED_USERS	"Trimmed $b%d %ss$b from the %s user list who were inactive for at least %s."
#define CSMSG_INCORRECT_ACCESS	"%s is %s $b%s$b, not %s %s."
#define CSMSG_USER_EXISTS	"%s is already on the $b%s$b user list (as %s %s)."

#define CSMSG_CANNOT_ADD	"You lack sufficient access to add users."
#define CSMSG_CANNOT_DEL	"You lack sufficient access to delete users."
#define CSMSG_CANNOT_CLVL	"You lack sufficient access to modify users' access."
#define CSMSG_CANNOT_TRIM	"You must include a minimum inactivity duration of at least 60 seconds to trim."

#define CSMSG_NO_SELF_CLVL	"You cannot change your own access."
#define CSMSG_NO_BUMP_ACCESS	"You cannot give users access greater than or equal to your own."
#define CSMSG_MULTIPLE_OWNERS   "There is more than one owner in %s; please use $bCLVL$b, $bDELOWNER$b and/or $bADDOWNER$b instead."
#define CSMSG_NO_TRANSFER_SELF  "You cannot give ownership to your own account."
#define CSMSG_OWNERSHIP_GIVEN   "Ownership of $b%s$b has been transferred to account $b%s$b."

/* Ban management */
#define CSMSG_BAN_ADDED		"Permanently banned $b%s$b from %s."
#define CSMSG_TIMED_BAN_ADDED	"Banned $b%s$b from %s for %s."
#define CSMSG_KICK_BAN_DONE    	"Kickbanned $b%s$b from %s."
#define CSMSG_BAN_DONE		"Banned $b%s$b from %s."
#define CSMSG_REASON_CHANGE	"Reason for ban $b%s$b changed."
#define CSMSG_BAN_EXTENDED	"Extended ban for $b%s$b expires in %s."
#define CSMSG_BAN_REMOVED	"Matching ban(s) for $b%s$b removed."
#define CSMSG_TRIMMED_BANS	"Trimmed $b%d bans$b from the %s ban list that were inactive for at least %s."
#define CSMSG_REDUNDANT_BAN    	"$b%s$b is already banned in %s."
#define CSMSG_INVALID_DURATION  "Timed bans must last for $b%s$b %s."
#define CSMSG_LAME_MASK 	"$b%s$b is a little too general. Try making it more specific."
#define CSMSG_MASK_PROTECTED	"Sorry, ban for $b%s$b conflicts with a protected user's hostmask."
#define CSMSG_NO_MATCHING_USERS "No one in $b%s$b has a hostmask matching $b%s$b."
#define CSMSG_BAN_NOT_FOUND	"Sorry, no ban found for $b%s$b."
#define CSMSG_BANLIST_FULL	"The $b%s$b channel ban list is $bfull$b."

#define CSMSG_INVALID_TRIM	"$b%s$b isn't a valid trim target."

/* Channel management */
#define CSMSG_CHANNEL_OPENED	"$b%s$b has been opened."
#define CSMSG_CLEARED_INFO_LINE	"Your info line has been cleared."
#define CSMSG_SET_INFO_LINE    	"Your info line has been set to '%s'."
#define CSMSG_INFO_LINE    	"Your info line is: %s."
#define CSMSG_NO_INFO_LINE    	"You don't have an info line set."
#define CSMSG_WIPED_INFO_LINE   "Removed $b%s$b's infoline in $b%s$b."

#define CSMSG_TOPIC_SET		"Topic is now '%s'."
#define CSMSG_TOPIC_CLEARED   	"Topic has been reset."
#define CSMSG_NO_TOPIC		"$b%s$b does not have a default topic."
#define CSMSG_TOPICMASK_CONFLICT1 "I do not know how to make that topic work with the current topic mask in $b%s$b, which is: %s"
#define CSMSG_TOPICMASK_CONFLICT2 "Please make sure your topic at most %d characters and matches the topic mask pattern."
#define CSMSG_TOPIC_LOCKED	"The %s topic is locked."
#define CSMSG_TOPIC_DENY	"You lack sufficient access to change the $b%s$b topic."
#define CSMSG_MASK_BUT_NO_TOPIC "Warning: $b%s$b does not have a default topic, but you just set the topic mask."
#define CSMSG_TOPIC_MISMATCH    "Warning: The default topic for $b%s$b does not match the topic mask; changing it anyway."

#define CSMSG_MODES_SET		"Channel modes are now $b%s$b."
#define CSMSG_NO_MODES		"$b%s$b does not have any default modes."
#define CSMSG_MODE_LOCKED	"Modes conflicting with $b%s$b are not allowed in %s."
#define CSMSG_CANNOT_SET        "That setting is above your current level, so you cannot change it."
#define CSMSG_BAD_SETLEVEL      "You cannot change any setting to above your level."

#define CSMSG_USER_PROTECTED   	"Sorry, $b%s$b is protected."
#define CSMSG_OPBY_LOCKED	"You may not op users who lack op or greater access."
#define CSMSG_PROCESS_FAILED	"$b$C$b could not process some of the nicks you provided."
#define CSMSG_PROCESS_USERS	"%s users in $b%s$b."

#define CSMSG_INVITED_USER	"Invited $b%s$b to join %s."
#define CSMSG_INVITING_YOU	"$b%s$b invites you to join %s%s%s"
#define CSMSG_CANNOT_INVITE	"You lack sufficient access to invite users."
#define CSMSG_ALREADY_PRESENT	"%s is $balready in %s$b."
#define CSMSG_YOU_ALREADY_PRESENT "You are already in $b%s$b."
#define CSMSG_LOW_CHANNEL_ACCESS "You lack sufficient access in %s to use this command."

#define CSMSG_KICK_DONE		"Kicked $b%s$b from %s."
#define CSMSG_NO_BANS		"No channel bans found on $b%s$b."
#define CSMSG_BANS_REMOVED	"Removed all channel bans from $b%s$b."

/* Channel userlist */
#define CSMSG_ACCESS_SEARCH_HDR	"%s $b%ss$b matching %s:"
#define CSMSG_ACCESS_HEADER	"%s $b%ss$b:"

/* Channel banlist */
#define CSMSG_BAN_HEADER     	"$b%-*s%-*s%-*s%-*s%s$b"
#define CSMSG_BAN_DATA     	"$b%-*s$b%-*s%-*s%-*s%s"

#define CSMSG_INVALID_ACCESS	"$b%s$b is an invalid access level."
#define CSMSG_CHANGED_ACCESS	"%s is now %s $b%s$b in %s."

/* Channel CTCP reactions */
#define CSMSG_BAD_CTCP_REASON   "CTCPs to this channel are forbidden."

/* Channel note list */
#define CSMSG_NOTELIST_HEADER   "Notes for $b%s$b:"
#define CSMSG_NOTE_FORMAT       "%s (set by %s): %s"
#define CSMSG_NOTELIST_END      "End of notes for $b%s$b."
#define CSMSG_NOTELIST_EMPTY    "There are no (visible) notes for $b%s$b."
#define CSMSG_NO_SUCH_NOTE      "Channel $b%s$b does not have a note named $b%s$b."
#define CSMSG_BAD_NOTE_TYPE     "Note type $b%s$b does not exist."
#define CSMSG_NOTE_SET          "Note $b%s$b set in channel $b%s$b."
#define CSMSG_NOTE_REMOVED      "Note $b%s$b removed in channel $b%s$b."
#define CSMSG_BAD_NOTE_ACCESS   "$b%s$b is not a valid note access type."
#define CSMSG_BAD_MAX_LENGTH    "$b%s$b is not a valid maximum length (must be between 20 and 450 inclusive)."
#define CSMSG_NOTE_MODIFIED     "Note type $b%s$b modified."
#define CSMSG_NOTE_CREATED      "Note type $b%s$b created."
#define CSMSG_NOTE_TYPE_USED    "Note type $b%s$b is in use; give the FORCE argument to delete it."
#define CSMSG_NOTE_DELETED      "Note type $b%s$b deleted."
#define CSMSG_NOTE_SUSPENDED    "You may not view or set notes for suspended channels."

/* Channel [un]suspension */
#define CSMSG_ALREADY_SUSPENDED	"$b%s$b is already suspended."
#define CSMSG_NOT_SUSPENDED	"$b%s$b is not suspended."
#define CSMSG_SUSPENDED		"$b$C$b access to $b%s$b has been temporarily suspended."
#define CSMSG_SUSPENDED_REASON	"$b$C$b access to $b%s$b has been temporarily suspended (%s)."
#define CSMSG_UNSUSPENDED	"$b$C$b access to $b%s$b has been restored."
#define CSMSG_SUSPEND_NODELETE  "$b%s$b is protected from unregistration, and cannot be suspended."
#define CSMSG_USER_SUSPENDED    "$b%s$b's access to $b%s$b has been suspended."
#define CSMSG_USER_UNSUSPENDED  "$b%s$b's access to $b%s$b has been restored."

/* Access information */
#define CSMSG_IS_CHANSERV	"$b$C$b is the $bchannel service bot$b."
#define CSMSG_ACCESS_SELF_ONLY  "You may only see the list of infolines for yourself (by using $b%s$b with no arguments)."
#define CSMSG_SQUAT_ACCESS      "You do not have access to any channels."
#define CSMSG_INFOLINE_LIST     "Showing all channel entries for account $b%s$b:"
#define CSMSG_USER_NO_ACCESS    "%s lacks access to %s."
#define CSMSG_USER_HAS_ACCESS	"%s is %s $b%s$b in %s."
#define CSMSG_HELPER_NO_ACCESS  "%s lacks access to %s but has $bsecurity override$b enabled."
#define CSMSG_HELPER_HAS_ACCESS "%s is %s $b%s$b in %s and has $bsecurity override$b enabled."
#define CSMSG_LAZY_SMURF_TARGET "%s is %s ($bIRCOp$b; not logged in)."
#define CSMSG_SMURF_TARGET     	"%s is %s ($b%s$b)."
#define CSMSG_LAME_SMURF_TARGET "%s is %s %s."

/* Seen information */
#define CSMSG_NEVER_SEEN       	"%s has never been seen in $b%s$b."
#define CSMSG_USER_SEEN		"%s was last seen in $b%s$b %s ago."
#define CSMSG_USER_VACATION	"%s is currently on vacation."
#define CSMSG_USER_PRESENT     	"%s is in the channel $bright now$b."

/* Names information */
#define CSMSG_CHANNEL_NAMES     "Users in $b%s$b:%s"
#define CSMSG_END_NAMES         "End of names in $b%s$b"

/* Channel information */
#define CSMSG_CHANNEL_INFO	"$b%s$b Information:"
#define CSMSG_CHANNEL_TOPIC	"$bDefault Topic:       $b%s"
#define CSMSG_CHANNEL_MODES	"$bMode Lock:           $b%s"
#define CSMSG_CHANNEL_NOTE      "$b%s:%*s$b%s"
#define CSMSG_CHANNEL_MAX	"$bRecord Visitors:     $b%i"
#define CSMSG_CHANNEL_BANS	"$bBan Count:           $b%i"
#define CSMSG_CHANNEL_USERS	"$b%s Count:%*s$b%i"
#define CSMSG_CHANNEL_REGISTRAR "$bRegistrar:           $b%s"
#define CSMSG_CHANNEL_SUSPENDED "$b%s$b is suspended:"
#define CSMSG_CHANNEL_HISTORY   "Suspension history for $b%s$b:"
#define CSMSG_CHANNEL_SUSPENDED_0 " by %s: %s"
#define CSMSG_CHANNEL_SUSPENDED_1 " by %s; expires in %s: %s"
#define CSMSG_CHANNEL_SUSPENDED_2 " by %s; expired %s ago: %s"
#define CSMSG_CHANNEL_SUSPENDED_3 " by %s; revoked %s ago: %s"
#define CSMSG_CHANNEL_SUSPENDED_4 " %s ago by %s: %s"
#define CSMSG_CHANNEL_SUSPENDED_5 " %s ago by %s; expires in %s: %s"
#define CSMSG_CHANNEL_SUSPENDED_6 " %s ago by %s; expired %s ago: %s"
#define CSMSG_CHANNEL_SUSPENDED_7 " %s ago by %s; revoked %s ago: %s"
#define CSMSG_CHANNEL_REGISTERED "$bRegistered:          $b%s ago."
#define CSMSG_CHANNEL_VISITED	"$bVisited:             $b%s ago."

#define CSMSG_PEEK_INFO		"$b%s$b Status:"
#define CSMSG_PEEK_TOPIC       	"$bTopic:          $b%s"
#define CSMSG_PEEK_MODES       	"$bModes:          $b%s"
#define CSMSG_PEEK_USERS        "$bTotal users:    $b%d"
#define CSMSG_PEEK_OPS       	"$bOps:$b"

/* Network information */
#define CSMSG_NETWORK_INFO	"Network Information:"
#define CSMSG_NETWORK_SERVERS	"$bServers:             $b%i"
#define CSMSG_NETWORK_USERS	"$bTotal Users:         $b%i"
#define CSMSG_NETWORK_BANS	"$bTotal Ban Count:     $b%i"
#define CSMSG_NETWORK_OPERS	"$bIRC Operators:       $b%i"
#define CSMSG_NETWORK_CHANNELS	"$bRegistered Channels: $b%i"
#define CSMSG_SERVICES_UPTIME	"$bServices Uptime:     $b%s"
#define CSMSG_BURST_LENGTH	"$bLast Burst Length:   $b%s"

/* Staff list */
#define CSMSG_NETWORK_STAFF	"$bOnline Network Staff:$b"
#define CSMSG_STAFF_OPERS	"$bIRC Operators:$b"
#define CSMSG_STAFF_HELPERS	"$bHelpers:$b"

/* Channel searches */
#define CSMSG_ACTION_INVALID	"$b%s$b is not a recognized search action."
#define CSMSG_INVALID_CFLAG	"$b%s$b is not a recognized channel flag."
#define CSMSG_UNVISITED_HEADER	"Showing a maximum of %d channels unvisited for $b%s$b:"
#define CSMSG_UNVISITED_DATA	"%s: $b%s$b"
#define CSMSG_SEARCH_DATA	"%s"

/* Channel configuration */
#define CSMSG_INVALID_OPTION 	"$b%s$b is not a valid %s option."

#define CSMSG_CHANNEL_OPTIONS	"Channel Options:"
#define CSMSG_STRING_VALUE    	"$b%s$b%s"
#define CSMSG_NUMERIC_VALUE    	"$b%s$b%d - %s"

#define CSMSG_GREETING_TOO_LONG "Your greeting ($b%d$b characters) must be shorter than $b%d$b characters."

#define CSMSG_INVALID_STR_SET 	"$b'%s'$b is an invalid %s setting."
#define CSMSG_INVALID_NUM_SET 	"$b'%d'$b is an invalid %s setting."

#define CSMSG_UNF_RESPONSE      "I don't want to be part of your sick fantasies!"
#define CSMSG_PING_RESPONSE     "Pong!"
#define CSMSG_WUT_RESPONSE      "wut"
#define CSMSG_BAD_NUMBER        "$b%s$b is an invalid number.  Please use a number greater than 1 with this command."

/* User settings */
#define CSMSG_USER_OPTIONS	"User Options:"

/* eject_user and unban_user flags */
#define ACTION_KICK		0x0001
#define ACTION_BAN		0x0002
#define ACTION_ADD_BAN		0x0004
#define ACTION_ADD_TIMED_BAN   	0x0008
#define ACTION_UNBAN		0x0010
#define ACTION_DEL_BAN		0x0020

/* The 40 allows for [+-ntlksimprD] and lots of fudge factor. */
#define MODELEN			40 + KEYLEN
#define PADLEN			21
#define ACCESSLEN		10

#define CSMSG_NETRIDE_DETECTED  "User from far side of netsplit should have been banned - bye."
#define CSMSG_KICK_REASON	"Bye."
#define CSMSG_KICK_EXTRA	4
#define CSMSG_KICK_FORMAT	"%s (%s)"
#define CSMSG_PROTECT_REASON	"That user is protected."

#define CSFUNC_ARGS		user, channel, argc, argv, cmd

static unsigned char *CHANSERV_DEFAULT_MODES = "+nt";

#define CHANSERV_FUNC(NAME) MODCMD_FUNC(NAME)
#define CHANSERV_SYNTAX() 	svccmd_send_help(user, chanserv, cmd)
#define REQUIRE_PARAMS(N)	if(argc < (N)) {            \
	chanserv_notice(user, MSG_MISSING_PARAMS, argv[0]); \
	CHANSERV_SYNTAX(); \
	return 0; }

static int eject_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], struct svccmd *cmd, int action);

struct userNode *chanserv;
dict_t note_types;
static dict_t plain_dnrs, mask_dnrs, handle_dnrs;
static struct log_type *CS_LOG;

static struct
{
    struct channelList  support_channels;

    unsigned long 	db_backup_frequency;
    unsigned long 	channel_expire_frequency;

    long 		info_delay;
    unsigned int 	adjust_delay;
    long	 	channel_expire_delay;
    unsigned int        nodelete_level;

    unsigned int	adjust_threshold;
    int			join_flood_threshold;

    unsigned int 	greeting_length;
    unsigned int        refresh_period;

    unsigned int        max_owned;
    unsigned int 	max_chan_users;
    unsigned int 	max_chan_bans;
    struct string_list  *set_shows;

    struct string_list  *eightball;
    unsigned char       trigger_char;

    const char          *ctcp_short_ban_duration;
    const char          *ctcp_long_ban_duration;

    const char          *irc_operator_epithet;
    const char          *network_helper_epithet;
    const char          *support_helper_epithet;
} chanserv_conf;

struct listData
{
    struct userNode *user;
    struct chanNode *channel;
    const char      *search;
    enum userLevel  lowest;
    enum userLevel  highest;
    struct userData **users;
    struct helpfile_table table;
};

enum note_access_type
{
    NOTE_SET_CHANNEL_ACCESS,
    NOTE_SET_CHANNEL_SETTER,
    NOTE_SET_PRIVILEGED
};

enum note_visible_type
{
    NOTE_VIS_ALL,
    NOTE_VIS_CHANNEL_USERS,
    NOTE_VIS_PRIVILEGED
};

struct note_type
{
    enum note_access_type set_access_type;
    union {
        unsigned int     min_opserv;
        enum userLevel   min_ulevel;
    } set_access;
    enum note_visible_type visible_type;
    unsigned int         max_length;
    unsigned int         refs;
    char                 name[1];
};

struct note
{
    struct note_type     *type;
    char                 setter[NICKSERV_HANDLE_LEN+1];
    char                 note[1];
};

static unsigned int registered_channels;
static unsigned int userCount[USER_LEVEL_HIGHEST + 1];
static unsigned int banCount;

char *accessNames[] =
{
    "nothing",
    "peon",
    "op",
    "master",
    "coowner",
    "owner",
    "helper"
};

char *accessTitles[] =
{
    "BUG:",
    "Peon",
    "Op",
    "Master",
    "Coowner",
    "Owner",
    "BUG:"
};

char accessChars[] =
{
    'X',
    '+',
    '@',
    '%',
    '*',
    '!',
    'X'
};

int levelMap[] =
{
    0,
    100,
    200,
    300,
    400,
    500,
    600
};

const char *CSMSG_VERSION_ID = "$Id: chanserv.c,v 1.402 2003/12/23 13:59:16 entrope Exp $";

struct userData *helperList;
struct chanData *channelList;
static struct module *chanserv_module;
static struct service *chanserv_service;

#define chanserv_notice(target, format...) send_message(target , chanserv , ## format)

#define GetChannelUser(channel, handle) _GetChannelUser(channel, handle, 1, 0)
#define GetChannelAccess(channel, handle) _GetChannelUser(channel, handle, 0, 0)
#define GetTrueChannelAccess(channel, handle) _GetChannelUser(channel, handle, 0, 1)

enum userLevel
user_level_from_name(const char *name)
{
    enum userLevel level;

    for(level = USER_LEVEL_LOWEST; level <= USER_LEVEL_HIGHEST; level++)
    {
	if(!irccasecmp(name, accessNames[level]))
	{
	    return level;
	}
    }

    return ulNone;
}

const char*
user_level_from_level(enum userLevel level) {
    return accessNames[level];
}

struct userData*
_GetChannelUser(struct chanData *channel, struct handle_info *handle, int override, int allow_suspended)
{
    struct userData *uData, **head;

    if(!channel || !handle)
        return NULL;

    if(override && HANDLE_FLAGGED(handle, HELPING)
       && ((handle->opserv_level >= chanserv_conf.nodelete_level) || !IsProtected(channel)))
    {
	for(uData = helperList;
	    uData && uData->handle != handle;
	    uData = uData->next);

	if(!uData)
	{
	    uData = malloc(sizeof(struct userData));
	    uData->handle = handle;

	    uData->access = ulHelper;
	    uData->seen = 0;

	    uData->info = NULL;

	    uData->prev = NULL;
	    uData->next = helperList;
	    if(helperList)
		helperList->prev = uData;
	    helperList = uData;
	}

	head = &helperList;
    }
    else
    {
	for(uData = channel->users; uData; uData = uData->next)
            if((uData->handle == handle) && (allow_suspended || !IsUserSuspended(uData)))
                break;

	head = &(channel->users);
    }

    if(uData && (uData != *head))
    {
	/* Shuffle the user to the head of whatever list he was in. */
	if(uData->next)
            uData->next->prev = uData->prev;
	if(uData->prev)
            uData->prev->next = uData->next;

	uData->prev = NULL;
	uData->next = *head;

	if(*head)
	    (**head).prev = uData;
	*head = uData;
    }

    return uData;
}

/* Returns non-zero if user has at least VALUE access (where VALUE is
 * one of 'a', 'p', 'o', 'm', 'c', 'w' or 'n').  exempt_owner is set
 * when handling !set, so the owner can set things to/from 'n'.
 */
static int check_user_level(struct userNode *user, struct chanNode *channel, char value, int use_override, int exempt_owner)
{
    struct userData *uData;
    if(value == 'a') return 1;
    uData = _GetChannelUser(channel->channel_info, user->handle_info, use_override, 0);
    if(!uData) return 0;
    switch(value) {
    case 'c': return uData->access >= ulCoowner;
    case 'm': return uData->access >= ulMaster;
    case 'n': return exempt_owner ? (uData->access >= ulOwner) : (uData->access >= ulHelper);
    case 'o': return uData->access >= ulOp;
    case 'p': return uData->access >= ulPeon;
    case 'w': return uData->access >= ulOwner;
    default:  return 1;
    }
}

/* Scan for other users authenticated to the same handle
   still in the channel. If so, keep them listed as present.

   user is optional, if not null, it skips checking that userNode
   (for the handle_part function) */
static void
scan_handle_presence(struct chanNode *channel, struct handle_info *handle, struct userNode *user)
{
    struct userData *uData;

    if(!channel->channel_info || IsSuspended(channel->channel_info))
        return;

    uData = GetTrueChannelAccess(channel->channel_info, handle);
    if(uData)
    {
        struct modeNode *mn = find_handle_in_channel(channel, handle, user);

	if(mn)
	{
	    uData->present = 1;
	    uData->seen = now;
	}
	else
	{
	    uData->present = 0;
	}
    }
}

static void
chanserv_ctcp_check(struct userNode *user, struct chanNode *channel, char *text, UNUSED_ARG(struct userNode *bot))
{
    struct userData *uData;
    enum userLevel minlvl;
    unsigned int eflags, argc;
    unsigned char *argv[4];
    static char bad_ctcp_reason[] = CSMSG_BAD_CTCP_REASON;

    /* Bail early if channel is inactive or doesn't restrict CTCPs, or sender is a service */
    if(!channel->channel_info
       || IsSuspended(channel->channel_info)
       || (channel->channel_info->options[optCTCPUsers] == 'a')
       || IsService(user)
       || !ircncasecmp(text, "ACTION ", 7))
        return;
    /* Figure out the minimum level needed to CTCP the channel */
    switch(channel->channel_info->options[optCTCPUsers]) {
    default: case 'p': minlvl = ulPeon; break;
    case 'o': minlvl = ulOp; break;
    case 'm': minlvl = ulMaster; break;
    case 'c': minlvl = ulCoowner; break;
    case 'w': minlvl = ulOwner; break;
    case 'n': minlvl = INT_MAX; break;
    }
    /* Exit if they're allowed to CTCP the channel. */
    uData = GetChannelUser(channel->channel_info, user->handle_info);
    if(uData && (uData->access >= minlvl))
        return;
    /* We need to enforce against them; do so. */
    eflags = 0;
    argv[0] = text;
    argv[1] = user->nick;
    argc = 2;
    if(GetUserMode(channel, user))
        eflags |= ACTION_KICK;
    switch(channel->channel_info->options[optCTCPReaction]) {
    default: case 'k': /* just do the kick */ break;
    case 'b':
        eflags |= ACTION_BAN;
        break;
    case 't':
        eflags |= ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN;
        argv[argc++] = (unsigned char*)chanserv_conf.ctcp_short_ban_duration;
        break;
    case 'T':
        eflags |= ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN;
        argv[argc++] = (unsigned char*)chanserv_conf.ctcp_long_ban_duration;
        break;
    }
    argv[argc++] = bad_ctcp_reason;
    eject_user(chanserv, channel, argc, argv, NULL, eflags);
}

struct note_type *
chanserv_create_note_type(const char *name)
{
    struct note_type *ntype = calloc(1, sizeof(*ntype) + strlen(name));
    strcpy(ntype->name, name);
    ntype->refs = 1;
    dict_insert(note_types, ntype->name, ntype);
    return ntype;
}

static void
chanserv_deref_note_type(void *data)
{
    struct note_type *ntype = data;

    if(--ntype->refs > 0)
	return;
    free(ntype);
}

static void
chanserv_flush_note_type(struct note_type *ntype)
{
    struct chanData *cData;
    for (cData = channelList; cData; cData = cData->next)
        dict_remove(cData->notes, ntype->name);
}

static void
chanserv_truncate_notes(struct note_type *ntype)
{
    struct chanData *cData;
    struct note *note;
    unsigned int size = sizeof(*note) + ntype->max_length;
    
    for (cData = channelList; cData; cData = cData->next) {
        note = dict_find(cData->notes, ntype->name, NULL);
        if (!note)
            continue;
	if (strlen(note->note) > ntype->max_length) {
	    dict_remove2(cData->notes, ntype->name, 1);
	    note = realloc(note, size);
	    note->note[ntype->max_length] = 0;
	    dict_insert(cData->notes, ntype->name, note);
	}
    }
}

static int note_type_visible_to_user(struct chanData *channel, struct note_type *ntype, struct userNode *user);

static struct note *
chanserv_add_channel_note(struct chanData *channel, struct note_type *type, const char *setter, const char *text)
{
    struct note *note;
    unsigned int len = strlen(text);

    if(len > type->max_length) len = type->max_length;
    note = calloc(1, sizeof(*note) + len);
    note->type = type;
    strncpy(note->setter, setter, sizeof(note->setter)-1);
    memcpy(note->note, text, len);
    note->note[len] = 0;
    dict_insert(channel->notes, type->name, note);
    type->refs++;
    return note;
}

static void
chanserv_free_note(void *data)
{
    struct note *note = data;

    chanserv_deref_note_type(note->type);

    /* Reference count starts at one, so removing a note
       should _never_ cause a note type to be freed; that
       is, the only way to free a note type must be to use
       OperServ's delnote command. It's possible to
       assert(note->type->refs > 0) here. */
    free(note);
}

static MODCMD_FUNC(cmd_createnote) {
    struct note_type *ntype;
    unsigned int arg = 1, existed = 0, max_length;

    if ((ntype = dict_find(note_types, argv[1], NULL))) {
        existed = 1;
    } else {
	ntype = chanserv_create_note_type(argv[arg]);
    }
    if (!irccasecmp(argv[++arg], "privileged")) {
        arg++;
        ntype->set_access_type = NOTE_SET_PRIVILEGED;
        ntype->set_access.min_opserv = strtoul(argv[arg], NULL, 0);
    } else if (!irccasecmp(argv[arg], "channel")) {
        enum userLevel ulvl = user_level_from_name(argv[++arg]);
        if (ulvl == ulNone) {
            send_message(user, cmd->parent->bot, CSMSG_INVALID_ACCESS, argv[arg]);
	    goto fail;
        }
        ntype->set_access_type = NOTE_SET_CHANNEL_ACCESS;
        ntype->set_access.min_ulevel = ulvl;
    } else if (!irccasecmp(argv[arg], "setter")) {
        ntype->set_access_type = NOTE_SET_CHANNEL_SETTER;
    } else {
        send_message(user, cmd->parent->bot, CSMSG_BAD_NOTE_ACCESS, argv[arg]);
	goto fail;
    }

    if (!irccasecmp(argv[++arg], "privileged")) {
        ntype->visible_type = NOTE_VIS_PRIVILEGED;
    } else if (!irccasecmp(argv[arg], "channel_users")) {
        ntype->visible_type = NOTE_VIS_CHANNEL_USERS;
    } else if (!irccasecmp(argv[arg], "all")) {
        ntype->visible_type = NOTE_VIS_ALL;
    } else {
        send_message(user, cmd->parent->bot, CSMSG_BAD_NOTE_ACCESS, argv[arg]);
	goto fail;
    }

    if ((arg+1) >= argc) {
        send_message(user, cmd->parent->bot, MSG_MISSING_PARAMS, argv[0]);
	goto fail;
    }
    max_length = strtoul(argv[++arg], NULL, 0);
    if (max_length < 20 || max_length > 450) {
        send_message(user, cmd->parent->bot, CSMSG_BAD_MAX_LENGTH, argv[arg]);
	goto fail;
    }
    if(existed && (max_length < ntype->max_length)) {
	ntype->max_length = max_length;
	chanserv_truncate_notes(ntype);
    }
    ntype->max_length = max_length;

    if (existed) {
    	send_message(user, cmd->parent->bot, CSMSG_NOTE_MODIFIED, ntype->name);    
    } else {
        send_message(user, cmd->parent->bot, CSMSG_NOTE_CREATED, ntype->name);
    }
    return 1;

fail:
    if (!existed) {
	dict_remove(note_types, ntype->name);
    }
    return 0;
}

static MODCMD_FUNC(cmd_removenote) {
    struct note_type *ntype;
    int force;

    ntype = dict_find(note_types, argv[1], NULL);
    force = (argc > 2) && !irccasecmp(argv[2], "force");
    if (!ntype) {
        send_message(user, cmd->parent->bot, CSMSG_BAD_NOTE_TYPE, argv[1]);
        return 0;
    }
    if (ntype->refs > 1) {
        if (!force) {
            send_message(user, cmd->parent->bot, CSMSG_NOTE_TYPE_USED, ntype->name);
            return 0;
        }
        chanserv_flush_note_type(ntype);
    }
    dict_remove(note_types, argv[1]);
    send_message(user, cmd->parent->bot, CSMSG_NOTE_DELETED, argv[1]);
    return 1;
}

static int
parse_mode_lock(unsigned char *mode[], int count, struct chanData *channel)
{
    int lock = -1, argo = 1, pos = 0, limit = 0, key = 0;
    long on = 0, off = 0;

    if(count < 1) return 0;

#define lock_mode(FLAG) do { if(lock == 1) on |= FLAG; else if(lock == 0) off |= FLAG; } while(0)

    while(mode[0][pos])
    {
	switch(mode[0][pos++])
	{
	case '+':
	    lock = 1;
	    break;
	case '-':
	    lock = 0;
	    break;
	case 'n': lock_mode(MODE_NOPRIVMSGS); break;
	case 't': lock_mode(MODE_TOPICLIMIT); break;
	case 's': lock_mode(MODE_SECRET); break;
	case 'p': lock_mode(MODE_PRIVATE); break;
	case 'i': lock_mode(MODE_INVITEONLY); break;
	case 'm': lock_mode(MODE_MODERATED); break;
        case 'D': lock_mode(MODE_DELAYJOINS); break;
        case 'r': lock_mode(MODE_REGONLY); break;
        case 'c': lock_mode(MODE_NOCOLORS); break;
        case 'C': lock_mode(MODE_NOCTCPS); break;
	case 'l':
	    if(lock)
	    {
		if(argo >= count) return 0;

		on |= MODE_LIMIT;
		limit = argo++;
	    }
	    else
	    {
		off |= MODE_LIMIT;
	    }
	    break;
	case 'k':
	    if(lock)
	    {
		if(argo >= count) return 0;

		on |= MODE_KEY;
		key = argo++;
	    }
	    else
	    {
		off |= MODE_KEY;
	    }
	    break;
	default:
	    return 0;
	}
    }

    if(lock == -1) return 0;

    if(key) safestrncpy(channel->key, mode[key], sizeof(channel->key));
    if(limit)
    {
	channel->limit = atoi(mode[limit]);

	if(channel->limit < 1)
	{
	    return 0;
	}
    }

    /* do not record conflicting mode locks ('+p-p', for example). */
    channel->mode_on = on & ~off;
    channel->mode_off = off & ~on;

    return 1;
}

static void
create_mode_lock(unsigned char *buffer, struct chanData *channel, int display)
{
    int pos = 0, n;
    long mode = channel->mode_on;
    char *key = NULL;

    for(n=0; n<2; n++)
    {
	if(!mode)
	{
	    mode = channel->mode_off;
	    continue;
	}

	buffer[pos++] = n == 0 ? '+' : '-';

	if(mode & MODE_NOPRIVMSGS) buffer[pos++] = 'n';
	if(mode & MODE_TOPICLIMIT) buffer[pos++] = 't';
	if(mode & MODE_SECRET) buffer[pos++] = 's';
	if(mode & MODE_PRIVATE) buffer[pos++] = 'p';
	if(mode & MODE_INVITEONLY) buffer[pos++] = 'i';
	if(mode & MODE_MODERATED) buffer[pos++] = 'm';
        if(mode & MODE_DELAYJOINS) buffer[pos++] = 'D';
        if(mode & MODE_REGONLY) buffer[pos++] = 'r';
        if(mode & MODE_NOCOLORS) buffer[pos++] = 'c';
        if(mode & MODE_NOCTCPS) buffer[pos++] = 'C';

	/* parse_mode_lock() restricts limit to positive numbers. */
 	if(mode & MODE_LIMIT) buffer[pos++] = 'l';

        /* Figure out if the 'k' character should be included in
           the mode string and if a key should be appended. */
	if(mode & MODE_KEY)
        {
            if(mode == channel->mode_on || display)
            {
                /* If the key is getting locked on, or if the mode
                   string is getting displayed (key locked on or
                   off), include the 'k' character. */
                buffer[pos++] = 'k';
                if(mode == channel->mode_on)
                {
                    /* If the key is getting locked on, include the
                       actual key (whether the mode's displayed or
                       not). */
                    key = channel->key;
                }
            }
            else
            {
                /* The key is locked off and the mode string isn't
                   displayed, so add in the 'k' and key if there's
                   a key to remove. If not and the key is the only
                   mode locked off, back up over the '-'. */
                if(channel->channel && channel->channel->key[0])
                {
                    buffer[pos++] = 'k';
                    key = channel->channel->key;
                }
                else if(mode == MODE_KEY)
                {
                    pos--;
                }
            }
        }

	mode = channel->mode_off;
    }

    if((channel->mode_on & MODE_LIMIT) && channel->limit)
    {
	buffer[pos++] = ' ';

	pos += sprintf(buffer + pos, "%d", channel->limit);
    }

    if(key)
    {
	int len = strlen(key);
	buffer[pos++] = ' ';

	memcpy(buffer + pos, key, len);
	pos += len;
    }

    buffer[pos] = 0;
}

static int
mode_lock_violated(struct chanNode *channel)
{
    chan_mode_t on, off;

    on = channel->channel_info->mode_on;
    off = channel->channel_info->mode_off;

    if((on & ~channel->modes) || (channel->modes & off))
    {
        return 1;
    }

    if(on & MODE_KEY)
    {
        if(strcmp(channel->key, channel->channel_info->key))
            return 1;
    }

    if(on & MODE_LIMIT)
    {
        if(channel->limit != channel->channel_info->limit)
            return 1;
    }

    return 0;
}

static char max_length_text[MAXLEN+1][16];

static struct helpfile_expansion
chanserv_expand_variable(const char *variable)
{
    struct helpfile_expansion exp;

    if(!irccasecmp(variable, "notes"))
    {
        dict_iterator_t it;
        exp.type = HF_TABLE;
        exp.value.table.length = 1;
        exp.value.table.width = 3;
        exp.value.table.flags = 0;
        exp.value.table.contents = calloc(dict_size(note_types)+1, sizeof(char**));
        exp.value.table.contents[0] = calloc(exp.value.table.width, sizeof(char*));
        exp.value.table.contents[0][0] = "Note Type";
        exp.value.table.contents[0][1] = "Visibility";
        exp.value.table.contents[0][2] = "Max Length";
        for (it=dict_first(note_types); it; it=iter_next(it))
        {
            struct note_type *ntype = iter_data(it);
            int row;

            if(!note_type_visible_to_user(NULL, ntype, message_dest)) continue;
            row = exp.value.table.length++;
            exp.value.table.contents[row] = calloc(exp.value.table.width, sizeof(char*));
            exp.value.table.contents[row][0] = ntype->name;
            exp.value.table.contents[row][1] = (ntype->visible_type == NOTE_VIS_ALL) ? "all" :
                (ntype->visible_type == NOTE_VIS_CHANNEL_USERS) ? "chan users" :
                "unknown";
            if(!max_length_text[ntype->max_length][0])
            {
                snprintf(max_length_text[ntype->max_length], sizeof(max_length_text[ntype->max_length]),
                         "%u", ntype->max_length);
            }
            exp.value.table.contents[row][2] = max_length_text[ntype->max_length];
        }
        return exp;
    }

    exp.type = HF_STRING;
    exp.value.str = NULL;
    return exp;
}

static struct chanData*
register_channel(struct chanNode *cNode, char *registrar, char *options, long flags)
{
    struct chanData *channel;

    channel = calloc(1, sizeof(struct chanData));

    channel->notes = dict_new();
    dict_set_free_data(channel->notes, chanserv_free_note);

    channel->registrar = registrar ? strdup(registrar) : NULL;
    channel->registered = now;
    channel->visited = now;
    channel->limitAdjusted = now;
    channel->flags = flags ? flags : CHANNEL_DEFAULT_FLAGS;

    safestrncpy(channel->options, options ? options : CHANNEL_DEFAULT_OPTIONS, sizeof(channel->options));

    if(strlen(channel->options) < (OPTION_STRING_SIZE - 1))
    {
	/* Use a sane default from the DEFAULT_OPTIONS in the
	   case of missing options. */
	memcpy(channel->options + strlen(channel->options), CHANNEL_DEFAULT_OPTIONS + strlen(channel->options), (OPTION_STRING_SIZE - 1) - strlen(channel->options));
    }

    channel->prev = NULL;
    channel->next = channelList;

    if(channelList)
    {
	channelList->prev = channel;
    }
    channelList = channel;

    channel->channel = cNode;
    LockChannel(cNode);
    cNode->channel_info = channel;

    registered_channels++;

    return channel;
}

static struct userData*
add_channel_user(struct chanData *channel, struct handle_info *handle, enum userLevel access, time_t seen, const char *info)
{
    struct userData *ud;

    if(access > USER_LEVEL_HIGHEST || access < USER_LEVEL_LOWEST)
	return NULL;

    ud = calloc(1, sizeof(*ud));
    ud->channel = channel;
    ud->handle = handle;
    ud->seen = seen;
    ud->access = access;
    ud->info = info ? strdup(info) : NULL;

    ud->prev = NULL;
    ud->next = channel->users;
    if(channel->users)
	channel->users->prev = ud;
    channel->users = ud;

    channel->userCount[0]++;
    channel->userCount[access]++;
    userCount[0]++;
    userCount[access]++;

    ud->u_prev = NULL;
    ud->u_next = ud->handle->channels;
    if(ud->u_next)
        ud->u_next->u_prev = ud;
    ud->handle->channels = ud;

    return ud;
}

static void unregister_channel(struct chanData *channel, const char *reason);

void
del_channel_user(struct userData *user, int do_gc)
{
    struct chanData *channel = user->channel;

    channel->userCount[0]--;
    channel->userCount[user->access]--;
    userCount[0]--;
    userCount[user->access]--;

    if(user->prev)
        user->prev->next = user->next;
    else
        channel->users = user->next;
    if(user->next)
        user->next->prev = user->prev;

    if(user->u_prev)
        user->u_prev->u_next = user->u_next;
    else
        user->handle->channels = user->u_next;
    if(user->u_next)
        user->u_next->u_prev = user->u_prev;

    free(user->info);
    free(user);
    if(do_gc && !channel->users && !IsProtected(channel))
        unregister_channel(channel, "lost all users.");
}

static void expire_ban(void *data);

static struct banData*
add_channel_ban(struct chanData *channel, const char *mask, char *owner, time_t set, time_t triggered, time_t expires, char *reason)
{
    struct banData *bd;

    if(!mask)
        return NULL;

    bd = malloc(sizeof(struct banData));

    bd->channel = channel;
    bd->set = set;
    bd->triggered = triggered;
    bd->expires = expires;

    safestrncpy(bd->mask, mask, sizeof(bd->mask));
    if(owner)
        safestrncpy(bd->owner, owner, sizeof(bd->owner));
    bd->reason = reason ? strdup(reason) : NULL;

    if(expires)
    {
	timeq_add(expires, expire_ban, bd);
    }

    bd->prev = NULL;
    bd->next = channel->bans;
    if(channel->bans)
    {
	channel->bans->prev = bd;
    }
    channel->bans = bd;
    channel->banCount++;
    banCount++;

    return bd;
}

static void
del_channel_ban(struct banData *ban)
{
    ban->channel->banCount--;
    banCount--;

    if(ban->prev)
    {
        ban->prev->next = ban->next;
    }
    else
    {
        ban->channel->bans = ban->next;
    }

    if(ban->next)
    {
        ban->next->prev = ban->prev;
    }

    if(ban->expires)
    {
	timeq_del(0, expire_ban, ban, TIMEQ_IGNORE_WHEN);
    }

    if(ban->reason)
    {
        free(ban->reason);
    }

    free(ban);
}

static void
expire_ban(void *data)
{
    struct banData *bd;
    struct banList bans;
    struct chanData *channel;
    struct string_list *removed;
    unsigned int i;
    char *ban;

    bd = data;
    channel = bd->channel;

    if(IsSuspended(channel))
    {
	bd->expires = 0;
	del_channel_ban(bd);
	return;
    }

    bans = channel->channel->banlist;
    removed = alloc_string_list(2);
    for(i=0; i<bans.used; i++)
    {
	if(!strcmp(bans.list[i]->ban, bd->mask))
	{
	    ban = strdup(bans.list[i]->ban);
	    string_list_append(removed, ban);
	}
    }

    bd->expires = 0;
    del_channel_ban(bd);

    if(removed->used)
    {
	DelChannelBan(removed->used, (const char **)removed->list, channel->channel, chanserv, 1);
    }

    free_string_list(removed);
}

static void chanserv_expire_suspension(void *data);

static void
unregister_channel(struct chanData *channel, const char *reason)
{
    char msgbuf[MAXLEN];

    /* After channel unregistration, the following must be cleaned
       up:
       - Channel information.
       - Channel users.
       - Channel bans.
       - Channel suspension data.
       - Timeq entries. (Except timed bans, which are handled elsewhere.)
    */

    if(!channel)
    {
	return;
    }

    timeq_del(0, NULL, channel, TIMEQ_IGNORE_FUNC | TIMEQ_IGNORE_WHEN);

    while(channel->users)
    {
	del_channel_user(channel->users, 0);
    }

    while(channel->bans)
    {
	del_channel_ban(channel->bans);
    }

    if(channel->registrar) free(channel->registrar);
    if(channel->greeting) free(channel->greeting);
    if(channel->user_greeting) free(channel->user_greeting);
    if(channel->topic_mask) free(channel->topic_mask);

    if(channel->prev) channel->prev->next = channel->next;
    else channelList = channel->next;

    if(channel->next) channel->next->prev = channel->prev;

    if(channel->suspended)
    {
	struct chanNode *cNode = channel->channel;
	struct suspended *suspended, *next_suspended;

        for(suspended = channel->suspended; suspended; suspended = next_suspended)
        {
            next_suspended = suspended->previous;
            free(suspended->suspender);
            free(suspended->reason);
            if(suspended->expires)
                timeq_del(suspended->expires, chanserv_expire_suspension, suspended, 0);
            free(suspended);
        }

	if(cNode)
	    cNode->channel_info = NULL;
    }
    channel->channel->channel_info = NULL;

    if(channel->notes)
        dict_delete(channel->notes);
    sprintf(msgbuf, "%s %s", channel->channel->name, reason);
    if(!IsSuspended(channel))
        DelChannelUser(chanserv, channel->channel, msgbuf, 0);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, msgbuf);
    UnlockChannel(channel->channel);
    free(channel);
    registered_channels--;
}

static void
expire_channels(UNUSED_ARG(void *data))
{
    struct chanData *channel, *next;
    struct userData *user;
    char delay[INTERVALLEN], reason[INTERVALLEN + 64];

    intervalString(delay, chanserv_conf.channel_expire_delay);
    sprintf(reason, "Channel registration automatically expired after %s of disuse.", delay);

    for(channel = channelList; channel; channel = next)
    {
	next = channel->next;

        /* See if the channel can be expired. */
        if(((now - channel->visited) <= chanserv_conf.channel_expire_delay)
           || IsProtected(channel))
            continue;

        /* Make sure there are no high-ranking users still in the channel. */
        for(user=channel->users; user; user=user->next)
            if(user->present && (user->access >= ulMaster))
                break;
        if(user)
            continue;

        /* Unregister the channel */
        log_module(CS_LOG, LOG_INFO, "(%s) Channel registration expired.", channel->channel->name);
        unregister_channel(channel, "registration expired.");
    }

    if(chanserv_conf.channel_expire_frequency)
    {
	timeq_add(now + chanserv_conf.channel_expire_frequency, expire_channels, NULL);
    }
}

static int
protect_user(struct userNode *victim, struct userNode *aggressor, struct chanData *channel)
{
    char protect = channel->options[optProtect];
    struct userData *cs_victim, *cs_aggressor;

    /* Don't protect if no one is to be protected, someone is attacking
       himself, or if the aggressor is an IRC Operator. */
    if(protect == 'n' || victim == aggressor || IsOper(aggressor))
    {
	return 0;
    }

    /* Don't protect if the victim isn't authenticated (because they
       can't be a channel user), unless we are to protect non-users
       also. */
    cs_victim = GetChannelAccess(channel, victim->handle_info);
    if(protect != 'a' && !cs_victim)
    {
        return 0;
    }

    /* Protect if the aggressor isn't a user because at this point,
       the aggressor can only be less than or equal to the victim. */
    cs_aggressor = GetChannelAccess(channel, aggressor->handle_info);
    if(!cs_aggressor)
    {
        return 1;
    }

    /* If the aggressor was a user, then the victim can't be helped. */
    if(!cs_victim)
    {
        return 0;
    }

    switch(protect)
    {
    case 'l':
	if(cs_victim->access > cs_aggressor->access)
            return 1;
	break;
    case 'a':
    case 'e':
	if(cs_victim->access >= cs_aggressor->access)
            return 1;
	break;
    }

    return 0;
}

static int
validate_op(struct userNode *user, struct chanNode *channel, struct userNode *victim)
{
    struct userData *cs_victim;

    if((!(cs_victim = GetChannelUser(channel->channel_info, victim->handle_info))
        || (cs_victim->access < ulOp))
       && !check_user_level(user, channel, channel->channel_info->options[optStrictOp], 1, 0))
    {
	chanserv_notice(user, CSMSG_OPBY_LOCKED);
	return 0;
    }

    return 1;
}

static int
validate_auto_op(struct userNode *user, struct chanNode *channel)
{
    struct userData *uData;

    if(!check_user_level(user, channel, channel->channel_info->options[optAutoOp], 0, 0))
        return 0;
    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    /* uData may be NULL if optAutoOp is "all" */
    return uData ? IsUserAutoOp(uData) : !channel->join_flooded;
}

static int
validate_deop(struct userNode *user, struct chanNode *channel, struct userNode *victim)
{
    if(IsService(victim))
    {
	chanserv_notice(user, MSG_SERVICE_IMMUNE, victim->nick);
	return 0;
    }

    if(protect_user(victim, user, channel->channel_info))
    {
	chanserv_notice(user, CSMSG_USER_PROTECTED, victim->nick);
	return 0;
    }

    return 1;
}

#define chanserv_get_handle_info(USER, NAME) smart_get_handle_info(chanserv, USER, NAME)

static struct do_not_register *
chanserv_add_dnr(const char *chan_name, const char *setter, const char *reason)
{
    struct do_not_register *dnr = calloc(1, sizeof(*dnr)+strlen(reason));
    safestrncpy(dnr->chan_name, chan_name, sizeof(dnr->chan_name));
    safestrncpy(dnr->setter, setter, sizeof(dnr->setter));
    strcpy(dnr->reason, reason);
    if(dnr->chan_name[0] == '*')
    {
        dict_insert(handle_dnrs, dnr->chan_name+1, dnr);
    }
    else if(strpbrk(dnr->chan_name, "*?"))
    {
        dict_insert(mask_dnrs, dnr->chan_name, dnr);
    }
    else
    {
        dict_insert(plain_dnrs, dnr->chan_name, dnr);
    }
    return dnr;
}

struct do_not_register *
chanserv_is_dnr(const char *chan_name, struct handle_info *handle)
{
    struct do_not_register *dnr;
    dict_iterator_t it;

    if(handle)
    {
        if((dnr = dict_find(handle_dnrs, handle->handle, NULL)))
            return dnr;
    }
    if(chan_name)
    {
        if((dnr = dict_find(plain_dnrs, chan_name, NULL)))
            return dnr;
        for(it = dict_first(mask_dnrs); it; it = iter_next(it))
        {
            if(match_ircglob(chan_name, iter_key(it)))
                return iter_data(it);
        }
    }
    return NULL;
}

#define IsCSChannelName(STR) (IsChannelName(STR) && ((STR)[0] == '#'))

static CHANSERV_FUNC(cmd_noregister)
{
    const char *target;
    struct do_not_register *dnr;

    if(argc < 2)
    {
        dict_iterator_t it;
        unsigned int matches = 0;

        reply(MSG_SEARCH_RESULTS, "do-not-registers");
        for(it = dict_first(handle_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            reply(CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
            matches++;
        }
        for(it = dict_first(plain_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            reply(CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
            matches++;
        }
        for(it = dict_first(mask_dnrs); it; it = iter_next(it))
        {
            dnr = iter_data(it);
            reply(CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
            matches++;
        }

        if(matches)
        {
            reply(MSG_MATCH_COUNT, matches);
        }
        else
        {
            reply(MSG_NO_MATCHES);
        }
        return 0;
    }

    if(IsCSChannelName(argv[1]) || (*argv[1] == '*'))
    {
        target = argv[1];

        if(argc > 2)
        {
            const char *reason = unsplit_string(argv + 2, argc - 2, NULL);
            if((*target == '*') && !get_handle_info(target + 1))
            {
                reply(MSG_HANDLE_UNKNOWN, target + 1);
                return 0;
            }
            chanserv_add_dnr(target, user->handle_info->handle, reason);
            reply(CSMSG_NOREGISTER_CHANNEL, target);
            return 1;
        }

        if(*target == '*')
        {
            struct handle_info *handle;

            handle = get_handle_info(target + 1);
            dnr = chanserv_is_dnr(NULL, handle);
        }
        else
        {
            dnr = chanserv_is_dnr(target, NULL);
        }

        if(dnr)
        {
            reply(CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        else
        {
            reply(MSG_SEARCH_RESULTS, "do-not-registers");
            reply(MSG_NO_MATCHES);
        }
        return 0;
    }

    reply(CSMSG_NOT_DNR, argv[1]);
    return 0;
}

static CHANSERV_FUNC(cmd_allowregister)
{
    const char *chan_name = argv[1];

    if((chan_name[0] == '*') && dict_find(handle_dnrs, chan_name+1, NULL))
    {
        dict_remove(handle_dnrs, chan_name+1);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else if(dict_find(plain_dnrs, chan_name, NULL))
    {
        dict_remove(plain_dnrs, chan_name);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else if(dict_find(mask_dnrs, chan_name, NULL))
    {
        dict_remove(mask_dnrs, chan_name);
        chanserv_notice(user, CSMSG_DNR_REMOVED, chan_name);
    }
    else
    {
        chanserv_notice(user, CSMSG_NO_SUCH_DNR, chan_name);
        return 0;
    }
    return 1;
}

unsigned int
chanserv_get_owned_count(struct handle_info *hi)
{
    struct userData *cList;
    unsigned int owned;

    for(owned=0, cList=hi->channels; cList; cList=cList->u_next)
    {
        if(cList->access == ulOwner)
            owned++;
    }
    return owned;
}

static CHANSERV_FUNC(cmd_register)
{
    struct handle_info *handle;
    struct chanData *cData;
    struct modeNode *mn;
    char reason[MAXLEN];
    char *chan_name;
    unsigned int new_channel, force=0;
    struct do_not_register *dnr;

    if(channel)
    {
        if(channel->channel_info)
        {
            chanserv_notice(user, CSMSG_ALREADY_REGGED, channel->name);
            return 0;
        }

        if(channel->bad_channel)
        {
            chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, channel->name);
            return 0;
        }

        if(!IsHelping(user) && (!(mn = GetUserMode(channel, user)) || !(mn->modes & MODE_CHANOP)))
        {
            chanserv_notice(user, CSMSG_MUST_BE_OPPED, channel->name);
            return 0;
        }

        new_channel = 0;
        chan_name = channel->name;
    }
    else
    {
        if((argc < 2) || !IsCSChannelName(argv[1]))
        {
            chanserv_notice(user, MSG_NOT_CHANNEL_NAME);
            return 0;
        }

        if(opserv_bad_channel(argv[1]))
        {
            chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, argv[1]);
            return 0;
        }

        new_channel = 1;
        chan_name = argv[1];
    }

    if(argc >= (new_channel+2))
    {
	if(!IsHelping(user))
	{
	    chanserv_notice(user, CSMSG_PROXY_FORBIDDEN);
	    return 0;
	}

	if(!(handle = chanserv_get_handle_info(user, argv[new_channel+1]))) return 0;
        force = (argc > (new_channel+2)) && !irccasecmp(argv[new_channel+2], "force");
        dnr = chanserv_is_dnr(chan_name, handle);
    }
    else
    {
	handle = user->handle_info;
        dnr = chanserv_is_dnr(chan_name, handle);
    }
    if(dnr && !force)
    {
        if(IsHelping(user))
        {
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        else
        {
                chanserv_notice(user, CSMSG_DNR_CHANNEL, chan_name);
        }
        return 0;
    }


    if((chanserv_get_owned_count(handle) >= chanserv_conf.max_owned) && !force)
    {
        chanserv_notice(user, CSMSG_OWN_TOO_MANY, handle->handle, chanserv_conf.max_owned);
        return 0;
    }

    if(new_channel)
    {
        channel = AddChannel(argv[1], now, NULL, NULL);
        AddChannelUser(chanserv, channel);
    }

    cData = register_channel(channel, user->handle_info->handle, NULL, 0);
    add_channel_user(cData, handle, USER_LEVEL_HIGHEST, 0, NULL);
    scan_handle_presence(channel, handle, NULL);

    if(!new_channel) AddChannelUser(chanserv, channel);
    AddChannelOp(1, &chanserv, channel, chanserv, !new_channel);

    parse_mode_lock(&CHANSERV_DEFAULT_MODES, 1, cData);
    mod_chanmode(channel, chanserv, CHANSERV_DEFAULT_MODES);
    irc_mode(chanserv, channel, CHANSERV_DEFAULT_MODES);

    /* Initialize the channel's max user record. */
    cData->max = channel->members.used;

    if(handle != user->handle_info)
    {
        chanserv_notice(user, CSMSG_PROXY_SUCCESS, handle->handle, channel->name);
    }
    else
    {
        chanserv_notice(user, CSMSG_REG_SUCCESS, channel->name);
    }

    sprintf(reason, "%s registered to %s by %s.", channel->name, handle->handle, user->handle_info->handle);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);
    return 1;
}

static const char *
make_confirmation_string(struct userData *uData)
{
    static char strbuf[16];
    char *src;
    unsigned int accum;

    accum = 0;
    for(src = uData->handle->handle; *src; )
    {
        accum = accum * 31 + toupper(*src++);
    }
    for(src = uData->channel->channel->name; *src; )
    {
        accum = accum * 31 + toupper(*src++);
    }
    sprintf(strbuf, "%08x", accum);
    return strbuf;
}

static CHANSERV_FUNC(cmd_unregister)
{
    char *name;
    char reason[MAXLEN];
    struct chanData *cData;
    struct userData *uData;

    cData = channel->channel_info;
    if(!cData)
    {
        reply(CSMSG_NOT_REGISTERED, channel->name);
        return 0;
    }

    uData = GetChannelUser(cData, user->handle_info);
    if(!uData || (uData->access < USER_LEVEL_HIGHEST))
    {
        reply(CSMSG_NO_ACCESS);
        return 0;
    }

    if(IsProtected(cData))
    {
        chanserv_notice(user, CSMSG_UNREG_NODELETE, channel->name);
        return 0;
    }

    if(!IsHelping(user))
    {
        const char *confirm_string;
        if(IsSuspended(cData))
        {
            reply(CSMSG_CHAN_SUSPENDED, channel->name, cData->suspended->reason);
            return 0;
        }
        confirm_string = make_confirmation_string(uData);
	if((argc < 2) || strcmp(argv[1], confirm_string))
	{
	    reply(CSMSG_CONFIRM_UNREG, confirm_string);
	    return 0;
	}
    }

    sprintf(reason, "unregistered by %s.", user->handle_info->handle);
    name = strdup(channel->name);
    unregister_channel(cData, reason);
    reply(CSMSG_UNREG_SUCCESS, name);
    free(name);
    return 1;
}

static CHANSERV_FUNC(cmd_move)
{
    struct chanNode *target;
    struct modeNode *mn;
    struct userData *uData;
    char reason[MAXLEN];
    struct do_not_register *dnr;

    REQUIRE_PARAMS(2);

    if(IsProtected(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_MOVE_NODELETE, channel->name);
	return 0;
    }

    if(!IsCSChannelName(argv[1]))
    {
        chanserv_notice(user, MSG_NOT_CHANNEL_NAME);
        return 0;
    }

    if(opserv_bad_channel(argv[1]))
    {
        chanserv_notice(user, CSMSG_ILLEGAL_CHANNEL, argv[1]);
        return 0;
    }

    if (!IsHelping(user) || (argc < 3) || irccasecmp(argv[2], "force"))
    {
        for(uData = channel->channel_info->users; uData; uData = uData->next)
        {
            if((uData->access == ulOwner) && (dnr = chanserv_is_dnr(argv[1], uData->handle)))
            {
                if(IsHelping(user))
                {
                    chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
                }
                else
                {
                    chanserv_notice(user, CSMSG_DNR_CHANNEL_MOVE, argv[1]);
                }
                return 0;
            }
        }
    }

    if (!(target = GetChannel(argv[1])))
    {
        target = AddChannel(argv[1], now, NULL, NULL);
        LockChannel(target);
        if(!IsSuspended(channel->channel_info))
        {
            AddChannelUser(chanserv, target);
        }
    }
    else if(target->channel_info)
    {
	chanserv_notice(user, CSMSG_ALREADY_REGGED, target->name);
	return 0;
    }
    else if((!(mn = GetUserMode(target, user)) || !(mn->modes && MODE_CHANOP))
            && !IsHelping(user))
    {
        chanserv_notice(user, CSMSG_MUST_BE_OPPED, target->name);
        return 0;
    }
    else if(!IsSuspended(channel->channel_info))
    {
	AddChannelUser(chanserv, target);
        AddChannelOp(1, &chanserv, target, chanserv, 1);
    }

    /* Move the channel_info to the target channel; it
       shouldn't be necessary to clear timeq callbacks
       for the old channel. */
    target->channel_info = channel->channel_info;
    target->channel_info->channel = target;
    channel->channel_info = NULL;

    chanserv_notice(user, CSMSG_MOVE_SUCCESS, target->name);

    sprintf(reason, "%s moved to %s by %s.", channel->name, target->name, user->handle_info->handle);
    if(!IsSuspended(target->channel_info))
    {
        char reason2[MAXLEN];
	sprintf(reason2, "Channel moved to %s by %s.", target->name, user->handle_info->handle);
	DelChannelUser(chanserv, channel, reason2, 0);
    }
    UnlockChannel(channel);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);
    return 1;
}

static void
merge_users(struct chanData *source, struct chanData *target)
{
    struct userData *suData, *tuData, *next;
    dict_iterator_t it;
    dict_t merge;

    merge = dict_new();

    /* Insert the source's users into the scratch area. */
    for(suData = source->users; suData; suData = suData->next)
    {
	dict_insert(merge, suData->handle->handle, suData);
    }

    /* Iterate through the target's users, looking for
       users common to both channels. The lower access is
       removed from either the scratch area or target user
       list. */
    for(tuData = target->users; tuData; tuData = next)
    {
	struct userData *choice;

	next = tuData->next;

	/* If a source user exists with the same handle as a target
	   channel's user, resolve the conflict by removing one. */
	suData = dict_find(merge, tuData->handle->handle, NULL);
	if(!suData)
	{
	    continue;
	}

	/* Pick the data we want to keep. */
	if(suData->access == tuData->access)
	{
	    /* If the access is the same, use the later seen time. */
	    choice = (suData->seen > tuData->seen) ? suData : tuData;
	}
	else
	{
	    /* Otherwise, keep the higher access level. */
	    choice = (suData->access > tuData->access) ? suData : tuData;
	}

	/* Remove the user that wasn't picked. */
	if(choice == tuData)
	{
	    dict_remove(merge, suData->handle->handle);
	    del_channel_user(suData, 0);
	}
	else
	{
	    del_channel_user(tuData, 0);
	}
    }

    /* Move the remaining users to the target channel. */
    for(it = dict_first(merge); it; it = iter_next(it))
    {
	suData = iter_data(it);

	/* Insert the user into the target channel's linked list. */
	suData->prev = NULL;
	suData->next = target->users;
        suData->channel = target;

	if(target->users)
	{
	    target->users->prev = suData;
	}
	target->users = suData;

	/* Update the user counts for the target channel; the
	   source counts are left alone. */
	target->userCount[0]++;
	target->userCount[suData->access]++;
    }

    /* Possible to assert (source->users == NULL) here. */
    source->users = NULL;
    dict_delete(merge);
}

static void
merge_bans(struct chanData *source, struct chanData *target)
{
    struct banData *sbData, *tbData, *sNext, *tNext, *tFront;

    /* Hold on to the original head of the target ban list
       to avoid comparing source bans with source bans. */
    tFront = target->bans;

    /* Perform a totally expensive O(n*m) merge, ick. */
    for(sbData = source->bans; sbData; sbData = sNext)
    {
	/* Flag to track whether the ban's been moved
	   to the destination yet. */
	int moved = 0;

	/* Possible to assert (sbData->prev == NULL) here. */
	sNext = sbData->next;

	for(tbData = tFront; tbData; tbData = tNext)
	{
	    tNext = tbData->next;

	    /* Perform two comparisons between each source
	       and target ban, conflicts are resolved by
	       keeping the broader ban and copying the later
	       expiration and triggered time. */
	    if(match_ircglobs(tbData->mask, sbData->mask))
	    {
		/* There is a broader ban in the target channel that
		   overrides one in the source channel; remove the 
		   source ban and break. */
		if(sbData->expires > tbData->expires)
		{
		    tbData->expires = sbData->expires;
		}
		if(sbData->triggered > tbData->triggered)
		{
		    tbData->triggered = sbData->triggered;
		}
		del_channel_ban(sbData);
		break;
	    }
	    else if(match_ircglobs(sbData->mask, tbData->mask))
	    {
		/* There is a broader ban in the source channel that
		   overrides one in the target channel; remove the
		   target ban, fall through and move the source over. */
		if(tbData->expires > sbData->expires)
		{
		    sbData->expires = tbData->expires;
		}
		if(tbData->triggered > sbData->triggered)
		{
		    sbData->triggered = tbData->triggered;
		}
		if(tbData == tFront)
		{
		    tFront = tNext;
		}
		del_channel_ban(tbData);
	    }

	    /* Source bans can override multiple target bans, so
	       we allow a source to run through this loop multiple
	       times, but we can only move it once. */
	    if(moved)
	    {
		continue;
	    }
	    moved = 1;

	    /* Remove the source ban from the source ban list. */
	    if(sbData->next)
	    {
		sbData->next->prev = sbData->prev;
	    }

	    /* Modify the source ban's associated channel. */
	    sbData->channel = target;

	    /* Insert the ban into the target channel's linked list. */
	    sbData->prev = NULL;
	    sbData->next = target->bans;

	    if(target->bans)
	    {
		target->bans->prev = sbData;
	    }
	    target->bans = sbData;

	    /* Update the user counts for the target channel. */
	    target->banCount++;
	}
    }

    /* Possible to assert (source->bans == NULL) here. */
    source->bans = NULL;
}

static void
merge_data(struct chanData *source, struct chanData *target)
{
    if(source->visited > target->visited)
    {
	target->visited = source->visited;
    }
}

static void
merge_channel(struct chanData *source, struct chanData *target)
{
    merge_users(source, target);
    merge_bans(source, target);
    merge_data(source, target);
}

static CHANSERV_FUNC(cmd_merge)
{
    struct userData *target_user;
    struct chanNode *target;
    char reason[MAXLEN];

    REQUIRE_PARAMS(2);

    /* Make sure the target channel exists and is registered to the user
       performing the command. */
    if(!(target = GetChannel(argv[1])))
    {
        chanserv_notice(user, MSG_INVALID_CHANNEL);
        return 0;
    }

    if(!target->channel_info)
    {
        chanserv_notice(user, CSMSG_NOT_REGISTERED, target->name);
        return 0;
    }

    if(IsProtected(channel->channel_info))
    {
        chanserv_notice(user, CSMSG_MERGE_NODELETE);
        return 0;
    }

    if(IsSuspended(target->channel_info))
    {
        chanserv_notice(user, CSMSG_MERGE_SUSPENDED);
        return 0;
    }

    if(channel == target)
    {
        chanserv_notice(user, CSMSG_MERGE_SELF);
        return 0;
    }

    target_user = GetChannelUser(target->channel_info, user->handle_info);
    if(!target_user || (target_user->access < ulOwner))
    {
        chanserv_notice(user, CSMSG_MERGE_NOTOWNER);
        return 0;
    }

    /* Merge the channel structures and associated data. */
    merge_channel(channel->channel_info, target->channel_info);
    sprintf(reason, "merged into %s by %s.", target->name, user->handle_info->handle);
    unregister_channel(channel->channel_info, reason);
    chanserv_notice(user, CSMSG_MERGE_SUCCESS, target->name);
    return 1;
}

static CHANSERV_FUNC(cmd_opchan)
{
    if(!IsHelping(user) && !channel->channel_info->may_opchan)
    {
        reply(CSMSG_ALREADY_OPCHANNED, channel->name);
        return 0;
    }
    channel->channel_info->may_opchan = 0;
    AddChannelOp(1, &chanserv, channel, chanserv, 1);
    reply(CSMSG_OPCHAN_DONE, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_adduser)
{
    struct userData *actee;
    struct userData *actor;
    struct handle_info *handle;
    enum userLevel access;

    REQUIRE_PARAMS(3);

    if(channel->channel_info->userCount[0] >= chanserv_conf.max_chan_users)
    {
	chanserv_notice(user, CSMSG_STRUCTURE_LIMIT, "user", chanserv_conf.max_chan_users);
	return 0;
    }

    access = user_level_from_name(argv[1]);
    if(access == ulNone)
    {
	chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[1]);
	return 0;
    }

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access <= access)
    {
	chanserv_notice(user, CSMSG_NO_BUMP_ACCESS);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[2]))) return 0;

    if((actee = GetTrueChannelAccess(channel->channel_info, handle)))
    {
        const char *title = accessNames[actee->access];
	chanserv_notice(user, CSMSG_USER_EXISTS, handle->handle, channel->name, preposition(title), title);
	return 0;
    }

    actee = add_channel_user(channel->channel_info, handle, access, 0, NULL);
    scan_handle_presence(channel, handle, NULL);

    chanserv_notice(user, CSMSG_ADDED_USER, accessNames[access], handle->handle, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_clvl)
{
    struct handle_info *handle;
    struct userData *victim;
    struct userData *actor;
    enum userLevel access, new_access = ulNone;
    int privileged = IsHelping(user) && ((user->handle_info->opserv_level >= chanserv_conf.nodelete_level) || !IsProtected(channel->channel_info));

    REQUIRE_PARAMS(3);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_CLVL);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[1]))) return 0;

    if(handle == user->handle_info && !privileged)
    {
	chanserv_notice(user, CSMSG_NO_SELF_CLVL);
	return 0;
    }

    if(!(victim = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if(actor->access <= victim->access && !privileged)
    {
	chanserv_notice(user, MSG_USER_OUTRANKED, handle->handle);
	return 0;
    }

    for(access = USER_LEVEL_LOWEST; access <= USER_LEVEL_HIGHEST; access++)
    {
	if(!irccasecmp(argv[2], accessNames[access]))
	{
	    new_access = access;
	    break;
	}
    }

    if(new_access == ulNone)
    {
	chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[2]);
	return 0;
    }

    if(new_access >= actor->access && !privileged)
    {
	chanserv_notice(user, CSMSG_NO_BUMP_ACCESS);
	return 0;
    }

    channel->channel_info->userCount[victim->access]--;
    userCount[victim->access]--;

    victim->access = new_access;

    channel->channel_info->userCount[victim->access]++;
    userCount[victim->access]++;

    chanserv_notice(user, CSMSG_CHANGED_ACCESS, handle->handle, preposition(accessNames[new_access]), accessNames[new_access], channel->name);

    return 1;
}

static CHANSERV_FUNC(cmd_deluser)
{
    struct handle_info *handle;
    struct userData *victim;
    struct userData *actor;
    enum userLevel access = ulNone;
    char *chan_name;

    REQUIRE_PARAMS(2);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_DEL);
	return 0;
    }

    if(!(handle = chanserv_get_handle_info(user, argv[argc-1]))) return 0;

    if(!(victim = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if(argc > 2)
    {
        access = user_level_from_name(argv[1]);
        if(access == ulNone)
        {
            chanserv_notice(user, CSMSG_INVALID_ACCESS, argv[1]);
            return 0;
        }
	if(access != victim->access)
	{
	    chanserv_notice(user, CSMSG_INCORRECT_ACCESS, handle->handle, preposition(accessNames[victim->access]), accessNames[victim->access], preposition(accessNames[access]), accessNames[access]);
	    return 0;
	}
    }
    else
    {
        access = victim->access;
    }

    if((actor->access <= victim->access) && !IsHelping(user))
    {
	chanserv_notice(user, MSG_USER_OUTRANKED, victim->handle->handle);
	return 0;
    }

    chan_name = strdup(channel->name);
    del_channel_user(victim, 1);
    chanserv_notice(user, CSMSG_DELETED_USER, preposition(accessNames[access]), accessNames[access], handle->handle, chan_name);
    free(chan_name);
    return 1;
}

static int
cmd_mdel_user(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned int argc, unsigned char *argv[], struct svccmd *cmd)
{
    struct userData *actor, *uData, *next;

    REQUIRE_PARAMS(2);

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(actor->access < ulMaster)
    {
	chanserv_notice(user, CSMSG_CANNOT_DEL);
	return 0;
    }

    if((actor->access <= access) && !IsHelping(user))
    {
	chanserv_notice(user, CSMSG_NO_ACCESS);
	return 0;
    }

    for(uData = channel->channel_info->users; uData; uData = next)
    {
	next = uData->next;

	if(uData->access == access && match_ircglob(uData->handle->handle, argv[1]))
	{
	    del_channel_user(uData, 1);
	}
    }

    chanserv_notice(user, CSMSG_DELETED_USERS, accessNames[access], argv[1], channel->name);

    return 1;
}

static CHANSERV_FUNC(cmd_mdelowner)
{
    return cmd_mdel_user(user, channel, ulOwner, argc, argv, cmd);
}

static CHANSERV_FUNC(cmd_mdelcoowner)
{
    return cmd_mdel_user(user, channel, ulCoowner, argc, argv, cmd);
}

static CHANSERV_FUNC(cmd_mdelmaster)
{
    return cmd_mdel_user(user, channel, ulMaster, argc, argv, cmd);
}

static CHANSERV_FUNC(cmd_mdelop)
{
    return cmd_mdel_user(user, channel, ulOp, argc, argv, cmd);
}

static CHANSERV_FUNC(cmd_mdelpeon)
{
    return cmd_mdel_user(user, channel, ulPeon, argc, argv, cmd);
}

static int
cmd_trim_bans(struct userNode *user, struct chanNode *channel, unsigned long duration)
{
    struct banData *bData, *next;
    char interval[INTERVALLEN];
    unsigned int count;
    time_t limit;

    count = 0;
    limit = now - duration;
    for(bData = channel->channel_info->bans; bData; bData = next)
    {
	next = bData->next;

        if((bData->triggered && bData->triggered >= limit) || (bData->set && bData->set >= limit))
        {
            continue;
        }

        del_channel_ban(bData);
        count++;
    }

    intervalString(interval, duration);
    chanserv_notice(user, CSMSG_TRIMMED_BANS, count, channel->name, interval);
    return 1;
}

static int
cmd_trim_users(struct userNode *user, struct chanNode *channel, enum userLevel access, unsigned long duration)
{
    struct userData *actor, *uData, *next;
    char interval[INTERVALLEN];
    unsigned int count;
    time_t limit;

    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if((actor->access <= access) && !IsHelping(user))
    {
	chanserv_notice(user, CSMSG_NO_ACCESS);
	return 0;
    }

    count = 0;
    limit = now - duration;
    for(uData = channel->channel_info->users; uData; uData = next)
    {
	next = uData->next;

	if((uData->seen > limit) || uData->present)
	{
	    continue;
	}

	if(uData->access == access || (access == ulNone && (uData->access < actor->access)))
	{
	    del_channel_user(uData, 1);
	    count++;
	}
    }

    intervalString(interval, duration);
    chanserv_notice(user, CSMSG_TRIMMED_USERS, count, (access == ulNone) ? "user" : accessNames[access], channel->name, interval);
    return 1;
}

static CHANSERV_FUNC(cmd_trim)
{
    unsigned long duration;
    enum userLevel level;

    REQUIRE_PARAMS(3);

    duration = ParseInterval(argv[2]);
    if(duration < 60)
    {
	chanserv_notice(user, CSMSG_CANNOT_TRIM);
	return 0;
    }

    if(!irccasecmp(argv[1], "bans"))
    {
	cmd_trim_bans(user, channel, duration);
	return 1;
    }

    if(!irccasecmp(argv[1], "users"))
    {
	cmd_trim_users(user, channel, ulNone, duration);
	return 1;
    }

    level = user_level_from_name(argv[1]);
    if(level != ulNone)
    {
	cmd_trim_users(user, channel, level, duration);
	return 1;
    }

    chanserv_notice(user, CSMSG_INVALID_TRIM, argv[1]);
    return 0;
}

/* If argc is 0 in cmd_up or cmd_down, no notices will be sent
   to the user. cmd_all takes advantage of this. */
static CHANSERV_FUNC(cmd_up)
{
    struct userData *uData;
    struct modeNode *mn;

    mn = GetUserMode(channel, user);
    if(!mn && argc)
    {
	chanserv_notice(user, MSG_CHANNEL_ABSENT, channel->name);
	return 0;
    }

    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    if(!uData)
    {
        if (argc) chanserv_notice(user, CSMSG_GODMODE_UP, argv[0]);
        return 0;
    }
    else if(uData->access >= ulOp)
    {
	if((mn->modes & MODE_CHANOP) && argc)
	{
	    chanserv_notice(user, CSMSG_ALREADY_UP, "opped", channel->name);
	    return 0;
	}

	AddChannelOp(1, &user, channel, chanserv, 1);
    }
    else
    {
	if((mn->modes & MODE_VOICE) && argc)
	{
	    chanserv_notice(user, CSMSG_ALREADY_UP, "voiced", channel->name);
	    return 0;
	}

	AddChannelVoice(1, &user, channel, chanserv, 1);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_down)
{
    char change[MAXLEN];
    struct modeNode *mn;

    mn = GetUserMode(channel, user);
    if(!mn && argc)
    {
	chanserv_notice(user, MSG_CHANNEL_ABSENT, channel->name);
	return 0;
    }

    if(!(mn->modes & (MODE_CHANOP | MODE_VOICE)) && argc)
    {
	chanserv_notice(user, CSMSG_ALREADY_DOWN, channel->name);
	return 0;
    }

    DelChannelOp(1, &user, channel, chanserv, 0);
    DelChannelVoice(1, &user, channel, chanserv, 0);

#ifdef WITH_PROTOCOL_P10
    sprintf(change, "-ov %s %s", user->numeric, user->numeric);
#else
    sprintf(change, "-ov %s %s", user->nick, user->nick);
#endif
    irc_mode(chanserv, channel, change);

    return 1;
}

static int cmd_all(struct userNode *user, UNUSED_ARG(struct chanNode *channel), UNUSED_ARG(unsigned int argc), UNUSED_ARG(unsigned char *argv[]), struct svccmd *cmd, modcmd_func_t mcmd)
{
    struct userData *cList;

    for(cList = user->handle_info->channels; cList; cList = cList->u_next)
    {
	if(IsSuspended(cList->channel)
           || IsUserSuspended(cList)
           || !GetUserMode(cList->channel->channel, user))
	{
	    continue;
	}

	mcmd(user, cList->channel->channel, 0, NULL, cmd);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_upall)
{
    return cmd_all(CSFUNC_ARGS, cmd_up);
}

static CHANSERV_FUNC(cmd_downall)
{
    return cmd_all(CSFUNC_ARGS, cmd_down);
}

typedef int validate_func_t(struct userNode *user, struct chanNode *channel, struct userNode *victim);
typedef void process_func_t(unsigned int num, struct userNode **newops, struct chanNode *channel, struct userNode *who, int announce);

static int
modify_users(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd), validate_func_t validate, process_func_t process, char *action)
{
    unsigned int i = 0, valid = 0, invalid = 0;
    struct userNode *victim, **victims;

    REQUIRE_PARAMS(2);

    victims = alloca(sizeof(struct userNode *) * (argc - 1));

    while(++i < argc)
    {
	victim = GetUserH(argv[i]);

	if(!victim
           || (validate && !validate(user, channel, victim))
           || !GetUserMode(channel, victim))
	{
	    invalid++;
	    continue;
	}

	victims[valid++] = victim;
    }

    if(invalid)
    {
	chanserv_notice(user, CSMSG_PROCESS_FAILED);

	if(invalid == (argc - 1))
	{
	    return 0;
	}
    }

    process(valid, victims, channel, chanserv, 1);
    if(valid && action)
    {
	chanserv_notice(user, CSMSG_PROCESS_USERS, action, channel->name);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_op)
{
    return modify_users(CSFUNC_ARGS, validate_op, AddChannelOp, "Opped");
}

static CHANSERV_FUNC(cmd_deop)
{
    return modify_users(CSFUNC_ARGS, validate_deop, DelChannelOp, "Deopped");
}

static CHANSERV_FUNC(cmd_voice)
{
    return modify_users(CSFUNC_ARGS, NULL, AddChannelVoice, "Voiced");
}

static CHANSERV_FUNC(cmd_devoice)
{
    return modify_users(CSFUNC_ARGS, NULL, DelChannelVoice, "Devoiced");
}

static int
bad_channel_ban(struct chanNode *channel, struct userNode *user, const char *ban, int *victimCount, struct userNode **victims)
{
    
    unsigned int ii;

    if(victimCount) *victimCount = 0;
    for(ii=0; ii<channel->members.used; ii++)
    {
        struct modeNode *mn = channel->members.list[ii];

        if(IsService(mn->user))
        {
            continue;
        }

        if(!user_matches_glob(mn->user, ban, 1))
        {
            continue;
        }

        if(protect_user(mn->user, user, channel->channel_info))
        {
            return 1;
        }

        if(victims)
        {
            victims[(*victimCount)++] = mn->user;
        }
    }
    return 0;
}

static int
eject_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd), int action)
{
    struct userNode *victim, **victims;
    unsigned int offset, n, victimCount = 0, duration = 0, present;
    char *reason = CSMSG_KICK_REASON, *ban, *name;

    offset = (action & ACTION_ADD_TIMED_BAN) ? 3 : 2;
    REQUIRE_PARAMS(offset);
    if(argc > offset)
    {
	reason = unsplit_string(argv + offset, argc - offset, NULL);
        if(strlen(reason) > (TOPICLEN - (NICKLEN + 3)))
        {
            /* Truncate the reason to a length of TOPICLEN, as
               the ircd does; however, leave room for an ellipsis
               and the kicker's nick. */
            sprintf(reason + (TOPICLEN - (NICKLEN + 6)), "...");
        }
    }

    if((victim = GetUserH(argv[1])))
    {
        /* XXX: The comparison with ACTION_KICK is just because all
         * other actions can work on users outside the channel, and we
         * want to allow those (e.g.  unbans) in that case.  If we add
         * some other ejection action for in-channel users, change
         * this too. */
	present = GetUserMode(channel, victim) ? 1 : 0;

        if((action == ACTION_KICK) && !present)
        {
            chanserv_notice(user, MSG_CHANNEL_USER_ABSENT, victim->nick, channel->name);
            return 0;
        }

	if(IsService(victim))
	{
	    chanserv_notice(user, MSG_SERVICE_IMMUNE, victim->nick);
	    return 0;
	}

	if(protect_user(victim, user, channel->channel_info))
	{
	    chanserv_notice(user, CSMSG_USER_PROTECTED, victim->nick);
	    return 0;
	}

	victims = &victim;
	victimCount = 1;

	ban = generate_hostmask(victim, GENMASK_STRICT_HOST|GENMASK_ANY_IDENT);
	name = victim->nick;
    }
    else
    {
	if(!is_ircmask(argv[1]))
	{
	    chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	    return 0;
	}

	victims = alloca(sizeof(struct userNode *) * channel->members.used);

        if(bad_channel_ban(channel, user, argv[1], &victimCount, victims))
        {
	    chanserv_notice(user, CSMSG_MASK_PROTECTED, argv[1]);
	    return 0;
	}

        if((victimCount > 4) && ((victimCount * 3) > channel->members.used) && !IsOper(user))
        {
            chanserv_notice(user, CSMSG_LAME_MASK, argv[1]);
            return 0;
        }

        if((action == ACTION_KICK) && (victimCount == 0))
        {
            chanserv_notice(user, CSMSG_NO_MATCHING_USERS, channel->name, argv[1]);
            return 0;
        }

	ban = strdup(argv[1]);
	name = ban;
	present = victimCount ? 1 : 0;
    }

    /* Truncate the ban in place if necessary; we must ensure
       that 'ban' is a valid ban mask before sanitizing it. */
    sanitize_ircmask(ban);

    if(action & ACTION_ADD_BAN)
    {
	struct banData *bData, *next;

	if(channel->channel_info->banCount >= chanserv_conf.max_chan_bans)
	{
	    chanserv_notice(user, CSMSG_STRUCTURE_LIMIT, "ban", chanserv_conf.max_chan_bans);
	    free(ban);
	    return 0;
	}

	if(action & ACTION_ADD_TIMED_BAN)
	{
	    duration = ParseInterval(argv[2]);

	    if(duration < 15)
	    {
		chanserv_notice(user, CSMSG_INVALID_DURATION, "at least", "15 seconds");

		free(ban);
		return 0;
	    }
	    else if(duration > (86400 * 365 * 2))
	    {
		chanserv_notice(user, CSMSG_INVALID_DURATION, "less than", "2 years");

		free(ban);
		return 0;
	    }
	}
	else
	{
	    duration = 0;
	}

	for(bData = channel->channel_info->bans; bData; bData = next)
	{
	    if(match_ircglobs(bData->mask, ban))
	    {
		int exact = !irccasecmp(bData->mask, ban);

		/* The ban is redundant; there is already a ban
		   with the same effect in place. */
		if(exact)
		{
		    if(bData->reason) free(bData->reason);
		    bData->reason = strdup(reason);
                    safestrncpy(bData->owner, (user->handle_info ? user->handle_info->handle : user->nick), sizeof(bData->owner));
		    chanserv_notice(user, CSMSG_REASON_CHANGE, ban);
		    if(!bData->expires) goto post_add_ban;
		}
		if(exact && bData->expires)
		{
		    int reset = 0;

		    /* If the ban matches an existing one exactly,
		       extend the expiration time if the provided
		       duration is longer. */
		    if(duration && ((time_t)(now + duration) > bData->expires))
		    {
			bData->expires = now + duration;
			reset = 1;
		    }
		    else if(!duration)
		    {
			bData->expires = 0;
			reset = 1;
		    }

		    if(reset)
		    {
			char length[INTERVALLEN];

			/* Delete the expiration timeq entry and
			   requeue if necessary. */
			timeq_del(0, expire_ban, bData, TIMEQ_IGNORE_WHEN);

			if(bData->expires)
			{
			    timeq_add(bData->expires, expire_ban, bData);
			}

			if(duration)
			{
			    intervalString(length, duration);
			    chanserv_notice(user, CSMSG_BAN_EXTENDED, ban, length);
			}
			else
			{
			    chanserv_notice(user, CSMSG_BAN_ADDED, name, channel->name);
			}

			goto post_add_ban;
		    }
		}
		chanserv_notice(user, CSMSG_REDUNDANT_BAN, name, channel->name);

		free(ban);
		return 0;
	    }

	    next = bData->next;
	    if(match_ircglobs(ban, bData->mask))
	    {
		/* The ban we are adding makes previously existing
		   bans redundant; silently remove them. */
		del_channel_ban(bData);
	    }
	}

	add_channel_ban(channel->channel_info, ban, (user->handle_info ? user->handle_info->handle : user->nick), now, (present ? now : 0), (duration ? now + duration : 0), reason);
    }

  post_add_ban:
    if(action & ACTION_BAN)
    {
	unsigned int exists;

	if(channel->banlist.used >= MAXBANS)
	{
	    chanserv_notice(user, CSMSG_BANLIST_FULL, channel->name);

	    free(ban);
	    return 0;
	}

	/* De-op/voice the users if we know they're in the channel. */
	if(present)
	{
	    DelChannelOp(victimCount, victims, channel, chanserv, 1);
	    DelChannelVoice(victimCount, victims, channel, chanserv, 1);
	}

	exists = ChannelBanExists(channel, ban);
	if(exists)
	{
	    if(action == ACTION_BAN)
	    {
		chanserv_notice(user, CSMSG_REDUNDANT_BAN, name, channel->name);

		free(ban);
		return 0;
	    }
	}
	else
	{
	    AddChannelBan(1, (const char **)&ban, channel, chanserv, now, 1);
	}
    }

    if(action & ACTION_KICK)
    {
        char kick_reason[MAXLEN];
	sprintf(kick_reason, CSMSG_KICK_FORMAT, reason, user->nick);

	for (n = 0; n < victimCount; n++)
	{
	    KickChannelUser(victims[n], channel, chanserv, kick_reason);
	}
    }

    if(action & ACTION_ADD_BAN)
    {
	if(duration)
	{
	    char interval[INTERVALLEN];

	    intervalString(interval, duration);
	    chanserv_notice(user, CSMSG_TIMED_BAN_ADDED, name, channel->name, interval);
	}
	else
	{
	    chanserv_notice(user, CSMSG_BAN_ADDED, name, channel->name);
	}
    }
    else if((action & (ACTION_BAN | ACTION_KICK)) == (ACTION_BAN | ACTION_KICK))
    {
	chanserv_notice(user, CSMSG_KICK_BAN_DONE, name, channel->name);
    }
    else if(action & ACTION_BAN)
    {
	chanserv_notice(user, CSMSG_BAN_DONE, name, channel->name);
    }
    else if(action & ACTION_KICK && victimCount)
    {
	chanserv_notice(user, CSMSG_KICK_DONE, name, channel->name);
    }

    free(ban);
    return 1;
}

static CHANSERV_FUNC(cmd_kickban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN);
}

static CHANSERV_FUNC(cmd_kick)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK);
}

static CHANSERV_FUNC(cmd_ban)
{
    return eject_user(CSFUNC_ARGS, ACTION_BAN);
}

static CHANSERV_FUNC(cmd_addban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN | ACTION_ADD_BAN);
}

static CHANSERV_FUNC(cmd_addtimedban)
{
    return eject_user(CSFUNC_ARGS, ACTION_KICK | ACTION_BAN | ACTION_ADD_BAN | ACTION_ADD_TIMED_BAN);
}

static int
unban_user(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd), int action)
{
    struct userNode *actee;
    char *mask = NULL;
    int acted = 0;

    REQUIRE_PARAMS(2);

    /* may want to allow a comma delimited list of users... */
    if(!(actee = GetUserH(argv[1])))
    {
	if(!is_ircmask(argv[1]))
	{
	    chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	    return 0;
	}

	mask = strdup(argv[1]);
    }

    /* We don't sanitize the mask here because ircu
       doesn't do it. */
    if(action & ACTION_UNBAN)
    {
	unsigned int i;
	struct banNode **ary = channel->banlist.list;
	struct string_list *removed;
	char *ban;

	removed = alloc_string_list(2);
	for (i=0; i<channel->banlist.used; i++)
	{
	    if(actee)
	    {
		if(user_matches_glob(actee, ary[i]->ban, 1))
		{
		    ban = strdup(ary[i]->ban);
		    string_list_append(removed, ban);
		}
	    }
	    else
	    {
		if(match_ircglobs(mask, ary[i]->ban))
		{
		    ban = strdup(ary[i]->ban);
		    string_list_append(removed, ban);
		}
	    }
	}

	if(removed->used)
	{
	    DelChannelBan(removed->used, (const char **)removed->list, channel, chanserv, 1);
	    acted = 1;
	}

	free_string_list(removed);
    }

    if(action & ACTION_DEL_BAN)
    {
	struct banData *ban, *next;

	ban = channel->channel_info->bans;
	while(ban)
	{
	    if(actee)
	    {
		for( ; ban && !user_matches_glob(actee, ban->mask, 1);
		     ban = ban->next);
	    }
	    else
	    {
		for( ; ban && !match_ircglobs(mask, ban->mask);
		     ban = ban->next);
	    }

	    if(!ban) break;

	    next = ban->next;
	    del_channel_ban(ban);
	    ban = next;
	    acted = 1;
	}
    }

    if(!acted)
    {
	chanserv_notice(user, CSMSG_BAN_NOT_FOUND, actee ? actee->nick : mask);
    }
    else
    {
	chanserv_notice(user, CSMSG_BAN_REMOVED, actee ? actee->nick : mask);
    }

    if(mask) free(mask);
    return 1;
}

static CHANSERV_FUNC(cmd_unban)
{
    return unban_user(CSFUNC_ARGS, ACTION_UNBAN);
}

static CHANSERV_FUNC(cmd_delban)
{
    /* it doesn't necessarily have to remove the channel ban - may want
       to make that an option. */
    return unban_user(CSFUNC_ARGS, ACTION_UNBAN | ACTION_DEL_BAN);
}

static CHANSERV_FUNC(cmd_unbanme)
{
    struct userData *uData = GetChannelUser(channel->channel_info, user->handle_info);
    long flags = ACTION_UNBAN;

    /* remove permanent bans if the user has the proper access. */
    if(uData->access >= ulMaster)
    {
	flags |= ACTION_DEL_BAN;
    }

    argv[1] = user->nick;
    return unban_user(user, channel, 2, argv, cmd, flags);
}

static CHANSERV_FUNC(cmd_unbanall)
{
    unsigned int i;
    struct banNode **ary = channel->banlist.list;
    struct string_list *removed;
    char *ban;

    if(!channel->banlist.used)
    {
	chanserv_notice(user, CSMSG_NO_BANS, channel->name);
	return 0;
    }

    removed = alloc_string_list(channel->banlist.used);
    for (i=0; i<channel->banlist.used; i++)
    {
	ban = strdup(ary[i]->ban);
	string_list_append(removed, ban);
    }
    DelChannelBan(removed->used, (const char **)removed->list, channel, chanserv, 1);
    free_string_list(removed);

    chanserv_notice(user, CSMSG_BANS_REMOVED, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_open)
{
    char change[MAXLEN];

    cmd_unbanme(CSFUNC_ARGS);

    if(channel->modes & MODE_KEY)
    {
	sprintf(change, "-ilk %s", channel->key);
    }
    else
    {
        strcpy(change, "-il");
    }

    mod_chanmode(channel, chanserv, change);
    irc_mode(chanserv, channel, change);

    chanserv_notice(user, CSMSG_CHANNEL_OPENED, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_access)
{
    struct userNode *target;
    struct handle_info *target_handle;
    struct userData *uData;
    int helping;
    char prefix[MAXLEN];

    if(!channel)
    {
        struct userData *uData;
        const char *chanName;
        int hide = 0;

        target_handle = user->handle_info;
        if(!target_handle)
        {
            chanserv_notice(user, MSG_AUTHENTICATE);
            return 0;
        }
        if(argc > 1)
        {
            if(!IsHelping(user))
            {
                chanserv_notice(user, CSMSG_ACCESS_SELF_ONLY, argv[0]);
                return 0;
            }

            if (!(target_handle = chanserv_get_handle_info(user, argv[1]))) return 0;
            hide = 1;
        }
        if(!target_handle->channels)
        {
            reply(CSMSG_SQUAT_ACCESS);
            return 1;
        }
        chanserv_notice(user, CSMSG_INFOLINE_LIST, target_handle->handle);
        for (uData = target_handle->channels; uData; uData = uData->u_next) {
            struct chanData *cData = uData->channel;

            if(uData->access == ulHelper) continue;
            if(IsProtected(cData) && hide && !GetTrueChannelAccess(cData, user->handle_info))
            {
                continue;
            }
            chanName = cData->channel->name;
            if(uData->info)
            {
                chanserv_notice(user, "[%s (%s)] %s", chanName, accessNames[uData->access], uData->info);
            } else {
                chanserv_notice(user, "[%s (%s)]", chanName, accessNames[uData->access]);
            }
        }
        return 1;
    }

    if(argc < 2)
    {
	target = user;
        target_handle = target->handle_info;
    }
    else if((target = GetUserH(argv[1])))
    {
        target_handle = target->handle_info;
    }
    else if(argv[1][0] == '*')
    {
        if(!(target_handle = get_handle_info(argv[1]+1)))
        {
            chanserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]+1);
            return 0;
        }
    }
    else
    {
        chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
        return 0;
    }

    assert(target || target_handle);

    if(target == chanserv)
    {
	chanserv_notice(user, CSMSG_IS_CHANSERV);
	return 1;
    }

    if(!target_handle)
    {
        if(IsOper(target))
        {
            chanserv_notice(user, CSMSG_LAZY_SMURF_TARGET, target->nick, chanserv_conf.irc_operator_epithet);
            return 0;
        }
	if(target != user)
	{
	    chanserv_notice(user, MSG_USER_AUTHENTICATE, target->nick);
	    return 0;
	}
        chanserv_notice(user, MSG_AUTHENTICATE);
        return 0;
    }

    if(target)
    {
        const char *epithet = NULL, *type = NULL;
        if(IsOper(target))
        {
            epithet = chanserv_conf.irc_operator_epithet;
            type = "IRCOp";
        }
        else if(IsNetworkHelper(target))
        {
            epithet = chanserv_conf.network_helper_epithet;
            type = "network helper";
        }
        else if(IsSupportHelper(target))
        {
            epithet = chanserv_conf.support_helper_epithet;
            type = "support helper";
        }
        if(epithet)
        {
            if (target_handle->epithet) {
                if (target_handle->epithet) epithet = target_handle->epithet;
                chanserv_notice(user, CSMSG_SMURF_TARGET, target->nick, epithet, type);
            } else {
                chanserv_notice(user, CSMSG_LAME_SMURF_TARGET, target->nick, preposition(type), type);
            }
        }
        sprintf(prefix, "%s (%s)", target->nick, target_handle->handle);
    }
    else
    {
        sprintf(prefix, "%s", target_handle->handle);
    }

    if(!channel->channel_info)
    {
        chanserv_notice(user, CSMSG_NOT_REGISTERED, channel->name);
        return 1;
    }

    helping = HANDLE_FLAGGED(target_handle, HELPING)
        && ((target_handle->opserv_level >= chanserv_conf.nodelete_level) || !IsProtected(channel->channel_info));
    if((uData = GetTrueChannelAccess(channel->channel_info, target_handle)))
    {
        chanserv_notice(user, (helping ? CSMSG_HELPER_HAS_ACCESS : CSMSG_USER_HAS_ACCESS), prefix, preposition(accessNames[uData->access]), accessNames[uData->access], channel->name);
        /* To prevent possible information leaks, only show infolines
         * if the requestor is in the channel or it's their own
         * handle. */
        if(uData->info && (GetUserMode(channel, user) || (target_handle == user->handle_info)))
        {
            chanserv_notice(user, "[%s] %s", (target ? target->nick : target_handle->handle), uData->info);
        }
        /* Likewise, only say it's suspended if the user has active
         * access in that channel or it's their own entry. */
        if(IsUserSuspended(uData)
           && (GetChannelUser(channel->channel_info, user->handle_info)
               || (user->handle_info == uData->handle)))
        {
            chanserv_notice(user, CSMSG_USER_SUSPENDED, (target ? target->nick : target_handle->handle), channel->name);
        }
    }
    else
    {
        chanserv_notice(user, (helping ? CSMSG_HELPER_NO_ACCESS : CSMSG_USER_NO_ACCESS), prefix, channel->name);
    }

    return 1;
}

static void
zoot_list(struct listData *list)
{
    struct userData *uData;
    unsigned int start, curr;
    struct helpfile_table tmp_table;
    const char **temp;

    if(list->table.length == 1)
    {
        if(list->search)
            chanserv_notice(list->user, CSMSG_ACCESS_SEARCH_HDR, list->channel->name, list->highest == list->lowest ? accessNames[list->lowest] : "user", list->search);
        else
            chanserv_notice(list->user, CSMSG_ACCESS_HEADER, list->channel->name, list->highest == list->lowest ? accessNames[list->lowest] : "user");
        chanserv_notice(list->user, "  None");
    }
    tmp_table.width = list->table.width;
    tmp_table.flags = list->table.flags;
    list->table.contents[0][0] = " ";
    for(start = curr = 1; curr < list->table.length; )
    {
        uData = list->users[curr-1];
        list->table.contents[curr++][0] = " ";
        if((curr == list->table.length) || (list->users[curr-1]->access != uData->access))
        {
            if(list->search)
                chanserv_notice(list->user, CSMSG_ACCESS_SEARCH_HDR, list->channel->name, accessNames[uData->access], list->search);
            else
                chanserv_notice(list->user, CSMSG_ACCESS_HEADER, list->channel->name, accessNames[uData->access]);
            temp = list->table.contents[--start];
            list->table.contents[start] = list->table.contents[0];
            tmp_table.contents = list->table.contents + start;
            tmp_table.length = curr - start;
            table_send(chanserv, list->user->nick, 0, NULL, tmp_table);
            list->table.contents[start] = temp;
            start = curr;
        }
    }
}

static void
def_list(struct listData *list)
{
    if(list->search)
        chanserv_notice(list->user, CSMSG_ACCESS_SEARCH_HDR, list->channel->name, list->highest == list->lowest ? accessNames[list->lowest] : "user", list->search);
    else
        chanserv_notice(list->user, CSMSG_ACCESS_HEADER, list->channel->name, list->highest == list->lowest ? accessNames[list->lowest] : "user");
    table_send(chanserv, list->user->nick, 0, NULL, list->table);
    if(list->table.length == 1)
        chanserv_notice(list->user, "  None");
}

static int
userData_access_comp(const void *arg_a, const void *arg_b)
{
    const struct userData *a = *(struct userData**)arg_a;
    const struct userData *b = *(struct userData**)arg_b;
    int res;
    if(a->access != b->access)
        res = b->access - a->access;
    else
        res = irccasecmp(a->handle->handle, b->handle->handle);
    return res;
}

static int
cmd_list_users(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd), enum userLevel display_lowest, enum userLevel display_highest)
{
    void (*send_list)(struct listData *);
    struct userData *uData;
    struct listData lData;
    enum userLevel display;
    unsigned int matches;
    const char **ary;

    lData.user = user;
    lData.channel = channel;
    lData.lowest = display_lowest;
    lData.highest = display_highest;
    lData.search = (argc > 1) ? argv[1] : NULL;
    send_list = zoot_list;

    if(user->handle_info)
    {
	switch(user->handle_info->userlist_style)
	{
	case HI_STYLE_DEF: send_list = def_list; break;
        case HI_STYLE_ZOOT: send_list = zoot_list; break;
	}
    }

    matches = 0;
    for(display = display_highest; display >= display_lowest; display--)
	matches += channel->channel_info->userCount[display];

    lData.users = alloca(matches * sizeof(struct userData *));
    matches = 0;
    for(uData = channel->channel_info->users; uData; uData = uData->next)
    {
	if((uData->access < display_lowest)
           || (uData->access > display_highest)
           || (lData.search && !match_ircglob(uData->handle->handle, lData.search)))
	    continue;
	lData.users[matches++] = uData;
    }
    qsort(lData.users, matches, sizeof(lData.users[0]), userData_access_comp);

    lData.table.length = matches+1;
    lData.table.width = 4;
    lData.table.flags = TABLE_NO_FREE;
    lData.table.contents = malloc(lData.table.length*sizeof(*lData.table.contents));
    ary = malloc(lData.table.width*sizeof(**lData.table.contents));
    lData.table.contents[0] = ary;
    ary[0] = "Access";
    ary[1] = "Account";
    ary[2] = "Last Seen";
    ary[3] = "Status";
    for(matches = 1; matches < lData.table.length; ++matches)
    {
        struct userData *uData = lData.users[matches-1];
        char seen[INTERVALLEN];

        ary = malloc(lData.table.width*sizeof(**lData.table.contents));
        lData.table.contents[matches] = ary;
        ary[0] = accessNames[uData->access];
        ary[1] = uData->handle->handle;
        if(uData->present)
            ary[2] = "Here";
        else if(!uData->seen)
            ary[2] = "Never";
        else
            ary[2] = intervalString(seen, now - uData->seen);
        ary[2] = strdup(ary[2]);
        if(IsUserSuspended(uData))
            ary[3] = "Suspended";
        else if(HANDLE_FLAGGED(uData->handle, FROZEN))
            ary[3] = "Vacation";
        else
            ary[3] = "Normal";
    }
    send_list(&lData);
    for(matches = 1; matches < lData.table.length; ++matches)
    {
        free((char*)lData.table.contents[matches][2]);
        free(lData.table.contents[matches]);
    }
    free(lData.table.contents[0]);
    free(lData.table.contents);
    return 1;
}

static CHANSERV_FUNC(cmd_users)
{
    return cmd_list_users(CSFUNC_ARGS, USER_LEVEL_LOWEST, USER_LEVEL_HIGHEST);
}

static CHANSERV_FUNC(cmd_wlist)
{
    return cmd_list_users(CSFUNC_ARGS, ulOwner, ulOwner);
}

static CHANSERV_FUNC(cmd_clist)
{
    return cmd_list_users(CSFUNC_ARGS, ulCoowner, ulCoowner);
}
static CHANSERV_FUNC(cmd_mlist)
{
    return cmd_list_users(CSFUNC_ARGS, ulMaster, ulMaster);
}
static CHANSERV_FUNC(cmd_olist)
{
    return cmd_list_users(CSFUNC_ARGS, ulOp, ulOp);
}
static CHANSERV_FUNC(cmd_plist)
{
    return cmd_list_users(CSFUNC_ARGS, ulPeon, ulPeon);
}

static CHANSERV_FUNC(cmd_bans)
{
    unsigned int banLength = 0, ownerLength = 0, durationLength = 0;
    unsigned int matches = 0, timed = 0, i;
    char t_buffer[INTERVALLEN], *triggered = NULL;
    char e_buffer[INTERVALLEN], *expires = NULL, *search = NULL;
    struct banData *ban, **bans;

    if(argc > 1)
    {
	search = argv[1];
    }

    if(!channel->channel_info->bans)
    {
	chanserv_notice(user, CSMSG_BAN_HEADER, 6, "Mask:", 8, "Set By:", 11, "Triggered:", 9, "Expires:", "Reason:");
	chanserv_notice(user, "$bNone.$b");
	return 0;
    }

    bans = alloca(channel->channel_info->banCount * sizeof(struct banData *));

    for(ban = channel->channel_info->bans; ban; ban = ban->next)
    {
	if(search && !match_ircglobs(search, ban->mask))
	    continue;

	bans[matches++] = ban;

	if((unsigned int)strlen(ban->mask) > banLength) banLength = strlen(ban->mask);
	if((unsigned int)strlen(ban->owner) > ownerLength) ownerLength = strlen(ban->owner);

	if(ban->expires) timed = 1;
    }

    if(!matches)
    {
	chanserv_notice(user, CSMSG_BAN_HEADER, 6, "Mask:", 8, "Set By:", 11, "Triggered:", 9, "Expires:", "Reason:");
	chanserv_notice(user, "$bNone.$b");
	return 0;
    }

    banLength = banLength < 6 ? 6 : banLength + 1;
    ownerLength = ownerLength < 8 ? 8 : ownerLength + 1;
    durationLength = timed ? INTERVALLEN : 0;

    chanserv_notice(user, CSMSG_BAN_HEADER, banLength, "Mask:", ownerLength, "Set By:", INTERVALLEN, "Triggered:", durationLength, timed ? "Expires:" : "", "Reason:");
    for(i = 0; i < matches; i++)
    {
	ban = bans[i];

	if(!timed)
	{
	    expires = "";
	}
	else if(ban->expires)
	{
	    intervalString(e_buffer, ban->expires - now);
	    expires = e_buffer;
	}
	else
	{
	    expires = "Never.";
	}

	if(ban->triggered)
	{
	    intervalString(t_buffer, now - ban->triggered);
	    triggered = t_buffer;
	}
	else
	{
	    triggered = "Never.";
	}

	chanserv_notice(user, CSMSG_BAN_DATA, banLength, ban->mask, ownerLength, ban->owner, INTERVALLEN, triggered, durationLength, expires, ban->reason);
    }

    return 1;
}

static int
bad_topic(struct chanNode *channel, struct userNode *user, const char *new_topic)
{
    struct chanData *cData = channel->channel_info;
    if (check_user_level(user, channel, cData->options[optEnfTopic], 1, 0)) return 0;
    if (cData->topic_mask)
    {
        return !match_ircglob(new_topic, cData->topic_mask);
    }
    else if (cData->topic[0])
    {
        return irccasecmp(new_topic, cData->topic);
    }
    else return 0;
}

static CHANSERV_FUNC(cmd_topic)
{
    struct chanData *cData;
    char *topic;

    cData = channel->channel_info;
    if(argc < 2)
    {
	if(cData->topic[0])
	{
	    SetChannelTopic(channel, chanserv, cData->topic, 1);
	    chanserv_notice(user, CSMSG_TOPIC_SET, cData->topic);
            return 1;
	}

        chanserv_notice(user, CSMSG_NO_TOPIC, channel->name);
        return 0;
    }

    topic = unsplit_string(argv + 1, argc - 1, NULL);
    /* If they say "!topic *", use an empty topic. */
    if ((topic[0] == '*') && (topic[1] == 0)) {
        topic[0] = 0;
    }
    if(bad_topic(channel, user, topic))
    {
        char *topic_mask = cData->topic_mask;
        if(topic_mask)
        {
            char new_topic[TOPICLEN+1], tchar;
            int pos=0, starpos=-1, dpos=0, len;

            while((tchar = topic_mask[pos++]) && (dpos <= TOPICLEN))
            {
                switch(tchar)
                {
                case '*':
                    if (starpos != -1) goto bad_mask;
                    len = strlen(topic);
                    if ((dpos + len) > TOPICLEN) len = TOPICLEN + 1 - dpos;
                    memcpy(new_topic+dpos, topic, len);
                    dpos += len;
                    starpos = pos;
                    break;
                case '\\': tchar = topic_mask[pos++]; /* and fall through */
                default: new_topic[dpos++] = tchar; break;
                }
            }
            if ((dpos > TOPICLEN) || tchar) {
            bad_mask:
                chanserv_notice(user, CSMSG_TOPICMASK_CONFLICT1, channel->name, topic_mask);
                chanserv_notice(user, CSMSG_TOPICMASK_CONFLICT2, TOPICLEN);
                return 0;
            }
            new_topic[dpos] = 0;
            SetChannelTopic(channel, chanserv, new_topic, 1);
        } else {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }
    }
    else
    {
        SetChannelTopic(channel, chanserv, topic, 1);
    }

    if(cData->flags & CHANNEL_TOPIC_SNARF)
    {
        /* Grab the topic and save it as the default topic. */
        safestrncpy(cData->topic, channel->topic, sizeof(cData->topic));
    }

    return 1;
}

static CHANSERV_FUNC(cmd_mode)
{
    char modes[MODELEN], *changes;

    if(argc < 2)
    {
	if(channel->channel_info->mode_on || channel->channel_info->mode_off)
	{
	    create_mode_lock(modes, channel->channel_info, 0);
	    mod_chanmode(channel, chanserv, modes);
	    irc_mode(chanserv, channel, modes);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NO_MODES, channel->name);
	}

	return 1;
    }

    changes = unsplit_string(argv + 1, argc - 1, NULL);
    if(!verify_mod_chanmode(channel, chanserv, changes, 0))
    {
	chanserv_notice(user, MSG_INVALID_MODES, changes);
	return 0;
    }

    if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
    {
        if(mode_lock_violated(channel))
	{
            /* verify_mod_chanmode() modified srvx's channel mode data
               but didn't send it out, so just undo it and don't worry
               about sending the corrections out. */
	    create_mode_lock(modes, channel->channel_info, 0);
	    mod_chanmode(channel, chanserv, modes);

	    create_mode_lock(modes, channel->channel_info, 1);
	    chanserv_notice(user, CSMSG_MODE_LOCKED, modes, channel->name);
	    return 0;
	}
    }

    irc_mode(chanserv, channel, changes);
    chanserv_notice(user, CSMSG_MODES_SET, changes);
    return 1;
}

static CHANSERV_FUNC(cmd_invite)
{
    struct userData *uData;
    struct userNode *invite;

    uData = GetChannelUser(channel->channel_info, user->handle_info);

    if(argc > 1)
    {
        if(!(invite = GetUserH(argv[1])))
	{
            chanserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
            return 0;
        }
    }
    else
    {
        invite = user;
    }

    if(GetUserMode(channel, invite))
    {
	chanserv_notice(user, CSMSG_ALREADY_PRESENT, invite->nick, channel->name);
	return 0;
    }

    if(user != invite)
    {
	char *reason = (argc > 2) ? unsplit_string(argv + 2, argc - 2, NULL) : "";

	chanserv_notice(invite, CSMSG_INVITING_YOU, user->nick, channel->name, (argc > 2) ? ": " : ".", reason);
    }
    irc_invite(chanserv, invite, channel);
    if(argc > 1)
    {
	chanserv_notice(user, CSMSG_INVITED_USER, argv[1], channel->name);
    }

    return 1;
}

static CHANSERV_FUNC(cmd_inviteme)
{
    struct userData *uData;

    if(GetUserMode(channel, user))
    {
	reply(CSMSG_YOU_ALREADY_PRESENT, channel->name);
	return 0;
    }
    if(channel->channel_info
       && !(channel->channel_info->flags & CHANNEL_PEON_INVITE)
       && (uData = GetChannelUser(channel->channel_info, user->handle_info))
       && (uData->access == ulPeon))
    {
        reply(CSMSG_LOW_CHANNEL_ACCESS, channel->name);
        return 0;
    }
    irc_invite(cmd->parent->bot, user, channel);
    return 1;
}

static CHANSERV_FUNC(cmd_version)
{
    chanserv_notice(user, MSG_VERSION);

    if(argc > 1 && !irccasecmp(argv[1], "cvs"))
    {
	chanserv_notice(user, CSMSG_VERSION_ID);
    }

    return 1;
}

static void
show_suspension_info(struct userNode *user, struct suspended *suspended)
{
    unsigned int combo;
    char buf1[INTERVALLEN], buf2[INTERVALLEN];

    /* We display things based on two dimensions:
     * - Issue time: present or absent
     * - Expiration: revoked, expired, expires in future, or indefinite expiration
     * (in order of precedence, so something both expired and revoked
     * only counts as revoked)
     */
    combo = (suspended->issued ? 4 : 0)
        + (suspended->revoked ? 3 : suspended->expires ? ((suspended->expires < now) ? 2 : 1) : 0);
    switch(combo) {
    case 0: /* no issue time, indefinite expiration */
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_0, suspended->suspender, suspended->reason);
        break;
    case 1: /* no issue time, expires in future */
        intervalString(buf1, suspended->expires-now);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_1, suspended->suspender, buf1, suspended->reason);
        break;
    case 2: /* no issue time, expired */
        intervalString(buf1, now-suspended->expires);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_2, suspended->suspender, buf1, suspended->reason);
        break;
    case 3: /* no issue time, revoked */
        intervalString(buf1, now-suspended->revoked);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_3, suspended->suspender, buf1, suspended->reason);
        break;
    case 4: /* issue time set, indefinite expiration */
        intervalString(buf1, now-suspended->issued);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_4, buf1, suspended->suspender, suspended->reason);
        break;
    case 5: /* issue time set, expires in future */
        intervalString(buf1, now-suspended->issued);
        intervalString(buf2, suspended->expires-now);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_5, buf1, suspended->suspender, buf2, suspended->reason);
        break;
    case 6: /* issue time set, expired */
        intervalString(buf1, now-suspended->issued);
        intervalString(buf2, now-suspended->expires);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_6, buf1, suspended->suspender, buf2, suspended->reason);
        break;
    case 7: /* issue time set, revoked */
        intervalString(buf1, now-suspended->issued);
        intervalString(buf2, now-suspended->revoked);
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED_7, buf1, suspended->suspender, buf2, suspended->reason);
        break;
    default:
        log_module(CS_LOG, LOG_ERROR, "Invalid combo value %d in show_suspension_info()", combo);
        return;
    }
}

static CHANSERV_FUNC(cmd_info)
{
    char modes[MODELEN], buffer[INTERVALLEN];
    struct userData *uData;
    struct chanData *cData;
    struct do_not_register *dnr;
    enum userLevel access;
    struct note *note;
    dict_iterator_t it;
    int privileged;

    cData = channel->channel_info;
    chanserv_notice(user, CSMSG_CHANNEL_INFO, channel->name);

    uData = GetChannelUser(cData, user->handle_info);
    if(uData && uData->access >= ulOp)
    {
        create_mode_lock(modes, cData, 1);
	chanserv_notice(user, CSMSG_CHANNEL_TOPIC, cData->topic);
	chanserv_notice(user, CSMSG_CHANNEL_MODES, modes[0] ? modes : "None.");
    }

    for(it = dict_first(cData->notes); it; it = iter_next(it))
    {
        int padding;

        note = iter_data(it);
        if(!note_type_visible_to_user(cData, note->type, user))
        {
            continue;
        }

        padding = PADLEN - 1 - strlen(iter_key(it));
        chanserv_notice(user, CSMSG_CHANNEL_NOTE, iter_key(it), padding > 0 ? padding : 1, "", note->note);
    }

    chanserv_notice(user, CSMSG_CHANNEL_MAX, cData->max);
    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	int ii = PADLEN - 7 - strlen(accessTitles[access]);
 
	chanserv_notice(user, CSMSG_CHANNEL_USERS, accessTitles[access], ii, "", cData->userCount[access]);
    }
    chanserv_notice(user, CSMSG_CHANNEL_USERS, "Total User", PADLEN - 17, "", cData->userCount[0]);
    chanserv_notice(user, CSMSG_CHANNEL_BANS, cData->banCount);

    intervalString(buffer, now - cData->visited);
    chanserv_notice(user, CSMSG_CHANNEL_VISITED, buffer);

    intervalString(buffer, now - cData->registered);
    chanserv_notice(user, CSMSG_CHANNEL_REGISTERED, buffer);

    privileged = IsStaff(user);
    if((uData && uData->access >= ulCoowner) || privileged)
    {
        if(cData->registrar)
        {
            chanserv_notice(user, CSMSG_CHANNEL_REGISTRAR, cData->registrar);
        }
    }

    if(privileged)
    {
	if((dnr = chanserv_is_dnr(channel->name, NULL)))
	{
	    chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
	}
    }

    if((cData->suspended) && (((uData && (uData->access >= ulCoowner)) || IsHelping(user))))
    {
        struct suspended *suspended;
        chanserv_notice(user, (IsSuspended(cData) ? CSMSG_CHANNEL_SUSPENDED : CSMSG_CHANNEL_HISTORY), channel->name);
        for(suspended = cData->suspended; suspended; suspended = suspended->previous)
        {
            show_suspension_info(user, suspended);
        }
    }
    else if(IsSuspended(cData))
    {
        chanserv_notice(user, CSMSG_CHANNEL_SUSPENDED, channel->name);
        show_suspension_info(user, cData->suspended);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_netinfo)
{
    extern time_t boot_time;
    extern unsigned long burst_length;
    char interval[INTERVALLEN];
    enum userLevel access;
    int i;

    chanserv_notice(user, CSMSG_NETWORK_INFO);
    chanserv_notice(user, CSMSG_NETWORK_SERVERS, dict_size(servers));
    chanserv_notice(user, CSMSG_NETWORK_USERS, dict_size(clients));
    chanserv_notice(user, CSMSG_NETWORK_OPERS, curr_opers.used);
    chanserv_notice(user, CSMSG_NETWORK_CHANNELS, registered_channels);
    chanserv_notice(user, CSMSG_NETWORK_BANS, banCount);
    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	i = PADLEN - 7 - strlen(accessTitles[access]);
 
	chanserv_notice(user, CSMSG_CHANNEL_USERS, accessTitles[access], i, "", userCount[access]);
    }
    chanserv_notice(user, CSMSG_CHANNEL_USERS, "Total User", PADLEN - 17, "", userCount[0]);
    intervalString(interval, time(NULL) - boot_time);
    chanserv_notice(user, CSMSG_SERVICES_UPTIME, interval);
    intervalString(interval, burst_length);
    chanserv_notice(user, CSMSG_BURST_LENGTH, interval);
    return 1;
}

static void
send_staff_list(struct userNode *to, struct userList *list, int skip_flags)
{
    struct helpfile_table table;
    unsigned int nn;
    struct userNode *user;
    unsigned char *nick;

    table.length = 0;
    table.width = 1;
    table.flags = TABLE_REPEAT_ROWS | TABLE_NO_FREE | TABLE_NO_HEADERS;
    table.contents = alloca(list->used*sizeof(*table.contents));
    for(nn=0; nn<list->used; nn++)
    {
        user = list->list[nn];
        if(user->modes & skip_flags) continue;
        table.contents[table.length] = alloca(table.width*sizeof(**table.contents));
        if(IsAway(user))
        {
            nick = alloca(strlen(user->nick)+3);
            sprintf(nick, "(%s)", user->nick);
        }
        else
        {
            nick = user->nick;
        }
        table.contents[table.length][0] = nick;
        table.length++;
    }
    table_send(chanserv, to->nick, 0, NULL, table);
}

static CHANSERV_FUNC(cmd_ircops)
{
    chanserv_notice(user, CSMSG_STAFF_OPERS);
    send_staff_list(user, &curr_opers, FLAGS_SERVICE);
    return 1;
}

static CHANSERV_FUNC(cmd_helpers)
{
    chanserv_notice(user, CSMSG_STAFF_HELPERS);
    send_staff_list(user, &curr_helpers, FLAGS_OPER);
    return 1;
}

static CHANSERV_FUNC(cmd_staff)
{
    chanserv_notice(user, CSMSG_NETWORK_STAFF);
    cmd_ircops(CSFUNC_ARGS);
    cmd_helpers(CSFUNC_ARGS);
    return 1;
}

static CHANSERV_FUNC(cmd_peek)
{
    struct modeNode *mn;
    char modes[MODELEN];
    unsigned int n;
    struct helpfile_table table;

    irc_make_chanmode(channel, modes);

    chanserv_notice(user, CSMSG_PEEK_INFO, channel->name);
    chanserv_notice(user, CSMSG_PEEK_TOPIC, channel->topic);
    chanserv_notice(user, CSMSG_PEEK_MODES, modes);
    chanserv_notice(user, CSMSG_PEEK_USERS, channel->members.used);
    chanserv_notice(user, CSMSG_PEEK_OPS);

    table.length = 0;
    table.width = 1;
    table.flags = TABLE_REPEAT_ROWS | TABLE_NO_FREE | TABLE_NO_HEADERS;
    table.contents = alloca(channel->members.used*sizeof(*table.contents));
    for(n = 0; n < channel->members.used; n++)
    {
	mn = channel->members.list[n];
	if(!(mn->modes & MODE_CHANOP) || IsLocal(mn->user)) continue;
        table.contents[table.length] = alloca(sizeof(**table.contents));
        table.contents[table.length][0] = mn->user->nick;
        table.length++;
    }
    table_send(chanserv, user->nick, 0, NULL, table);

    return 1;
}

static CHANSERV_FUNC(cmd_setinfo)
{
    struct userData *uData;
    char *infoline;

    uData = GetChannelAccess(channel->channel_info, user->handle_info);

    if(!uData)
    {
	/* If they got past the command restrictions (which require access)
         * but fail this test, we have some fool with security override on. */
	chanserv_notice(user, CSMSG_NOT_USER, channel->name);
	return 0;
    }

    if(argc < 2)
    {
	if(uData->info)
	{
	    chanserv_notice(user, CSMSG_INFO_LINE, uData->info);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NO_INFO_LINE);
	}

	return 1;
    }

    infoline = unsplit_string(argv + 1, argc - 1, NULL);
    if(uData->info) free(uData->info);
    if(infoline[0] == '*' && infoline[1] == 0)
    {
	uData->info = NULL;
	chanserv_notice(user, CSMSG_CLEARED_INFO_LINE);
    }
    else
    {
	uData->info = strdup(infoline);
	chanserv_notice(user, CSMSG_SET_INFO_LINE, infoline);
    }

    return 1;
}

static MODCMD_FUNC(cmd_wipeinfo)
{
    struct handle_info *victim;
    struct userData *ud, *actor;

    REQUIRE_PARAMS(2);
    actor = GetChannelUser(channel->channel_info, user->handle_info);
    if(!(victim = chanserv_get_handle_info(user, argv[1]))) return 0;
    if(!(ud = GetTrueChannelAccess(channel->channel_info, victim)))
    {
        reply(CSMSG_NO_CHAN_USER, argv[1], channel->name);
        return 0;
    }
    if((ud->access >= actor->access) && (ud != actor))
    {
        reply(MSG_USER_OUTRANKED, victim->handle);
        return 0;
    }
    if(ud->info) free(ud->info);
    ud->info = NULL;
    reply(CSMSG_WIPED_INFO_LINE, argv[1], channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_seen)
{
    struct userData *uData;
    struct handle_info *handle;
    char seen[INTERVALLEN];

    REQUIRE_PARAMS(2);

    if(!irccasecmp(argv[1], chanserv->nick))
    {
	chanserv_notice(user, CSMSG_IS_CHANSERV);
	return 1;
    }

    if(!(handle = get_handle_info(argv[1])))
    {
	chanserv_notice(user, MSG_HANDLE_UNKNOWN, argv[1]);
	return 0;
    }

    if(!(uData = GetTrueChannelAccess(channel->channel_info, handle)))
    {
	chanserv_notice(user, CSMSG_NO_CHAN_USER, handle->handle, channel->name);
	return 0;
    }

    if(uData->present)
    {
	chanserv_notice(user, CSMSG_USER_PRESENT, handle->handle);
    }
    else
    {
	if(uData->seen)
	{
	    intervalString(seen, now - uData->seen);
	    chanserv_notice(user, CSMSG_USER_SEEN, handle->handle, channel->name, seen);
	}
	else
	{
	    chanserv_notice(user, CSMSG_NEVER_SEEN, handle->handle, channel->name);
	}

        if(HANDLE_FLAGGED(handle, FROZEN))
        {
            chanserv_notice(user, CSMSG_USER_VACATION, handle->handle);
        }
    }

    return 1;
}

static MODCMD_FUNC(cmd_names)
{
    struct userNode *targ;
    struct userData *targData;
    unsigned int ii, pos, len;
    char buf[400];

    for (ii=pos=0; ii<channel->members.used; ++ii) {
        targ = channel->members.list[ii]->user;
        targData = GetTrueChannelAccess(channel->channel_info, targ->handle_info);
        if (!targData) continue;
        if (pos + strlen(targ->nick) + strlen(targ->handle_info->handle) + 6 > sizeof(buf)) {
            buf[pos] = 0;
            send_message(user, cmd->parent->bot, CSMSG_CHANNEL_NAMES, channel->name, buf);
            pos = 0;
        }
        buf[pos++] = ' ';
        if (IsUserSuspended(targData)) buf[pos++] = 's';
        buf[pos++] = accessChars[targData->access];
        len = strlen(targ->nick);
        memcpy(buf+pos, targ->nick, len);
        pos += len;
        buf[pos++] = '(';
        len = strlen(targ->handle_info->handle);
        memcpy(buf+pos, targ->handle_info->handle, len);
        pos += len;
        buf[pos++] = ')';
    }
    buf[pos] = 0;
    send_message(user, cmd->parent->bot, CSMSG_CHANNEL_NAMES, channel->name, buf);
    send_message(user, cmd->parent->bot, CSMSG_END_NAMES, channel->name);
    return 1;
}

static int
note_type_visible_to_user(struct chanData *channel, struct note_type *ntype, struct userNode *user)
{
    switch(ntype->visible_type)
    {
    case NOTE_VIS_ALL: return 1;
    case NOTE_VIS_CHANNEL_USERS: return !channel || !user || (user->handle_info && GetChannelUser(channel, user->handle_info));
    case NOTE_VIS_PRIVILEGED: default: return user && (IsOper(user) || IsSupportHelper(user) || IsNetworkHelper(user));
    }
}

static int
note_type_settable_by_user(struct chanNode *channel, struct note_type *ntype, struct userNode *user)
{
    struct userData *uData;

    switch(ntype->set_access_type)
    {
    case NOTE_SET_CHANNEL_ACCESS:
        if(!user->handle_info)
            return 0;
        if(!(uData = GetChannelUser(channel->channel_info, user->handle_info)))
            return 0;
        return uData->access >= ntype->set_access.min_ulevel;
    case NOTE_SET_CHANNEL_SETTER:
        return check_user_level(user, channel, channel->channel_info->options[optSetters], 1, 0);
    case NOTE_SET_PRIVILEGED: default: return IsHelping(user);
    }
}

static CHANSERV_FUNC(cmd_note)
{
    struct chanData *cData;
    struct note *note;
    struct note_type *ntype;

    cData = channel->channel_info;
    if(!cData)
    {
        reply(CSMSG_NOT_REGISTERED, channel->name);
        return 0;
    }

    /* If no arguments, show all visible notes for the channel. */
    if(argc < 2)
    {
        dict_iterator_t it;
        unsigned int count;

        for(count=0, it=dict_first(cData->notes); it; it=iter_next(it))
        {
            note = iter_data(it);
            if(!note_type_visible_to_user(cData, note->type, user))
                continue;
            if(!count++)
                chanserv_notice(user, CSMSG_NOTELIST_HEADER, channel->name);
            chanserv_notice(user, CSMSG_NOTE_FORMAT, iter_key(it), note->setter, note->note);
        }
        if(count)
            chanserv_notice(user, CSMSG_NOTELIST_END, channel->name);
        else
            chanserv_notice(user, CSMSG_NOTELIST_EMPTY, channel->name);
    }
    /* If one argument, show the named note. */
    else if(argc == 2)
    {
        if((note = dict_find(cData->notes, argv[1], NULL))
           && note_type_visible_to_user(cData, note->type, user))
        {
            chanserv_notice(user, CSMSG_NOTE_FORMAT, note->type->name, note->setter, note->note);
        }
        else if((ntype = dict_find(note_types, argv[1], NULL))
                && note_type_visible_to_user(NULL, ntype, user))
        {
            chanserv_notice(user, CSMSG_NO_SUCH_NOTE, channel->name, ntype->name);
            return 0;
        }
        else
        {
            chanserv_notice(user, CSMSG_BAD_NOTE_TYPE, argv[1]);
            return 0;
        }
    }
    /* Assume they're trying to set a note. */
    else
    {
        char *note_text;
        ntype = dict_find(note_types, argv[1], NULL);
        if(!ntype)
        {
            chanserv_notice(user, CSMSG_BAD_NOTE_TYPE, argv[1]);
            return 0;
        }
        else if(note_type_settable_by_user(channel, ntype, user))
        {
            note_text = unsplit_string(argv+2, argc-2, NULL);
            chanserv_add_channel_note(cData, ntype, user->handle_info->handle, note_text);
            chanserv_notice(user, CSMSG_NOTE_SET, ntype->name, channel->name);

            if(ntype->visible_type == NOTE_VIS_PRIVILEGED)
            {
                /* The note is viewable to staff only, so return 0
                   to keep the invocation from getting logged (or
                   regular users can see it in !events). */
                return 0;
            }
        }
        else
        {
            chanserv_notice(user, CSMSG_NO_ACCESS);
            return 0;
        }
    }
    return 1;
}

static CHANSERV_FUNC(cmd_delnote)
{
    struct note *note;

    REQUIRE_PARAMS(2);
    if(!(note = dict_find(channel->channel_info->notes, argv[1], NULL))
       || !note_type_settable_by_user(channel, note->type, user))
    {
        chanserv_notice(user, CSMSG_NO_SUCH_NOTE, channel->name, argv[1]);
        return 0;
    }
    dict_remove(channel->channel_info->notes, note->type->name);
    chanserv_notice(user, CSMSG_NOTE_REMOVED, argv[1], channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_events)
{
    struct logSearch discrim;
    struct logReport report;
    unsigned int matches, limit;

    limit = (argc > 1) ? atoi(argv[1]) : 10;
    if(limit < 1 || limit > 200) limit = 10;

    memset(&discrim, 0, sizeof(discrim));
    discrim.masks.bot = chanserv;
    discrim.masks.channel_name = channel->name;
    if(argc > 2) discrim.masks.command = argv[2];
    discrim.limit = limit;
    discrim.max_time = INT_MAX;
    discrim.severities = 1 << LOG_COMMAND;
    report.reporter = chanserv;
    report.user = user;
    chanserv_notice(user, MSG_SEARCH_RESULTS, "channel events");
    matches = log_entry_search(&discrim, log_report_entry, &report);
    if(matches)
    {
	chanserv_notice(user, MSG_MATCH_COUNT, matches);
    }
    else
    {
	chanserv_notice(user, MSG_NO_MATCHES);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_say)
{
    char *msg;
    if (channel) {
        REQUIRE_PARAMS(2);
        msg = unsplit_string(argv + 1, argc - 1, NULL);
        send_channel_message(channel, cmd->parent->bot, "%s", msg);
    } else if (GetUserH(argv[1])) {
        REQUIRE_PARAMS(3);
        msg = unsplit_string(argv + 2, argc - 2, NULL);
        send_target_message(1, argv[1], cmd->parent->bot, "%s", msg);
    } else {
        send_message(user, cmd->parent->bot, "You must specify the name of a channel or user.");
        return 0;
    }
    return 1;
}

static CHANSERV_FUNC(cmd_emote)
{
    char *msg;
    assert(argc >= 2);
    if (channel) {
        /* CTCP is so annoying. */
        msg = unsplit_string(argv + 1, argc - 1, NULL);
        send_channel_message(channel, cmd->parent->bot, "\001ACTION %s\001", msg);
    } else if (GetUserH(argv[1])) {
        msg = unsplit_string(argv + 2, argc - 2, NULL);
        send_target_message(1, argv[1], cmd->parent->bot, "\001ACTION %s\001", msg);
    } else {
        send_message(user, cmd->parent->bot, "You must specify the name of a channel or user.");
        return 0;
    }
    return 1;
}

struct channelList *
chanserv_support_channels(void)
{
    return &chanserv_conf.support_channels;
}

static CHANSERV_FUNC(cmd_expire)
{
    int channel_count = registered_channels;
    expire_channels(NULL);
    chanserv_notice(user, CSMSG_CHANNELS_EXPIRED, channel_count - registered_channels);
    return 1;
}

static void
chanserv_expire_suspension(void *data)
{
    struct suspended *suspended = data;
    struct chanNode *channel;

    if(!suspended->expires || (now < suspended->expires)) suspended->revoked = now;
    channel = suspended->cData->channel;
    AddChannelUser(chanserv, channel);
    AddChannelOp(1, &chanserv, channel, chanserv, 1);
    suspended->cData->channel = channel;
    suspended->cData->flags &= ~CHANNEL_SUSPENDED;
}

static CHANSERV_FUNC(cmd_csuspend)
{
    struct suspended *suspended;
    char reason[MAXLEN];
    time_t expiry;
    struct userData *uData;

    if(IsProtected(channel->channel_info))
    {
        chanserv_notice(user, CSMSG_SUSPEND_NODELETE, channel->name);
        return 0;
    }

    if(IsSuspended(channel->channel_info))
    {
	chanserv_notice(user, CSMSG_ALREADY_SUSPENDED, channel->name);
	return 0;
    }

    REQUIRE_PARAMS(3);

    if(!strcmp(argv[1], "0"))
    { 
        expiry = 0;
    }
    else
    {
        unsigned int duration;
        if(!(duration = ParseInterval(argv[1])))
        {
            chanserv_notice(user, MSG_INVALID_DURATION, argv[1]);
            return 0;
        }
        expiry = now + duration;
    }

    unsplit_string(argv + 2, argc - 2, reason);

    suspended = calloc(1, sizeof(*suspended));
    suspended->revoked = 0;
    suspended->issued = now;
    suspended->suspender = strdup(user->handle_info->handle);
    suspended->expires = expiry;
    if(suspended->expires)
    {
        timeq_add(suspended->expires, chanserv_expire_suspension, suspended);
    }
    suspended->reason = strdup(reason);
    suspended->cData = channel->channel_info;
    suspended->previous = suspended->cData->suspended;
    suspended->cData->suspended = suspended;

    /* Mark all users in channel as absent. */
    for(uData = channel->channel_info->users; uData; uData = uData->next)
    {
        if(uData->present)
        {
            uData->seen = now;
            uData->present = 0;
        }
    }

    /* Mark the channel as suspended, then part. */
    channel->channel_info->flags |= CHANNEL_SUSPENDED;
    DelChannelUser(chanserv, channel, suspended->reason, 0);
    chanserv_notice(user, CSMSG_SUSPENDED, channel->name);

    sprintf(reason, "%s suspended by %s.", channel->name, suspended->suspender);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);
    return 1;
}

static CHANSERV_FUNC(cmd_cunsuspend)
{
    struct suspended *suspended;
    char message[MAXLEN];

    if(!IsSuspended(channel->channel_info))
    {
        chanserv_notice(user, CSMSG_NOT_SUSPENDED, channel->name);
        return 0;
    }

    suspended = channel->channel_info->suspended;

    /* Expire the suspension and join ChanServ to the channel. */
    timeq_del(suspended->expires, chanserv_expire_suspension, suspended, 0);
    chanserv_expire_suspension(suspended);

    assert(channel == suspended->cData->channel);

    chanserv_notice(user, CSMSG_UNSUSPENDED, channel->name);
    sprintf(message, "%s unsuspended by %s.", channel->name, user->handle_info->handle);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, message);
    return 1;
}

typedef struct chanservSearch
{
    char *name;
    char *registrar;

    time_t unvisited;
    time_t registered;

    unsigned long flags;
    unsigned int limit;
} *search_t;

typedef void (*channel_search_func)(struct chanData *channel, void *data);

static search_t
chanserv_search_create(struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    search_t search;
    unsigned int i;

    search = malloc(sizeof(struct chanservSearch));
    memset(search, 0, sizeof(*search));
    search->limit = 25;

    for(i = 0; i < argc; i++)
    {
	/* Assume all criteria require arguments. */
	if(i == (argc - 1))
	{
	    chanserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
	    return NULL;
	}

	if(!irccasecmp(argv[i], "name"))
	{
	    search->name = argv[++i];
	}
	else if(!irccasecmp(argv[i], "registrar"))
	{
	    search->registrar = argv[++i];
	}
	else if(!irccasecmp(argv[i], "unvisited"))
	{
	    search->unvisited = ParseInterval(argv[++i]);
	}
	else if(!irccasecmp(argv[i], "registered"))
	{
	    search->registered = ParseInterval(argv[++i]);
	}
	else if(!irccasecmp(argv[i], "flags"))
	{
	    i++;
	    if(!irccasecmp(argv[i], "nodelete"))
	    {
		search->flags |= CHANNEL_NODELETE;
	    }
	    else if(!irccasecmp(argv[i], "suspended"))
	    {
		search->flags |= CHANNEL_SUSPENDED;
	    }
	    else
	    {
		chanserv_notice(user, CSMSG_INVALID_CFLAG, argv[i]);
		goto fail;
	    }
	}
	else if(!irccasecmp(argv[i], "limit"))
	{
	    search->limit = strtoul(argv[++i], NULL, 10);
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_CRITERIA, argv[i]);
	    goto fail;
	}
    }

    if(search->name && !strcmp(search->name, "*"))
    {
	search->name = 0;
    }
    if(search->registrar && !strcmp(search->registrar, "*"))
    {
	search->registrar = 0;
    }

    return search;
  fail:
    free(search);
    return NULL;
}

static int
chanserv_channel_match(struct chanData *channel, search_t search)
{
    const char *name = channel->channel->name;
    if((search->name && !match_ircglob(name, search->name)) ||
       (search->registrar && !channel->registrar) ||
       (search->registrar && !match_ircglob(channel->registrar, search->registrar)) ||
       (search->unvisited && (now - channel->visited) < search->unvisited) ||
       (search->registered && (now - channel->registered) > search->registered) ||
       (search->flags && ((search->flags & channel->flags) != search->flags)))
    {
	return 0;
    }

    return 1;
}

static unsigned int
chanserv_channel_search(search_t search, channel_search_func smf, void *data)
{
    struct chanData *channel;
    unsigned int matches = 0;

    for(channel = channelList; channel && matches < search->limit; channel = channel->next)
    {
	if(!chanserv_channel_match(channel, search)) continue;
	matches++;
	smf(channel, data);
    }

    return matches;
}

static void
search_count(UNUSED_ARG(struct chanData *channel), UNUSED_ARG(void *data))
{
}

static void
search_print(struct chanData *channel, void *data)
{
    chanserv_notice(data, CSMSG_SEARCH_DATA, channel->channel->name);
}

static CHANSERV_FUNC(cmd_search)
{
    search_t search;
    unsigned int matches;
    channel_search_func action;

    REQUIRE_PARAMS(3);

    if(!irccasecmp(argv[1], "count"))
    {
	action = search_count;
    }
    else if(!irccasecmp(argv[1], "print"))
    {
	action = search_print;
    }
    else
    {
	chanserv_notice(user, CSMSG_ACTION_INVALID, argv[1]);
	return 0;
    }

    search = chanserv_search_create(user, argc - 2, argv + 2);
    if(!search) return 0;

    if(action == search_count)
    {
	search->limit = INT_MAX;
    }

    if(action == search_print)
    {
	chanserv_notice(user, MSG_SEARCH_RESULTS, "channels");
    }

    matches = chanserv_channel_search(search, action, user);

    if(matches)
    {
	chanserv_notice(user, MSG_MATCH_COUNT, matches);
    }
    else
    {
	chanserv_notice(user, MSG_NO_MATCHES);
    }

    free(search);
    return 1;
}

static CHANSERV_FUNC(cmd_unvisited)
{
    struct chanData *cData;
    time_t interval = chanserv_conf.channel_expire_delay;
    char buffer[INTERVALLEN];
    unsigned int limit = 25, matches = 0;

    if(argc > 1)
    {
	interval = ParseInterval(argv[1]);
	if(argc > 2) limit = atoi(argv[2]);
    }

    intervalString(buffer, interval);
    chanserv_notice(user, CSMSG_UNVISITED_HEADER, limit, buffer);

    for(cData = channelList; cData && matches < limit; cData = cData->next)
    {
	if((now - cData->visited) < interval) continue;

	intervalString(buffer, now - cData->visited);
	chanserv_notice(user, CSMSG_UNVISITED_DATA, cData->channel->name, buffer);
	matches++;
    }

    return 1;
}

#define CHANNEL_OPTION_FUNC(NAME) MODCMD_FUNC(chan_##NAME)

#define CHANNEL_BINARY_OPTION(args...) return channel_binary_option(args, CSFUNC_ARGS);
#define CHANNEL_MULTIPLE_OPTION(args...) return channel_multiple_option(args, values, ArrayLength(values), CSFUNC_ARGS);

static CHANNEL_OPTION_FUNC(opt_defaulttopic)
{
    if(argc > 1)
    {
        char *topic;

        if(!check_user_level(user, channel, channel->channel_info->options[optEnfTopic], 1, 0))
        {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }

	topic = unsplit_string(argv+1, argc-1, NULL);

	if(topic[0] == '*' && topic[1] == 0)
	{
	    channel->channel_info->topic[0] = 0;
	}
	else
	{
	    safestrncpy(channel->channel_info->topic, topic, sizeof(channel->channel_info->topic));
            if(channel->channel_info->topic_mask
               && !match_ircglob(channel->channel_info->topic, channel->channel_info->topic_mask))
            {
                chanserv_notice(user, CSMSG_TOPIC_MISMATCH, channel->name);
            }
	}
        SetChannelTopic(channel, chanserv, channel->channel_info->topic, 1);
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "DefaultTopic ", channel->channel_info->topic[0] ? channel->channel_info->topic : "None.");
    return 1;
}

static CHANNEL_OPTION_FUNC(opt_topicmask)
{
    if(argc > 1)
    {
        struct chanData *cData = channel->channel_info;
        char *mask;

        if(!check_user_level(user, channel, cData->options[optEnfTopic], 1, 0))
        {
            chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
            return 0;
        }

	mask = unsplit_string(argv+1, argc-1, NULL);

        if(cData->topic_mask) free(cData->topic_mask);
	if(mask[0] == '*' && mask[1] == 0)
	{
	    cData->topic_mask = 0;
	}
	else
	{
            cData->topic_mask = strdup(mask);
            if(!cData->topic[0])
            {
                chanserv_notice(user, CSMSG_MASK_BUT_NO_TOPIC, channel->name);
            }
            else if(!match_ircglob(cData->topic, cData->topic_mask))
            {
                chanserv_notice(user, CSMSG_TOPIC_MISMATCH, channel->name);
            }
	}
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, "TopicMask    ", channel->channel_info->topic_mask ? channel->channel_info->topic_mask : "None.");
    return 1;
}

int opt_greeting_common(struct userNode *user, UNUSED_ARG(struct chanNode *channel), int argc, unsigned char *argv[], char *name, char **data)
{
    if(argc > 1)
    {
        char *greeting = unsplit_string(argv+1, argc-1, NULL);
        char *previous;

        previous = *data;
	if(greeting[0] == '*' && greeting[1] == 0)
	{
	    *data = NULL;
	}
	else
	{
	    unsigned int length;

	    length = strlen(greeting);
	    if(length > chanserv_conf.greeting_length)
	    {
		chanserv_notice(user, CSMSG_GREETING_TOO_LONG, length, chanserv_conf.greeting_length);
		return 0;
	    }

	    *data = strdup(greeting);
	}
        if(previous)
        {
            free(previous);
        }
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, name, *data ? *data : "None.");
    return 1;
}

static CHANNEL_OPTION_FUNC(opt_greeting)
{
    opt_greeting_common(user, channel, argc, argv, "Greeting     ", &channel->channel_info->greeting);
    return 1;
}

static CHANNEL_OPTION_FUNC(opt_usergreeting)
{
    opt_greeting_common(user, channel, argc, argv, "UserGreeting ", &channel->channel_info->user_greeting);
    return 1;
}

static CHANNEL_OPTION_FUNC(opt_modes)
{
    char modes[MODELEN];

    if(argc > 1)
    {
        if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
        {
            chanserv_notice(user, CSMSG_NO_ACCESS);
            return 0;
        }
	if(argv[1][0] == '*' && argv[1][1] == 0)
	{
	    channel->channel_info->mode_on = 0;
	    channel->channel_info->mode_off = 0;
	    modes[0] = 0;
	}
	else
	{
	    if(!parse_mode_lock(argv+1, argc-1, channel->channel_info))
	    {
		chanserv_notice(user, CSMSG_INVALID_STR_SET, argv[1], "mode lock");
		return 0;
	    }

	    create_mode_lock(modes, channel->channel_info, 0);
	    mod_chanmode(channel, chanserv, modes);
	    irc_mode(chanserv, channel, modes);
	}
    }

    create_mode_lock(modes, channel->channel_info, 1);
    chanserv_notice(user, CSMSG_STRING_VALUE, "Modes        ", strlen(modes) ? modes : "None.");
    return 1;
}

struct valueData
{
    char *description;
    char value;
};

static int
channel_binary_option(char *name, unsigned long mask, struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd))
{
    struct chanData *cData = channel->channel_info;
    int value;

    if(argc > 1)
    {
	/* Set flag according to value. */
	if(enabled_string(argv[1]))
	{
	    cData->flags |= mask;
	    value = 1;
	}
	else if(disabled_string(argv[1]))
	{
	    cData->flags &= ~mask;
	    value = 0;
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_BINARY, argv[1]);
	    return 0;
	}
    }
    else
    {
	/* Find current option value. */
	value = (cData->flags & mask) ? 1 : 0;
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, name, value ? "Enabled." : "Disabled.");
    return 1;
}

static int
channel_multiple_option(char *name, char *description, enum opt option, int check_level, struct valueData *values, int count, struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd))
{
    struct chanData *cData = channel->channel_info;
    int index;

    if(argc > 1)
    {
	/* Set flag according to value. */
        if((check_level != -1) && !check_user_level(user, channel, cData->options[check_level], 1, 1))
        {
            chanserv_notice(user, CSMSG_CANNOT_SET);
            return 0;
        }

        index = atoi(argv[1]);

	if(index < 0 || index >= count)
	{
	    chanserv_notice(user, CSMSG_INVALID_NUM_SET, index, description);
            /* Show possible values. */
            for(index = 0; index < count; index++)
            {
                chanserv_notice(user, CSMSG_NUMERIC_VALUE, name, index, values[index].description);
            }
	    return 0;
	}

        if((check_level != -1) && !check_user_level(user, channel, values[index].value, 1, 1))
        {
            chanserv_notice(user, CSMSG_BAD_SETLEVEL);
            return 0;
        }

	cData->options[option] = values[index].value;
    }
    else
    {
	/* Find current option value. */
	for(index = 0; index < count && cData->options[option] != values[index].value; index++);
        if(index == count)
        {
            /* Somehow, the option value is corrupt; reset it to the default. */
            cData->options[option] = CHANNEL_DEFAULT_OPTIONS[option];
            for(index = 0; index < count && cData->options[option] != values[index].value; index++);
        }
    }

    chanserv_notice(user, CSMSG_NUMERIC_VALUE, name, index, values[index].description);
    return 1;
}

static CHANNEL_OPTION_FUNC(opt_nodelete)
{
    if((argc > 1) && (!IsOper(user) || !user->handle_info || (user->handle_info->opserv_level < chanserv_conf.nodelete_level)))
    {
	chanserv_notice(user, MSG_SETTING_PRIVILEGED, argv[0]);
	return 0;
    }

    CHANNEL_BINARY_OPTION("NoDelete     ", CHANNEL_NODELETE);
}

static CHANNEL_OPTION_FUNC(opt_userinfo)
{
    CHANNEL_BINARY_OPTION("UserInfo     ", CHANNEL_INFO_LINES);
}

static CHANNEL_OPTION_FUNC(opt_voice)
{
    CHANNEL_BINARY_OPTION("Voice        ", CHANNEL_VOICE_ALL);
}

static CHANNEL_OPTION_FUNC(opt_dynlimit)
{
    CHANNEL_BINARY_OPTION("DynLimit     ", CHANNEL_DYNAMIC_LIMIT);
}

static CHANNEL_OPTION_FUNC(opt_topicsnarf)
{
    if((argc > 0) && !check_user_level(user, channel, channel->channel_info->options[optEnfTopic], 1, 0))
    {
        chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
        return 0;
    }
    CHANNEL_BINARY_OPTION("TopicSnarf   ", CHANNEL_TOPIC_SNARF);
}

static CHANNEL_OPTION_FUNC(opt_peoninvite)
{
    CHANNEL_BINARY_OPTION("PeonInvite   ", CHANNEL_PEON_INVITE);
}

static CHANNEL_OPTION_FUNC(opt_strictop)
{
    struct valueData values[] =
    {
	{"Anyone may op unknown users.", 'a'},
	{"Ops and above may op unknown users.", 'o'},
	{"Masters and above may op unknown users.", 'm'},
	{"Coowners and above may op unknown users.", 'c'},
	{"Owners may op unknown users.", 'w'},
	{"No one may op unknown users.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("StrictOp     ", "strict op", optStrictOp, optStrictOp);
}

static CHANNEL_OPTION_FUNC(opt_autoop)
{
    struct valueData values[] =
    {
	{"Everyone will be automatically opped.", 'a'},
	{"Ops and above will be automatically opped.", 'o'},
	{"Masters and above will be automatically opped.", 'm'},
	{"Coowners and above will be automatically opped.", 'c'},
	{"Owners will be automatically opped.", 'w'},
	{"No one will be automatically opped.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("AutoOp       ", "auto op", optAutoOp, optAutoOp);
}

static CHANNEL_OPTION_FUNC(opt_protect)
{
    struct valueData values[] =
    {
	{"Non-users and users will be protected from those of equal or lower access.", 'a'},
	{"Users will be protected from those of equal or lower access.", 'e'},
	{"Users will be protected from those of lower access.", 'l'},
	{"No users will be protected.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("Protect      ", "user protection", optProtect, -1);
}

static CHANNEL_OPTION_FUNC(opt_enfmodes)
{
    struct valueData values[] =
    {
	{"Anyone may override the mode lock.", 'a'},
	{"Ops and above may override the mode lock.", 'o'},
	{"Masters and above may override the mode lock.", 'm'},
	{"Coowners and above may override the mode lock.", 'c'},
	{"Owners may override the mode lock.", 'w'},
	{"No one may override the mode lock.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("EnfModes     ", "mode lock enforcement", optEnfModes, optEnfModes);
}

static CHANNEL_OPTION_FUNC(opt_enftopic)
{
    struct valueData values[] =
    {
	{"Anyone may override the topic lock.", 'a'},
	{"Ops and above may override the topic lock.", 'o'},
	{"Masters and above may override the topic lock.", 'm'},
	{"Coowners and above may override the topic lock.", 'c'},
	{"Owners may override the topic lock.", 'w'},
	{"No one may override the topic lock.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("EnfTopic     ", "topic lock enforcement", optEnfTopic, optEnfTopic);
}

static CHANNEL_OPTION_FUNC(opt_pubcmd)
{
    struct valueData values[] =
    {
	{"Anyone may use public commands.", 'a'},
        {"Peons and above may use public commands.", 'p'},
	{"Ops and above may use public commands.", 'o'},
	{"Masters and above may use public commands.", 'm'},
	{"Coowners and above may use public commands.", 'c'},
	{"Owners may use public commands.", 'w'},
	{"No one may use public commands.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("PubCmd       ", "public command", optPubCmd, optPubCmd);
}

static CHANNEL_OPTION_FUNC(opt_toys)
{
    struct valueData values[] =
    {
	{"Toys are completely disabled.", 'd'},
	{"Toys will only reply privately.", 'n'},
	{"Toys will reply publicly.", 'p'}
    };

    CHANNEL_MULTIPLE_OPTION("Toys         ", "toy control", optToys, -1);
}

static CHANNEL_OPTION_FUNC(opt_setters)
{
    struct valueData values[] =
    {
	{"Masters and above may change settings.", 'm'},
	{"Co-owners and above may change settings.", 'c'},
	{"Owner may change settings.", 'w'}
    };

    CHANNEL_MULTIPLE_OPTION("Setters      ", "setters", optSetters, optSetters);
}

static CHANNEL_OPTION_FUNC(opt_ctcpusers)
{
    struct valueData values[] =
    {
	{"Anyone may send CTCPs to the channel.", 'a'},
        {"Peons and above may send CTCPs to the channel.", 'p'},
	{"Ops and above may send CTCPs to the channel.", 'o'},
	{"Masters and above may send CTCPs to the channel.", 'm'},
	{"Coowners and above may send CTCPs to the channel.", 'c'},
	{"Owners may send CTCPs to the channel.", 'w'},
	{"No one may send CTCPs to the channel.", 'n'}
    };

    CHANNEL_MULTIPLE_OPTION("CTCPUsers    ", "CTCP allowed to", optCTCPUsers, optCTCPUsers);
}

static CHANNEL_OPTION_FUNC(opt_ctcpreaction)
{
    struct valueData values[] =
    {
        {"Kick on disallowed CTCPs", 'k'},
        {"Kickban on disallowed CTCPs", 'b'},
        {"Short timed ban on disallowed CTCPs", 't'},
        {"Long timed ban on disallowed CTCPs", 'T'}
    };

    CHANNEL_MULTIPLE_OPTION("CTCPReaction ", "CTCP reaction", optCTCPReaction, -1);
}

static CHANNEL_OPTION_FUNC(opt_topicrefresh)
{
    struct valueData values[] =
    {
        {"Never refresh topic.", 'n'},
        {"Refresh every 3 hours.", '1'},
        {"Refresh every 6 hours.", '2'},
        {"Refresh every 12 hours.", '3'},
        {"Refresh every 24 hours.", '4'}
    };

    CHANNEL_MULTIPLE_OPTION("TopicRefresh ", "topic refresh", optTopicRefresh, -1);
}

static struct svccmd_list set_shows_list;

static void
handle_svccmd_unbind(struct svccmd *target) {
    unsigned int ii;
    for (ii=0; ii<set_shows_list.used; ++ii) {
        if (target == set_shows_list.list[ii]) {
            set_shows_list.used = 0;
        }
    }
}

static CHANSERV_FUNC(cmd_set)
{
    struct svccmd *subcmd;
    char buf[MAXLEN];
    unsigned int ii;

    /* Check if we need to (re-)initialize set_shows_list. */
    if(!set_shows_list.used)
    {
        if(!set_shows_list.size)
        {
            set_shows_list.size = chanserv_conf.set_shows->used;
            set_shows_list.list = calloc(set_shows_list.size, sizeof(set_shows_list.list[0]));
        }
        for(ii = 0; ii < chanserv_conf.set_shows->used; ii++)
        {
            const char *name = chanserv_conf.set_shows->list[ii];
            sprintf(buf, "%s %s", argv[0], name);
            subcmd = dict_find(cmd->parent->commands, buf, NULL);
            if(!subcmd)
            {
                log_module(CS_LOG, LOG_ERROR, "Unable to find set option \"%s\".", name);
                continue;
            }
            svccmd_list_append(&set_shows_list, subcmd);
        }
    }

    if(argc < 2)
    {
	chanserv_notice(user, CSMSG_CHANNEL_OPTIONS);
        for(ii = 0; ii < set_shows_list.used; ii++)
        {
            subcmd = set_shows_list.list[ii];
            subcmd->command->func(user, channel, 1, argv+1, subcmd);
        }
	return 1;
    }

    sprintf(buf, "%s %s", argv[0], argv[1]);
    subcmd = dict_find(cmd->parent->commands, buf, NULL);
    if(!subcmd)
    {
        reply(CSMSG_INVALID_OPTION, argv[1], argv[0]);
        return 0;
    }
    if((argc > 2)
       && !check_user_level(user, channel, channel->channel_info->options[optSetters], 1, 0))
    {
        reply(CSMSG_NO_ACCESS);
        return 0;
    }

    return subcmd->command->func(user, channel, argc - 1, argv + 1, subcmd);
}

#define USER_OPTION_FUNC(NAME) MODCMD_FUNC(user_##NAME)

#define USER_BINARY_OPTION(args...) return user_binary_option(args, CSFUNC_ARGS);

static int
user_binary_option(char *name, unsigned long mask, struct userNode *user, struct chanNode *channel, int argc, unsigned char *argv[], UNUSED_ARG(struct svccmd *cmd))
{
    struct userData *uData;
    int value;

    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    if(!uData)
    {
        chanserv_notice(user, CSMSG_NOT_USER, channel->name);
        return 0;
    }

    if(argc > 1)
    {
	/* Set flag according to value. */
	if(enabled_string(argv[1]))
	{
	    uData->flags |= mask;
	    value = 1;
	}
	else if(disabled_string(argv[1]))
	{
	    uData->flags &= ~mask;
	    value = 0;
	}
	else
	{
	    chanserv_notice(user, MSG_INVALID_BINARY, argv[1]);
	    return 0;
	}
    }
    else
    {
	/* Find current option value. */
	value = (uData->flags & mask) ? 1 : 0;
    }

    chanserv_notice(user, CSMSG_STRING_VALUE, name, value ? "Enabled." : "Disabled.");
    return 1;
}

static USER_OPTION_FUNC(opt_noautoop)
{
    struct userData *uData;

    uData = GetChannelAccess(channel->channel_info, user->handle_info);
    if(!uData)
    {
        chanserv_notice(user, CSMSG_NOT_USER, channel->name);
        return 0;
    }
    USER_BINARY_OPTION(uData->access > ulPeon ? "NoAutoOp     " : "NoAutoVoice  ", USER_AUTO_OP);
}

static USER_OPTION_FUNC(opt_autoinvite)
{
    USER_BINARY_OPTION("AutoInvite   ", USER_AUTO_INVITE);
}

struct svccmd_list uset_shows_list;

static CHANSERV_FUNC(cmd_uset)
{
    struct svccmd *subcmd;
    char buf[MAXLEN];
    unsigned int ii;

    /* Check if we need to (re-)initialize uset_shows_list. */
    if(!uset_shows_list.used)
    {
        char *options[] =
        {
            "NoAutoOp", "AutoInvite"
        };

        if(!uset_shows_list.size)
        {
            uset_shows_list.size = ArrayLength(options);
            uset_shows_list.list = calloc(uset_shows_list.size, sizeof(uset_shows_list.list[0]));
        }
        for(ii = 0; ii < ArrayLength(options); ii++)
        {
            const char *name = options[ii];
            sprintf(buf, "%s %s", argv[0], name);
            subcmd = dict_find(cmd->parent->commands, buf, NULL);
            if(!subcmd)
            {
                log_module(CS_LOG, LOG_ERROR, "Unable to find uset option %s.", name);
                continue;
            }
            svccmd_list_append(&uset_shows_list, subcmd);
        }
    }

    if(argc < 2)
    {
	/* Do this so options are presented in a consistent order. */
	chanserv_notice(user, CSMSG_USER_OPTIONS);
        for(ii = 0; ii < uset_shows_list.used; ii++)
        {
            uset_shows_list.list[ii]->command->func(user, channel, 1, argv+1, uset_shows_list.list[ii]);
        }
	return 1;
    }

    sprintf(buf, "%s %s", argv[0], argv[1]);
    subcmd = dict_find(cmd->parent->commands, buf, NULL);
    if(!subcmd)
    {
        reply(CSMSG_INVALID_OPTION, argv[1], argv[0]);
        return 0;
    }

    return subcmd->command->func(user, channel, argc - 1, argv + 1, subcmd);
}

static CHANSERV_FUNC(cmd_giveownership)
{
    struct handle_info *new_owner_hi;
    struct userData *new_owner, *curr_user;
    struct chanData *cData = channel->channel_info;
    unsigned int force;
    struct do_not_register *dnr;
    char reason[MAXLEN];

    REQUIRE_PARAMS(2);
    curr_user = GetChannelAccess(cData, user->handle_info);
    force = IsHelping(user) && (argc > 2) && !irccasecmp(argv[2], "force");
    if(!curr_user || (curr_user->access != ulOwner))
    {
        switch (cData->userCount[ulOwner]) {
        case 0:
            curr_user = 0;
            break;
        case 1:
            for(curr_user = channel->channel_info->users;
                curr_user && (curr_user->access != ulOwner);
                curr_user = curr_user->next) ;
            break;
        default:
            chanserv_notice(user, CSMSG_MULTIPLE_OWNERS, channel->name);
            return 0;
        }
    }
    if(!(new_owner_hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    if(new_owner_hi == user->handle_info)
    {
        chanserv_notice(user, CSMSG_NO_TRANSFER_SELF);
        return 0;
    }
    new_owner = GetChannelAccess(cData, new_owner_hi);
    if(!new_owner)
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, new_owner_hi->handle, channel->name);
        return 0;
    }
    if((chanserv_get_owned_count(new_owner_hi) >= chanserv_conf.max_owned) && !force)
    {
        chanserv_notice(user, CSMSG_OWN_TOO_MANY, new_owner_hi->handle, chanserv_conf.max_owned);
        return 0;
    }
    if((dnr = chanserv_is_dnr(NULL, new_owner_hi)) && !force) {
        if(IsHelping(user))
        {
            chanserv_notice(user, CSMSG_DNR_INFO, dnr->chan_name, dnr->setter, dnr->reason);
        }
        else
        {
            chanserv_notice(user, CSMSG_DNR_ACCOUNT, new_owner_hi->handle);
        }
        return 0;
    }
    userCount[new_owner->access]--;
    cData->userCount[new_owner->access]--;
    new_owner->access = ulOwner;
    cData->userCount[new_owner->access]++;
    userCount[new_owner->access]++;
    if(curr_user)
    {
        userCount[curr_user->access]--;
        cData->userCount[curr_user->access]--;
        curr_user->access = ulCoowner;
        cData->userCount[curr_user->access]++;
        userCount[curr_user->access]++;
    }
    chanserv_notice(user, CSMSG_OWNERSHIP_GIVEN, channel->name, new_owner_hi->handle);

    sprintf(reason, "%s ownership transferred to %s by %s.", channel->name, new_owner_hi->handle, user->handle_info->handle);
    global_message(MESSAGE_RECIPIENT_OPERS | MESSAGE_RECIPIENT_HELPERS, reason);

    return 1;
}

static CHANSERV_FUNC(cmd_suspend)
{
    struct handle_info *hi;
    struct userData *self, *target;

    REQUIRE_PARAMS(2);
    if(!(hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    self = GetChannelUser(channel->channel_info, user->handle_info);
    if(!(target = GetTrueChannelAccess(channel->channel_info, hi)))
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, hi->handle, channel->name);
        return 0;
    }
    if(target->access >= self->access)
    {
        chanserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
        return 0;
    }
    if(target->flags & USER_SUSPENDED)
    {
        chanserv_notice(user, CSMSG_ALREADY_SUSPENDED, hi->handle);
        return 0;
    }
    if(target->present)
    {
        target->present = 0;
        target->seen = now;
    }
    target->flags |= USER_SUSPENDED;
    chanserv_notice(user, CSMSG_USER_SUSPENDED, hi->handle, channel->name);
    return 1;
}

static CHANSERV_FUNC(cmd_unsuspend)
{
    struct handle_info *hi;
    struct userData *self, *target;

    REQUIRE_PARAMS(2);
    if(!(hi = chanserv_get_handle_info(user, argv[1]))) return 0;
    self = GetChannelUser(channel->channel_info, user->handle_info);
    if(!(target = GetTrueChannelAccess(channel->channel_info, hi)))
    {
        chanserv_notice(user, CSMSG_NO_CHAN_USER, hi->handle, channel->name);
        return 0;
    }
    if(target->access >= self->access)
    {
        chanserv_notice(user, MSG_USER_OUTRANKED, hi->handle);
        return 0;
    }
    if(!(target->flags & USER_SUSPENDED))
    {
        chanserv_notice(user, CSMSG_NOT_SUSPENDED, hi->handle);
        return 0;
    }
    target->flags &= ~USER_SUSPENDED;
    chanserv_notice(user, CSMSG_USER_UNSUSPENDED, hi->handle, channel->name);
    return 1;
}

static MODCMD_FUNC(cmd_deleteme)
{
    struct handle_info *hi;
    struct userData *target;
    const char *confirm_string;
    char *channel_name;

    hi = user->handle_info;
    if(!(target = GetTrueChannelAccess(channel->channel_info, hi)))
    {
        reply(CSMSG_NO_CHAN_USER, hi->handle, channel->name);
        return 0;
    }
    if(target->access == ulOwner)
    {
        reply(CSMSG_NO_OWNER_DELETEME, channel->name);
        return 0;
    }
    confirm_string = make_confirmation_string(target);
    if((argc < 2) || strcmp(argv[1], confirm_string))
    {
        reply(CSMSG_CONFIRM_DELETEME, confirm_string);
        return 0;
    }
    confirm_string = user_level_from_level(target->access);
    channel_name = strdup(channel->name);
    del_channel_user(target, 1);
    reply(CSMSG_DELETED_YOU, confirm_string, channel_name);
    free(channel_name);
    return 1;
}

static void
chanserv_refresh_topics(UNUSED_ARG(void *data))
{
    unsigned int refresh_num = (now - self->link) / chanserv_conf.refresh_period;
    struct chanData *cData;
    char opt;

    for(cData = channelList; cData; cData = cData->next)
    {
        if(IsSuspended(cData)) continue;
        opt = cData->options[optTopicRefresh];
        if(opt == 'n') continue;
        if((refresh_num - cData->last_refresh) < (unsigned int)(1 << (opt - '1'))) continue;
        if(cData->topic[0]) SetChannelTopic(cData->channel, chanserv, cData->topic, 1);
        cData->last_refresh = refresh_num;
    }
    timeq_add(now + chanserv_conf.refresh_period, chanserv_refresh_topics, NULL);
}

static CHANSERV_FUNC(cmd_unf)
{
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "%s: %s", user->nick, CSMSG_UNF_RESPONSE);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        chanserv_notice(user, CSMSG_UNF_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_ping)
{
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "%s: %s", user->nick, CSMSG_PING_RESPONSE);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        chanserv_notice(user, CSMSG_PING_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_wut)
{
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "%s: %s", user->nick, CSMSG_WUT_RESPONSE);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        chanserv_notice(user, CSMSG_WUT_RESPONSE);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_8ball)
{
    unsigned int i, j, accum;
    const char *resp;

    REQUIRE_PARAMS(2);
    accum = 0;
    for (i=1; i<argc; i++) {
        for (j=0; argv[i][j]; j++) {
            accum = (accum << 5) - accum + toupper(argv[i][j]);
        }
    }
    resp = chanserv_conf.eightball->list[accum % chanserv_conf.eightball->used];
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "%s: %s", user->nick, resp);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        chanserv_notice(user, "%s", resp);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_d)
{
    int sides, rolled;

    REQUIRE_PARAMS(2);
    if((sides = atoi(argv[1])) <= 1)
    {
        chanserv_notice(user, CSMSG_BAD_NUMBER, argv[1]);
        return 0;
    }
    rolled = (rand() % sides)+1;

    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "%s: A %d shows on the %d-sided die.", user->nick, rolled, sides);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        chanserv_notice(user, "A %d shows on the %d-sided die.", rolled, sides);
    }
    return 1;
}

static CHANSERV_FUNC(cmd_huggle)
{
    if(channel)
    {
        char response[MAXLEN];
        sprintf(response, "\001ACTION huggles %s\001", user->nick);
        irc_privmsg(chanserv, channel->name, response);
    }
    else
    {
        /* CTCP must be via PRIVMSG, never notice */
        irc_privmsg(chanserv, user->nick, "\001ACTION huggles you\001");
    }
    return 1;
}

static void
chanserv_adjust_limit(void *data)
{
    struct chanData *cData = data;
    struct chanNode *channel;
    unsigned int limit;
    char change[10];

    if(IsSuspended(cData)) return;

    channel = cData->channel;
    limit = channel->members.used + chanserv_conf.adjust_threshold + 5;
    cData->limitAdjusted = now;

    if(cData->limit)
    {
	if(limit > cData->limit)
	{
	    limit = cData->limit;
	}
	else if(limit == cData->limit)
	{
	    return;
	}
    }

    sprintf(change, "+l %d", limit);

    channel->limit = limit;
    irc_mode(chanserv, channel, change);
}

static void
handle_new_channel(struct chanNode *channel)
{
    struct chanData *cData;

    if(!(cData = channel->channel_info)) return;

    if(cData->mode_on || cData->mode_off)
    {
	char modes[MODELEN];
	create_mode_lock(modes, cData, 0);
	mod_chanmode(channel, chanserv, modes);
	irc_mode(chanserv, channel, modes);
    }

    if(self->uplink && !self->uplink->burst && channel->channel_info->topic[0])
    {
        SetChannelTopic(channel, chanserv, channel->channel_info->topic, 1);
    }
}

/* Welcome to my worst nightmare. Warning: Read (or modify)
   the code below at your own risk. */
static int
handle_join(struct modeNode *mNode)
{
    struct userNode *user = mNode->user;
    struct chanNode *channel = mNode->channel;
    struct chanData *cData;
    struct userData *uData;
    struct banData *bData;
    struct handle_info *handle;
    unsigned int modes;
    char *greeting;
    char info[MAXLEN];

    if(IsLocal(user) || !channel->channel_info || IsSuspended(channel->channel_info))
    {
        return 0;
    }

    cData = channel->channel_info;
    if(channel->members.used > cData->max)
    {
        cData->max = channel->members.used;
    }

    /* Check for bans.  If they're joining through a ban, one of two
     * cases applies:
     *   1: Join during a netburst, by riding the break.  Kick them.
     *   2: They're allowed to join through the ban (an invite in
     *   ircu2.10, or a +e on Hybrid, or something).
     * If they're not joining through a ban, and the banlist is not
     * full, see if they're on the banlist for the channel.  If so,
     * kickban them.
     */
    if(user->uplink->burst)
    {
        unsigned int ii;
        for(ii = 0; ii < channel->banlist.used; ii++)
        {
            if(user_matches_glob(user, channel->banlist.list[ii]->ban, 1))
            {
                /* Riding a netburst.  Naughty. */
                KickChannelUser(user, channel, chanserv, CSMSG_NETRIDE_DETECTED);
                return 1;
            }
        }
    }

    if(channel->banlist.used < MAXBANS)
    {
        /* Not joining through a ban. */
        for(bData = cData->bans;
                bData && !user_matches_glob(user, bData->mask, 1);
                bData = bData->next);

        if(bData)
        {
            char *ban;
            char kick_reason[MAXLEN];
            sprintf(kick_reason, CSMSG_KICK_FORMAT, bData->reason, bData->owner);

            bData->triggered = now;
            if(bData != cData->bans)
            {
                /* Shuffle the ban to the head of the list. */
                if(bData->next) bData->next->prev = bData->prev;
                if(bData->prev) bData->prev->next = bData->next;

                bData->prev = NULL;
                bData->next = cData->bans;

                if(cData->bans)
                {
                    cData->bans->prev = bData;
                }
                cData->bans = bData;
            }

            ban = bData->mask;
            AddChannelBan(1, (const char **)&ban, channel, chanserv, now, 1);
            KickChannelUser(user, channel, chanserv, kick_reason);
            return 1;
        }
    }

    /* ChanServ will not modify the limits in join-flooded channels.
       It will also skip DynLimit processing when the user (or srvx)
       is bursting in, because there are likely more incoming. */
    if(cData->flags & CHANNEL_DYNAMIC_LIMIT && !channel->join_flooded &&
       !user->uplink->burst)
    {
        if((channel->limit - channel->members.used) < chanserv_conf.adjust_threshold)
        {
            /* The user count has begun "bumping" into the channel
               limit, so set a timer to raise the limit a bit. Any
               previous timers are removed so three incoming users
               within the delay results in one limit change, not
               three. */

            timeq_del(0, chanserv_adjust_limit, cData, TIMEQ_IGNORE_WHEN);
            timeq_add(now + chanserv_conf.adjust_delay, chanserv_adjust_limit, cData);
        }
    }

    modes = 0;
    info[0] = 0;
    if(validate_auto_op(user, channel))
    {
        modes |= MODE_CHANOP;
    }
    else if((cData->flags & CHANNEL_VOICE_ALL) && !channel->join_flooded)
    {
        modes |= MODE_VOICE;
    }

    greeting = NULL;
    if(user->handle_info)
    {
        handle = user->handle_info;

        if(IsHelper(user) && !IsHelping(user))
        {
            unsigned int ii;
            for(ii = 0; ii < chanserv_conf.support_channels.used; ++ii)
            {
                if(channel == chanserv_conf.support_channels.list[ii])
                {
                    HANDLE_SET_FLAG(user->handle_info, HELPING);
                    break;
                }
            }
        }

        uData = GetTrueChannelAccess(cData, handle);
        if(uData && !IsUserSuspended(uData))
        {
            /* Ops and above were handled by the above case. */
            if((uData->access == ulPeon) && IsUserAutoOp(uData))
            {
                modes |= MODE_VOICE;
            }
            if(uData->access >= ulMaster)
            {
                cData->visited = now;
            }

            if(!user->uplink->burst)
            {
                if(uData->info
                   && (cData->flags & CHANNEL_INFO_LINES)
                   && ((now - uData->seen) >= chanserv_conf.info_delay)
                   && !uData->present)
                {
                    sprintf(info, "[%s] %s", user->nick, uData->info);
                }

                if(cData->user_greeting)
                {
                    greeting = cData->user_greeting;
                }
            }

            uData->seen = now;
            uData->present = 1;
        }
    }
    if(cData->greeting && !greeting && !user->uplink->burst)
    {
        /* Not a user, or a suspended user -- greet them. */
        greeting = cData->greeting;
    }

    if(modes & MODE_CHANOP)
    {
        AddChannelOp(1, &user, channel, chanserv, 1);
    }
    else if(modes & MODE_VOICE)
    {
        AddChannelVoice(1, &user, channel, chanserv, 1);
    }
    if(greeting)
    {
        chanserv_notice(user, "(%s) %s", channel->name, greeting);
    }
    if(info[0])
    {
        irc_privmsg(chanserv, channel->name, info);
    }
    return 0;
}

static void
handle_auth(struct userNode *user, UNUSED_ARG(struct handle_info *old_handle))
{
    struct userData *channel;
    unsigned int ii, jj;

    if(!user->handle_info)
    {
	return;
    }

    for(channel = user->handle_info->channels; channel; channel = channel->u_next)
    {
        struct chanNode *cn;
        if(IsSuspended(channel->channel) || !(cn = channel->channel->channel))
        {
            continue;
        }
        if(!GetUserMode(cn, user) && !IsUserSuspended(channel))
        {
            if(IsUserAutoInvite(channel)
               && (cn->modes & (MODE_KEY | MODE_INVITEONLY))
               && !self->burst)
            {
                irc_invite(chanserv, user, cn);
            }
            continue;
        }

	if(channel->access >= ulMaster)
	{
	    channel->channel->visited = now;
	}

        if(validate_auto_op(user, cn))
        {
            AddChannelOp(1, &user, cn, chanserv, 1);
        }
        else if((channel->access == ulPeon)
                && IsUserAutoOp(channel))
        {
            AddChannelVoice(1, &user, cn, chanserv, 1);
        }

	channel->seen = now;
	channel->present = 1;
    }

    for(ii = 0; ii < user->channels.used; ++ii)
    {
        struct chanNode *channel = user->channels.list[ii]->channel;
        struct banData *ban;

        if((user->channels.list[ii]->modes & (MODE_CHANOP|MODE_VOICE))
           || !channel->channel_info)
            continue;
        for(jj = 0; jj < channel->banlist.used; ++jj)
            if(user_matches_glob(user, channel->banlist.list[jj]->ban, 1))
                break;
        if(jj < channel->banlist.used)
            continue;
        for(ban = channel->channel_info->bans; ban; ban = ban->next)
        {
            const char *ban_mask = ban->mask;
            char kick_reason[MAXLEN];

            if(!user_matches_glob(user, ban->mask, 1))
                continue;
            sprintf(kick_reason, CSMSG_KICK_FORMAT, ban->reason, ban->owner);
            AddChannelBan(1, &ban_mask, channel, chanserv, now, 1);
            KickChannelUser(user, channel, chanserv, kick_reason);
            break;
        }
    }

    if(IsSupportHelper(user))
    {
        for(ii = 0; ii < chanserv_conf.support_channels.used; ++ii)
        {
            if(GetUserMode(chanserv_conf.support_channels.list[ii], user))
            {
                HANDLE_SET_FLAG(user->handle_info, HELPING);
                break;
            }
        }
    }
}

static void
handle_part(struct userNode *user, struct chanNode *channel, UNUSED_ARG(const char *reason))
{
    struct chanData *cData;
    struct userData *uData;
    struct handle_info *handle;

    cData = channel->channel_info;
    if(!cData || IsSuspended(cData) || IsLocal(user)) return;

    if((cData->flags & CHANNEL_DYNAMIC_LIMIT) && !channel->join_flooded)
    {
	/* Allow for a bit of padding so that the limit doesn't
	   track the user count exactly, which could get annoying. */
	if((channel->limit - channel->members.used) > chanserv_conf.adjust_threshold + 5)
	{
	    timeq_del(0, chanserv_adjust_limit, cData, TIMEQ_IGNORE_WHEN);
	    timeq_add(now + chanserv_conf.adjust_delay, chanserv_adjust_limit, cData);
	}
    }

    if((handle = user->handle_info) && (uData = GetTrueChannelAccess(cData, handle)))
    {
	uData->seen = now;
	scan_handle_presence(channel, handle, user);
    }

    if(IsHelping(user) && IsSupportHelper(user))
    {
        unsigned int ii, jj;
        for(ii = 0; ii < chanserv_conf.support_channels.used; ++ii)
        {
            for(jj = 0; jj < user->channels.used; ++jj)
            {
                if(user->channels.list[jj]->channel == chanserv_conf.support_channels.list[ii])
                    break;
            }
            if(jj < user->channels.used)
                break;
        }
        if(ii == chanserv_conf.support_channels.used)
            HANDLE_CLEAR_FLAG(user->handle_info, HELPING);
    }
}

static void
handle_kick(struct userNode *kicker, struct userNode *victim, struct chanNode *channel)
{
    char *reason = CSMSG_PROTECT_REASON;

    if(!channel->channel_info || !kicker || IsService(kicker) || (kicker == victim) || (kicker->handle_info && kicker->handle_info == victim->handle_info) || IsSuspended(channel->channel_info)) return;

    if(protect_user(victim, kicker, channel->channel_info))
    {
	KickChannelUser(kicker, channel, chanserv, reason);
    }
}

static int
handle_topic(struct userNode *user, struct chanNode *channel, const char *old_topic)
{
    struct chanData *cData;

    if(!channel->channel_info || !user || IsSuspended(channel->channel_info) || IsService(user)) return 0;

    cData = channel->channel_info;
    if(bad_topic(channel, user, channel->topic))
    {
        chanserv_notice(user, CSMSG_TOPIC_LOCKED, channel->name);
        if(cData->topic_mask && match_ircglob(old_topic, cData->topic_mask))
        {
            SetChannelTopic(channel, chanserv, old_topic, 1);
        }
        else
        {
            SetChannelTopic(channel, chanserv, cData->topic, 1);
        }
        return 1;
    }
    if(cData->flags & CHANNEL_TOPIC_SNARF)
    {
	/* Grab the topic and save it as the default topic. */
	safestrncpy(cData->topic, channel->topic, sizeof(cData->topic));
    }
    return 0;
}

static void
handle_mode(struct chanNode *channel, struct userNode *user, int change_type, UNUSED_ARG(void *change))
{
    if(!channel->channel_info || IsLocal(user) || IsSuspended(channel->channel_info) || IsService(user)) return;

    if(change_type == MODE_CHANGE_MISC)
    {
	if(!check_user_level(user, channel, channel->channel_info->options[optEnfModes], 1, 0))
	{
            if(mode_lock_violated(channel))
	    {
		char correct[MODELEN];

		create_mode_lock(correct, channel->channel_info, 0);
		mod_chanmode(channel, chanserv, correct);
		irc_mode(chanserv, channel, correct);

		create_mode_lock(correct, channel->channel_info, 1);
		chanserv_notice(user, CSMSG_MODE_LOCKED, correct, channel->name);
	    }
	}
    }
    else if(change_type == MODE_CHANGE_OP)
    {
	struct userNode *victim = change;

	if(!IsService(victim) && !validate_op(user, channel, victim))
	{
	    DelChannelOp(1, &victim, channel, chanserv, 1);
	}
    }
    else if(change_type == MODE_CHANGE_DEOP)
    {
	struct userNode *victim = change;

	if(protect_user(victim, user, channel->channel_info))
	{
	    char change[MAXLEN];

	    DelChannelOp(1, &user, channel, chanserv, 0);
	    AddChannelOp(1, &victim, channel, chanserv, 0);

#ifdef WITH_PROTOCOL_P10
	    sprintf(change, "+o-o %s %s", victim->numeric, user->numeric);
#else
	    sprintf(change, "+o-o %s %s", victim->nick, user->nick);
#endif

	    irc_mode(chanserv, channel, change);
	    chanserv_notice(user, CSMSG_USER_PROTECTED, victim->nick);
	}
    }
    else if(change_type == MODE_CHANGE_BAN)
    {
	if(bad_channel_ban(channel, user, change, NULL, NULL))
	{
	    char *remove = change;
	    DelChannelBan(1, (const char **)&remove, channel, chanserv, 1);
	    chanserv_notice(user, CSMSG_MASK_PROTECTED, remove);
	}	
    }
}

static void
handle_nick_change(struct userNode *user, UNUSED_ARG(const char *old_nick))
{
    struct chanNode *channel;
    struct banData *bData;
    unsigned int ii, jj;
    char kick_reason[MAXLEN];

    for(ii = 0; ii < user->channels.used; ++ii)
    {
        channel = user->channels.list[ii]->channel;
        /* Need not check for bans if they're opped or voiced. */
        if(user->channels.list[ii]->modes & (MODE_CHANOP|MODE_VOICE))
            continue;
        /* Need not check for bans unless channel registration is active. */
        if(!channel->channel_info || IsSuspended(channel->channel_info))
            continue;
        /* Look for a matching ban already on the channel. */
        for(jj = 0; jj < channel->banlist.used; ++jj)
            if(user_matches_glob(user, channel->banlist.list[jj]->ban, 1))
                break;
        /* Need not act if we found one. */
        if(jj < channel->banlist.used)
            continue;
        /* Look for a matching ban in this channel. */
        for(bData = channel->channel_info->bans; bData; bData = bData->next)
        {
            const char *ban = bData->mask;
            if(!user_matches_glob(user, ban, 1))
                continue;
            /* note: &bData->mask == bData->mask, thus we have ban */
            AddChannelBan(1, (const char**)&ban, channel, chanserv, now, 1);
            sprintf(kick_reason, CSMSG_KICK_FORMAT, bData->reason, bData->owner);
            KickChannelUser(user, channel, chanserv, kick_reason);
            break; /* we don't need to check any more bans in the channel */
        }
    }
}

static void handle_rename(struct handle_info *handle, const char *old_handle)
{
    struct do_not_register *dnr = dict_find(handle_dnrs, old_handle, NULL);

    if(dnr)
    {
        dict_remove2(handle_dnrs, old_handle, 1);
        safestrncpy(dnr->chan_name + 1, handle->handle, sizeof(dnr->chan_name) - 1);
        dict_insert(handle_dnrs, dnr->chan_name + 1, dnr);
    }
}

static void
handle_unreg(UNUSED_ARG(struct userNode *user), struct handle_info *handle)
{
    struct userNode *h_user;

    if(handle->channels)
    {
        for(h_user = handle->users; h_user; h_user = h_user->next_authed)
        {
            chanserv_notice(h_user, CSMSG_HANDLE_UNREGISTERED);
        }

        while(handle->channels)
        {
            del_channel_user(handle->channels, 1);
        }
    }
}

static void
handle_server_link(UNUSED_ARG(struct server *server))
{
    struct chanData *cData;

    for(cData = channelList; cData; cData = cData->next)
    {
        if(!IsSuspended(cData))
        {
            cData->may_opchan = 1;
        }
    }
}

static void
chanserv_conf_read(void)
{
    dict_t conf_node;
    const char *str;
    struct string_list *strlist;
    struct chanNode *chan;
    unsigned int ii;

    if(!(conf_node = conf_get_data(CHANSERV_CONF_NAME, RECDB_OBJECT)))
    {
	log_module(CS_LOG, LOG_ERROR, "Invalid config node `%s'.", CHANSERV_CONF_NAME);
	return;
    }
    for(ii = 0; ii < chanserv_conf.support_channels.used; ++ii)
        UnlockChannel(chanserv_conf.support_channels.list[ii]);
    chanserv_conf.support_channels.used = 0;
    if((strlist = database_get_data(conf_node, KEY_SUPPORT_CHANNEL, RECDB_STRING_LIST)))
    {
        for(ii = 0; ii < strlist->used; ++ii)
        {
            const char *str2 = database_get_data(conf_node, KEY_SUPPORT_CHANNEL_MODES, RECDB_QSTRING);
            if(!str2)
                str2 = "+nt";
            chan = AddChannel(strlist->list[ii], now, str2, NULL);
            LockChannel(chan);
            channelList_append(&chanserv_conf.support_channels, chan);
        }
    }
    else if((str = database_get_data(conf_node, KEY_SUPPORT_CHANNEL, RECDB_QSTRING)))
    {
        const char *str2;
        str2 = database_get_data(conf_node, KEY_SUPPORT_CHANNEL_MODES, RECDB_QSTRING);
        if(!str2)
            str2 = "+nt";
        chan = AddChannel(str, now, str2, NULL);
        LockChannel(chan);
        channelList_append(&chanserv_conf.support_channels, chan);
    }
    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    chanserv_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;
    str = database_get_data(conf_node, KEY_INFO_DELAY, RECDB_QSTRING);
    chanserv_conf.info_delay = str ? ParseInterval(str) : 180;
    str = database_get_data(conf_node, KEY_MAX_GREETLEN, RECDB_QSTRING);
    chanserv_conf.greeting_length = str ? atoi(str) : 120;
    str = database_get_data(conf_node, KEY_ADJUST_THRESHOLD, RECDB_QSTRING);
    chanserv_conf.adjust_threshold = str ? atoi(str) : 15;
    str = database_get_data(conf_node, KEY_ADJUST_DELAY, RECDB_QSTRING);
    chanserv_conf.adjust_delay = str ? ParseInterval(str) : 30;
    str = database_get_data(conf_node, KEY_CHAN_EXPIRE_FREQ, RECDB_QSTRING);
    chanserv_conf.channel_expire_frequency = str ? ParseInterval(str) : 86400;
    str = database_get_data(conf_node, KEY_CHAN_EXPIRE_DELAY, RECDB_QSTRING);
    chanserv_conf.channel_expire_delay = str ? ParseInterval(str) : 86400*30;
    str = database_get_data(conf_node, KEY_NODELETE_LEVEL, RECDB_QSTRING);
    chanserv_conf.nodelete_level = str ? atoi(str) : 1;
    str = database_get_data(conf_node, KEY_MAX_CHAN_USERS, RECDB_QSTRING);
    chanserv_conf.max_chan_users = str ? atoi(str) : 512;
    str = database_get_data(conf_node, KEY_MAX_CHAN_BANS, RECDB_QSTRING);
    chanserv_conf.max_chan_bans = str ? atoi(str) : 512;
    str = database_get_data(conf_node, KEY_NICK, RECDB_QSTRING);
    if(str) NickChange(chanserv, str, 0);
    str = database_get_data(conf_node, KEY_TRIGGER, RECDB_QSTRING);
    chanserv_conf.trigger_char = str ? str[0] : 0;
    str = database_get_data(conf_node, KEY_REFRESH_PERIOD, RECDB_QSTRING);
    chanserv_conf.refresh_period = str ? ParseInterval(str) : 3*60*60;
    str = database_get_data(conf_node, KEY_CTCP_SHORT_BAN_DURATION, RECDB_QSTRING);
    chanserv_conf.ctcp_short_ban_duration = str ? str : "3m";
    str = database_get_data(conf_node, KEY_CTCP_LONG_BAN_DURATION, RECDB_QSTRING);
    chanserv_conf.ctcp_long_ban_duration = str ? str : "1h";
    str = database_get_data(conf_node, KEY_MAX_OWNED, RECDB_QSTRING);
    chanserv_conf.max_owned = str ? atoi(str) : 5;
    str = database_get_data(conf_node, KEY_IRC_OPERATOR_EPITHET, RECDB_QSTRING);
    chanserv_conf.irc_operator_epithet = str ? str : "a megalomaniacal power hungry tyrant";
    str = database_get_data(conf_node, KEY_NETWORK_HELPER_EPITHET, RECDB_QSTRING);
    chanserv_conf.network_helper_epithet = str ? str : "a wannabe tyrant";
    str = database_get_data(conf_node, KEY_SUPPORT_HELPER_EPITHET, RECDB_QSTRING);
    chanserv_conf.support_helper_epithet = str ? str : "a wannabe tyrant";
    free_string_list(chanserv_conf.set_shows);
    strlist = database_get_data(conf_node, "set_shows", RECDB_STRING_LIST);
    if(strlist)
    {
        strlist = string_list_copy(strlist);
    }
    else
    {
        static const char *list[] = {
            /* multiple choice options */
            "DefaultTopic", "TopicMask", "Greeting", "UserGreeting", "Modes",
            "PubCmd", "StrictOp", "AutoOp", "EnfModes", "EnfTopic", "Protect",
            "Toys", "Setters", "TopicRefresh", "CtcpUsers", "CtcpReaction",
            /* binary options */
            "Voice", "UserInfo", "DynLimit", "TopicSnarf", "PeonInvite", "NoDelete",
            /* delimiter */
            NULL };
        unsigned int ii;
        strlist = alloc_string_list(ArrayLength(list)-1);
        for(ii=0; list[ii]; ii++)
        {
            string_list_append(strlist, strdup(list[ii]));
        }
    }
    chanserv_conf.set_shows = strlist;
    /* We don't look things up now, in case the list refers to options
     * defined by modules initialized after this point.  Just mark the
     * function list as invalid, so it will be initialized.
     */
    set_shows_list.used = 0;
    free_string_list(chanserv_conf.eightball);
    strlist = database_get_data(conf_node, KEY_8BALL_RESPONSES, RECDB_STRING_LIST);
    if(strlist) {
        strlist = string_list_copy(strlist);
    } else {
        strlist = alloc_string_list(4);
        string_list_append(strlist, strdup("Yes."));
        string_list_append(strlist, strdup("No."));
        string_list_append(strlist, strdup("Maybe so."));
    }
    chanserv_conf.eightball = strlist;
}

static enum userLevel
leveltoaccess(int level)
{
    enum userLevel access;

    for(access = USER_LEVEL_HIGHEST; access >= USER_LEVEL_LOWEST; access--)
    {
	if(level >= levelMap[access])
	{
	    return access;
	}
    }

    return ulNone;
}

static void
chanserv_note_type_read(const char *key, struct record_data *rd)
{
    dict_t obj;
    struct note_type *ntype;
    const char *str;

    if(!(obj = GET_RECORD_OBJECT(rd)))
    {
        log_module(CS_LOG, LOG_ERROR, "Invalid note type %s.", key);
        return;
    }
    if(!(ntype = chanserv_create_note_type(key)))
    {
        log_module(CS_LOG, LOG_ERROR, "Memory allocation failed for note %s.", key);
        return;
    }

    /* Figure out set access */
    if((str = database_get_data(obj, KEY_NOTE_OPSERV_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_PRIVILEGED;
        ntype->set_access.min_opserv = strtoul(str, NULL, 0);
    }
    else if((str = database_get_data(obj, KEY_NOTE_CHANNEL_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_CHANNEL_ACCESS;
        ntype->set_access.min_ulevel = leveltoaccess(strtoul(str, NULL, 0));
    }
    else if((str = database_get_data(obj, KEY_NOTE_SETTER_ACCESS, RECDB_QSTRING)))
    {
        ntype->set_access_type = NOTE_SET_CHANNEL_SETTER;
    }
    else
    {
        log_module(CS_LOG, LOG_ERROR, "Could not find access type for note %s; defaulting to OpServ access level 0.", key);
        ntype->set_access_type = NOTE_SET_PRIVILEGED;
        ntype->set_access.min_opserv = 0;
    }

    /* Figure out visibility */
    if((str = database_get_data(obj, KEY_NOTE_VISIBILITY, RECDB_QSTRING)))
    {
        if(!irccasecmp(str, KEY_NOTE_VIS_PRIVILEGED))
        {
            ntype->visible_type = NOTE_VIS_PRIVILEGED;
        }
        else if(!irccasecmp(str, KEY_NOTE_VIS_CHANNEL_USERS))
        {
            ntype->visible_type = NOTE_VIS_CHANNEL_USERS;
        }
        else if(!irccasecmp(str, KEY_NOTE_VIS_ALL))
        {
            ntype->visible_type = NOTE_VIS_ALL;
        }
        else
        {
            ntype->visible_type = NOTE_VIS_PRIVILEGED;
        }
    }
    else
    {
        ntype->set_access_type = NOTE_VIS_PRIVILEGED;
    }

    str = database_get_data(obj, KEY_NOTE_MAX_LENGTH, RECDB_QSTRING);
    ntype->max_length = str ? strtoul(str, NULL, 0) : 400;
}

static void
user_read_helper(const char *key, struct record_data *rd, struct chanData *chan)
{
    enum userLevel access;
    char *seen, *inf, *flags;
    time_t last_seen;
    int level;
    struct handle_info *handle;
    struct userData *uData;

    if(rd->type != RECDB_OBJECT || !dict_size(rd->d.object))
    {
	log_module(CS_LOG, LOG_ERROR, "Invalid user in %s.", chan->channel->name);
	return;
    }

    level = atoi(database_get_data(rd->d.object, KEY_LEVEL, RECDB_QSTRING));
    access = leveltoaccess(level);

    if(access > USER_LEVEL_HIGHEST || access < USER_LEVEL_LOWEST)
    {
	log_module(CS_LOG, LOG_ERROR, "Invalid access level for %s in %s.", key, chan->channel->name);
	return;
    }

    inf = database_get_data(rd->d.object, KEY_INFO, RECDB_QSTRING);
    seen = database_get_data(rd->d.object, KEY_SEEN, RECDB_QSTRING);
    last_seen = seen ? (signed)strtoul(seen, NULL, 0) : now;
    flags = database_get_data(rd->d.object, KEY_FLAGS, RECDB_QSTRING);
    handle = get_handle_info(key);
    if(!handle)
    {
        log_module(CS_LOG, LOG_ERROR, "Nonexistent account %s in %s.", key, chan->channel->name);
        return;
    }

    uData = add_channel_user(chan, handle, access, last_seen, inf);
    uData->flags = flags ? strtoul(flags, NULL, 0) : 0;
}

static void
ban_read_helper(const char *key, struct record_data *rd, struct chanData *chan)
{
    struct banData *bData;
    char *set, *triggered, *s_duration, *s_expires, *reason, *owner;
    time_t set_time, triggered_time, expires_time;

    if(rd->type != RECDB_OBJECT || !dict_size(rd->d.object))
    {
	log_module(CS_LOG, LOG_ERROR, "Invalid ban in %s.", chan->channel->name);
	return;
    }

    set = database_get_data(rd->d.object, KEY_SET, RECDB_QSTRING);
    triggered = database_get_data(rd->d.object, KEY_TRIGGERED, RECDB_QSTRING);
    s_duration = database_get_data(rd->d.object, KEY_DURATION, RECDB_QSTRING);
    s_expires = database_get_data(rd->d.object, KEY_EXPIRES, RECDB_QSTRING);
    owner = database_get_data(rd->d.object, KEY_OWNER, RECDB_QSTRING);
    reason = database_get_data(rd->d.object, KEY_REASON, RECDB_QSTRING);

    set_time = set ? (time_t)strtoul(set, NULL, 0) : now;
    triggered_time = triggered ? (time_t)strtoul(triggered, NULL, 0) : 0;
    if(s_expires)
    {
        expires_time = strtoul(s_expires, NULL, 0);
    }
    else if(s_duration)
    {
        expires_time = set_time + strtoul(s_duration, NULL, 0);
    }
    else
    {
        expires_time = 0;
    }

    if(expires_time && (expires_time < now))
    {
	/* The ban has expired. */
	return;
    }

    bData = add_channel_ban(chan, key, owner, set_time, triggered_time, expires_time, reason);
}

static struct suspended *
chanserv_read_suspended(dict_t obj)
{
    struct suspended *suspended = calloc(1, sizeof(*suspended));
    char *str;
    dict_t previous;

    str = database_get_data(obj, KEY_EXPIRES, RECDB_QSTRING);
    suspended->expires = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(obj, KEY_REVOKED, RECDB_QSTRING);
    suspended->revoked = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(obj, KEY_ISSUED, RECDB_QSTRING);
    suspended->issued = str ? strtoul(str, NULL, 0) : 0;
    suspended->suspender = strdup(database_get_data(obj, KEY_SUSPENDER, RECDB_QSTRING));
    suspended->reason = strdup(database_get_data(obj, KEY_REASON, RECDB_QSTRING));
    previous = database_get_data(obj, KEY_PREVIOUS, RECDB_OBJECT);
    suspended->previous = previous ? chanserv_read_suspended(previous) : NULL;
    return suspended;
}

static int
chanserv_channel_read(const char *key, struct record_data *hir)
{
    struct chanNode *cNode;
    struct chanData *cData;
    dict_t channel, obj;
    dict_iterator_t it;
    char *str, *registrar, *options;
    unsigned long num_flags;
    struct suspended *suspended;

    channel = hir->d.object;

    if((str = database_get_data(channel, KEY_FLAGS, RECDB_QSTRING)))
    {
	num_flags = base64toint(str, FLAGS_STRING_LENGTH);
	options = str + FLAGS_STRING_LENGTH;
    }
    else
    {
	num_flags = CHANNEL_DEFAULT_FLAGS;
	options = CHANNEL_DEFAULT_OPTIONS;
    }
   
    if((obj = database_get_data(hir->d.object, KEY_SUSPENDED, RECDB_OBJECT)))
    {
        suspended = chanserv_read_suspended(obj);
        /* We could use suspended->expires and suspended->revoked to
         * set the CHANNEL_SUSPENDED flag, but we don't. */
    }
    else if(num_flags & CHANNEL_SUSPENDED)
    {
        suspended = calloc(1, sizeof(*suspended));
        suspended->issued = 0;
        suspended->revoked = 0;
        str = database_get_data(hir->d.object, KEY_SUSPEND_EXPIRES, RECDB_QSTRING);
        suspended->expires = str ? strtoul(str, NULL, 0) : 0;
        suspended->suspender = strdup(database_get_data(hir->d.object, KEY_SUSPENDER, RECDB_QSTRING));
        str = database_get_data(hir->d.object, KEY_SUSPEND_REASON, RECDB_QSTRING);
        suspended->reason = strdup(str ? str : "No reason");
        suspended->previous = NULL;
    }
    else
    {
        suspended = NULL;
    }

    if((num_flags & CHANNEL_SUSPENDED)
       && suspended->expires
       && (suspended->expires < now))
    {
        num_flags &= ~CHANNEL_SUSPENDED;
    }

    cNode = AddChannel(key, now, NULL, NULL);
    if(!cNode)
    {
        log_module(CS_LOG, LOG_ERROR, "Unable to create registered channel %s.", key);
        return 0;
    }
    if((num_flags & CHANNEL_SUSPENDED) && (suspended->expires > now))
    {
        timeq_add(suspended->expires, chanserv_expire_suspension, suspended);
    }
    else
    {
        AddChannelUser(chanserv, cNode);
        AddChannelOp(1, &chanserv, cNode, chanserv, 0);
    }

    registrar = database_get_data(channel, KEY_REGISTRAR, RECDB_QSTRING);
    cData = register_channel(cNode, registrar, options, num_flags);
    if(!cData)
    {
        log_module(CS_LOG, LOG_ERROR, "Unable to register channel %s from database.", key);
	return 0;
    }

    if((cData->suspended = suspended)) suspended->cData = cData;
    str = database_get_data(channel, KEY_REGISTERED, RECDB_QSTRING);
    cData->registered = str ? (time_t)strtoul(str, NULL, 0) : now;
    str = database_get_data(channel, KEY_VISITED, RECDB_QSTRING);
    cData->visited = str ? (time_t)strtoul(str, NULL, 0) : now;
    str = database_get_data(channel, KEY_MAX, RECDB_QSTRING);
    cData->max = str ? strtoul(str, NULL, 0) : 0;
    str = database_get_data(channel, KEY_GREETING, RECDB_QSTRING);
    cData->greeting = str ? strdup(str) : NULL;
    str = database_get_data(channel, KEY_USER_GREETING, RECDB_QSTRING);
    cData->user_greeting = str ? strdup(str) : NULL;
    str = database_get_data(channel, KEY_TOPIC_MASK, RECDB_QSTRING);
    cData->topic_mask = str ? strdup(str) : NULL;
    str = database_get_data(channel, KEY_TOPIC, RECDB_QSTRING);
    if(str) safestrncpy(cData->topic, str, sizeof(cData->topic));

    if((str = database_get_data(channel, KEY_MODES, RECDB_QSTRING)))
    {
	unsigned char *split_modes[3];
	int count;

	count = split_line(str, 0, ArrayLength(split_modes), split_modes);
	parse_mode_lock(split_modes, count, cData);
        if(!IsSuspended(cData))
        {
            assert(cNode);
            mod_chanmode(cNode, chanserv, str);
        }
    }

    obj = database_get_data(channel, KEY_USERS, RECDB_OBJECT);
    for(it = dict_first(obj); it; it = iter_next(it))
    {
	user_read_helper(iter_key(it), iter_data(it), cData);
    }

    if(!cData->users && !IsProtected(cData))
    {
        log_module(CS_LOG, LOG_ERROR, "Channel %s had no users in database, unregistering it.", key);
	unregister_channel(cData, "has empty user list.");
        return 0;
    }

    obj = database_get_data(channel, KEY_BANS, RECDB_OBJECT);
    for(it = dict_first(obj); it; it = iter_next(it))
    {
        ban_read_helper(iter_key(it), iter_data(it), cData);
    }

    obj = database_get_data(channel, KEY_NOTES, RECDB_OBJECT);
    for(it = dict_first(obj); it; it = iter_next(it))
    {
        struct note_type *ntype = dict_find(note_types, iter_key(it), NULL);
        struct record_data *rd = iter_data(it);
        const char *note, *setter;

        if(rd->type != RECDB_OBJECT)
        {
            log_module(CS_LOG, LOG_ERROR, "Bad record type for note %s in channel %s.", iter_key(it), key);
        }
        else if(!ntype)
        {
            log_module(CS_LOG, LOG_ERROR, "Bad note type name %s in channel %s.", iter_key(it), key);
        }
        else if(!(note = database_get_data(rd->d.object, KEY_NOTE_NOTE, RECDB_QSTRING)))
        {
            log_module(CS_LOG, LOG_ERROR, "Missing note text for note %s in channel %s.", iter_key(it), key);
        }
        else
        {
            setter = database_get_data(rd->d.object, KEY_NOTE_SETTER, RECDB_QSTRING);
            if(!setter) setter = "<unknown>";
            chanserv_add_channel_note(cData, ntype, setter, note);
        }
    }

    return 0;
}

static void
chanserv_dnr_read(const char *key, struct record_data *hir)
{
    const char *setter, *reason;
    setter = database_get_data(hir->d.object, KEY_DNR_SETTER, RECDB_QSTRING);
    if(!setter)
    {
        log_module(CS_LOG, LOG_ERROR, "Missing setter for DNR %s.", key);
        return;
    }
    reason = database_get_data(hir->d.object, KEY_DNR_REASON, RECDB_QSTRING);
    if(!reason)
    {
        log_module(CS_LOG, LOG_ERROR, "Missing reason for DNR %s.", key);
        return;
    }
    chanserv_add_dnr(key, setter, reason);
}

static int
chanserv_saxdb_read(struct dict *database)
{
    struct dict *section;
    dict_iterator_t it;

    if((section = database_get_data(database, KEY_NOTE_TYPES, RECDB_OBJECT)))
    {
        for(it = dict_first(section); it; it = iter_next(it))
        {
            chanserv_note_type_read(iter_key(it), iter_data(it));
        }
    }

    if((section = database_get_data(database, KEY_CHANNELS, RECDB_OBJECT)))
    {
	for(it = dict_first(section); it; it = iter_next(it))
	{
	    chanserv_channel_read(iter_key(it), iter_data(it));
	}
    }

    if((section = database_get_data(database, KEY_DNR, RECDB_OBJECT)))
    {
        for(it = dict_first(section); it; it = iter_next(it))
        {
            chanserv_dnr_read(iter_key(it), iter_data(it));
        }
    }
    return 0;
}

static int
chanserv_write_users(struct saxdb_context *ctx, struct userData *uData)
{
    int high_present = 0;
    saxdb_start_record(ctx, KEY_USERS, 1);
    for(; uData; uData = uData->next)
    {
        if((uData->access >= ulMaster) && uData->present)
            high_present = 1;
        saxdb_start_record(ctx, uData->handle->handle, 0);
        saxdb_write_int(ctx, KEY_LEVEL, levelMap[uData->access]);
        saxdb_write_int(ctx, KEY_SEEN, uData->seen);
        if(uData->flags)
            saxdb_write_int(ctx, KEY_FLAGS, uData->flags);
	if(uData->info)
            saxdb_write_string(ctx, KEY_INFO, uData->info);
        saxdb_end_record(ctx);
    }
    saxdb_end_record(ctx);
    return high_present;
}

static void
chanserv_write_bans(struct saxdb_context *ctx, struct banData *bData)
{
    if(!bData) return;
    saxdb_start_record(ctx, KEY_BANS, 1);
    for(; bData; bData = bData->next)
    {
        saxdb_start_record(ctx, bData->mask, 0);
        saxdb_write_int(ctx, KEY_SET, bData->set);
        if(bData->triggered) saxdb_write_int(ctx, KEY_TRIGGERED, bData->triggered);
        if(bData->expires) saxdb_write_int(ctx, KEY_EXPIRES, bData->expires);
        if(bData->owner[0]) saxdb_write_string(ctx, KEY_OWNER, bData->owner);
        if(bData->reason) saxdb_write_string(ctx, KEY_REASON, bData->reason);
        saxdb_end_record(ctx);
    }
    saxdb_end_record(ctx);
}

static void
chanserv_write_suspended(struct saxdb_context *ctx, const char *name, struct suspended *susp)
{
    saxdb_start_record(ctx, name, 0);
    saxdb_write_string(ctx, KEY_SUSPENDER, susp->suspender);
    saxdb_write_string(ctx, KEY_REASON, susp->reason);
    if(susp->issued) saxdb_write_int(ctx, KEY_ISSUED, susp->issued);
    if(susp->expires) saxdb_write_int(ctx, KEY_EXPIRES, susp->expires);
    if(susp->revoked) saxdb_write_int(ctx, KEY_REVOKED, susp->revoked);
    if(susp->previous) chanserv_write_suspended(ctx, KEY_PREVIOUS, susp->previous);
    saxdb_end_record(ctx);
}

static void
chanserv_write_channel(struct saxdb_context *ctx, struct chanData *channel)
{
    char buf[MAXLEN];
    int high_present;

    saxdb_start_record(ctx, channel->channel->name, 1);

    saxdb_write_int(ctx, KEY_REGISTERED, channel->registered);
    saxdb_write_int(ctx, KEY_MAX, channel->max);
    if(channel->topic[0]) saxdb_write_string(ctx, KEY_TOPIC, channel->topic);
    if(channel->registrar) saxdb_write_string(ctx, KEY_REGISTRAR, channel->registrar);
    if(channel->greeting) saxdb_write_string(ctx, KEY_GREETING, channel->greeting);
    if(channel->user_greeting) saxdb_write_string(ctx, KEY_USER_GREETING, channel->user_greeting);
    if(channel->topic_mask) saxdb_write_string(ctx, KEY_TOPIC_MASK, channel->topic_mask);
    if(channel->suspended) chanserv_write_suspended(ctx, "suspended", channel->suspended);

    inttobase64(buf, channel->flags, FLAGS_STRING_LENGTH);
    safestrncpy(buf + FLAGS_STRING_LENGTH, channel->options, sizeof(buf)-FLAGS_STRING_LENGTH-1);
    saxdb_write_string(ctx, KEY_FLAGS, buf);

    if(channel->mode_on || channel->mode_off)
    {
	create_mode_lock(buf, channel, 1);
        saxdb_write_string(ctx, KEY_MODES, buf);
    }

    high_present = chanserv_write_users(ctx, channel->users);
    chanserv_write_bans(ctx, channel->bans);

    if(dict_size(channel->notes))
    {
        dict_iterator_t it;

        saxdb_start_record(ctx, KEY_NOTES, 1);
        for(it = dict_first(channel->notes); it; it = iter_next(it))
        {
	    struct note *note = iter_data(it);
            saxdb_start_record(ctx, iter_key(it), 0);
            saxdb_write_string(ctx, KEY_NOTE_SETTER, note->setter);
            saxdb_write_string(ctx, KEY_NOTE_NOTE, note->note);
            saxdb_end_record(ctx);
        }
        saxdb_end_record(ctx);
    }

    saxdb_write_int(ctx, KEY_VISITED, high_present ? now : channel->visited);
    saxdb_end_record(ctx);
}

static void
chanserv_write_note_type(struct saxdb_context *ctx, struct note_type *ntype)
{
    const char *str;

    saxdb_start_record(ctx, ntype->name, 0);
    switch(ntype->set_access_type)
    {
    case NOTE_SET_CHANNEL_ACCESS:
        saxdb_write_int(ctx, KEY_NOTE_CHANNEL_ACCESS, levelMap[ntype->set_access.min_ulevel]);
        break;
    case NOTE_SET_CHANNEL_SETTER:
        saxdb_write_int(ctx, KEY_NOTE_SETTER_ACCESS, 1);
        break;
    case NOTE_SET_PRIVILEGED: default:
        saxdb_write_int(ctx, KEY_NOTE_OPSERV_ACCESS, ntype->set_access.min_opserv);
        break;
    }
    switch(ntype->visible_type)
    {
    case NOTE_VIS_ALL: str = KEY_NOTE_VIS_ALL; break;
    case NOTE_VIS_CHANNEL_USERS: str = KEY_NOTE_VIS_CHANNEL_USERS; break;
    case NOTE_VIS_PRIVILEGED: default: str = KEY_NOTE_VIS_PRIVILEGED; break;
    }
    saxdb_write_string(ctx, KEY_NOTE_VISIBILITY, str);
    saxdb_write_int(ctx, KEY_NOTE_MAX_LENGTH, ntype->max_length);
    saxdb_end_record(ctx);
}

static void
write_dnrs_helper(struct saxdb_context *ctx, struct dict *dnrs)
{
    struct do_not_register *dnr;
    dict_iterator_t it;

    for(it = dict_first(dnrs); it; it = iter_next(it))
    {
        dnr = iter_data(it);
        saxdb_start_record(ctx, dnr->chan_name, 0);
        saxdb_write_string(ctx, KEY_DNR_SETTER, dnr->setter);
        saxdb_write_string(ctx, KEY_DNR_REASON, dnr->reason);
        saxdb_end_record(ctx);
    }
}

static int
chanserv_saxdb_write(struct saxdb_context *ctx)
{
    dict_iterator_t it;
    struct chanData *channel;

    /* Notes */
    saxdb_start_record(ctx, KEY_NOTE_TYPES, 1);
    for(it = dict_first(note_types); it; it = iter_next(it))
    {
        chanserv_write_note_type(ctx, iter_data(it));
    }
    saxdb_end_record(ctx);

    /* DNRs */
    saxdb_start_record(ctx, KEY_DNR, 1);
    write_dnrs_helper(ctx, handle_dnrs);
    write_dnrs_helper(ctx, plain_dnrs);
    write_dnrs_helper(ctx, mask_dnrs);
    saxdb_end_record(ctx);

    /* Channels */
    saxdb_start_record(ctx, KEY_CHANNELS, 1);
    for(channel = channelList; channel; channel = channel->next)
    {
        chanserv_write_channel(ctx, channel);
    }
    saxdb_end_record(ctx);

    return 0;
}

static void
chanserv_db_cleanup(void) {
    unsigned int ii;
    unreg_part_func(handle_part);
    while(channelList)
        unregister_channel(channelList, "terminating.");
    for(ii = 0; ii < chanserv_conf.support_channels.used; ++ii)
        UnlockChannel(chanserv_conf.support_channels.list[ii]);
    free(chanserv_conf.support_channels.list);
    dict_delete(handle_dnrs);
    dict_delete(plain_dnrs);
    dict_delete(mask_dnrs);
    dict_delete(note_types);
    if(chanserv_conf.eightball) free_string_list(chanserv_conf.eightball);
    if(chanserv_conf.set_shows) free_string_list(chanserv_conf.set_shows);
    if(set_shows_list.list) free(set_shows_list.list);
    if(uset_shows_list.list) free(uset_shows_list.list);
    while(helperList)
    {
        struct userData *helper = helperList;
        helperList = helperList->next;
        free(helper);
    }
}

#define DEFINE_COMMAND(NAME, MIN_ARGC, FLAGS, OPTIONS...) modcmd_register(chanserv_module, #NAME, cmd_##NAME, MIN_ARGC, FLAGS, ## OPTIONS)
#define DEFINE_CHANNEL_OPTION(NAME) modcmd_register(chanserv_module, "set "#NAME, chan_opt_##NAME, 1, 0, NULL)
#define DEFINE_USER_OPTION(NAME) modcmd_register(chanserv_module, "uset "#NAME, user_opt_##NAME, 1, 0, NULL)

void
init_chanserv(const char *nick)
{
    chanserv = AddService(nick, "Channel Services");
    CS_LOG = log_register_type("ChanServ", "file:chanserv.log");
    conf_register_reload(chanserv_conf_read);

    reg_server_link_func(handle_server_link);

    reg_new_channel_func(handle_new_channel);
    reg_join_func(handle_join);
    reg_part_func(handle_part);
    reg_kick_func(handle_kick);
    reg_topic_func(handle_topic);
    reg_mode_change_func(handle_mode);
    reg_nick_change_func(handle_nick_change);

    reg_auth_func(handle_auth);
    reg_handle_rename_func(handle_rename);
    reg_unreg_func(handle_unreg);

    handle_dnrs = dict_new();
    dict_set_free_data(handle_dnrs, free);
    plain_dnrs = dict_new();
    dict_set_free_data(plain_dnrs, free);
    mask_dnrs = dict_new();
    dict_set_free_data(mask_dnrs, free);

    reg_svccmd_unbind_func(handle_svccmd_unbind);
    chanserv_module = module_register("ChanServ", CS_LOG, "chanserv.help", chanserv_expand_variable);
    DEFINE_COMMAND(register, 1, MODCMD_REQUIRE_AUTHED, "flags", "+acceptchan,+helping", NULL);
    DEFINE_COMMAND(noregister, 1, MODCMD_REQUIRE_AUTHED, "flags", "+helping", NULL);
    DEFINE_COMMAND(allowregister, 2, 0, "template", "noregister", NULL);
    DEFINE_COMMAND(move, 1, MODCMD_REQUIRE_AUTHED|MODCMD_REQUIRE_REGCHAN, "template", "register", NULL);
    DEFINE_COMMAND(csuspend, 2, MODCMD_REQUIRE_AUTHED|MODCMD_REQUIRE_REGCHAN, "flags", "+helping", NULL);
    DEFINE_COMMAND(cunsuspend, 1, MODCMD_REQUIRE_AUTHED|MODCMD_REQUIRE_REGCHAN, "flags", "+helping", NULL);
    DEFINE_COMMAND(createnote, 5, 0, "access", "800", NULL);
    DEFINE_COMMAND(removenote, 2, 0, "access", "800", NULL);

    DEFINE_COMMAND(unregister, 1, MODCMD_REQUIRE_AUTHED|MODCMD_REQUIRE_REGCHAN, "flags", "+loghostmask", NULL);
    DEFINE_COMMAND(merge, 2, MODCMD_REQUIRE_AUTHED|MODCMD_REQUIRE_REGCHAN, "access", "owner", NULL);

    DEFINE_COMMAND(adduser, 3, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(deluser, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(suspend, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(unsuspend, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(deleteme, 1, MODCMD_REQUIRE_CHANUSER, NULL);

    DEFINE_COMMAND(mdelowner, 2, MODCMD_REQUIRE_CHANUSER, "flags", "+helping", NULL);
    DEFINE_COMMAND(mdelcoowner, 2, MODCMD_REQUIRE_CHANUSER, "access", "owner", NULL);
    DEFINE_COMMAND(mdelmaster, 2, MODCMD_REQUIRE_CHANUSER, "access", "coowner", NULL);
    DEFINE_COMMAND(mdelop, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(mdelpeon, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);

    DEFINE_COMMAND(trim, 3, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(opchan, 1, MODCMD_REQUIRE_REGCHAN, "access", "peon", NULL);
    DEFINE_COMMAND(clvl, 3, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);
    DEFINE_COMMAND(giveownership, 2, MODCMD_REQUIRE_CHANUSER, "access", "owner", "flags", "+loghostmask", NULL);

    DEFINE_COMMAND(up, 1, MODCMD_REQUIRE_CHANUSER, NULL);
    DEFINE_COMMAND(down, 1, MODCMD_REQUIRE_REGCHAN, NULL);
    DEFINE_COMMAND(upall, 1, MODCMD_REQUIRE_AUTHED, NULL);
    DEFINE_COMMAND(downall, 1, MODCMD_REQUIRE_AUTHED, NULL);
    DEFINE_COMMAND(op, 2, MODCMD_REQUIRE_CHANNEL, "access", "op", NULL);
    DEFINE_COMMAND(deop, 2, MODCMD_REQUIRE_CHANNEL, "template", "op", NULL);
    DEFINE_COMMAND(voice, 2, MODCMD_REQUIRE_CHANNEL, "template", "op", NULL);
    DEFINE_COMMAND(devoice, 2, MODCMD_REQUIRE_CHANNEL, "template", "op", NULL);

    DEFINE_COMMAND(kickban, 2, MODCMD_REQUIRE_REGCHAN, "template", "op", NULL);
    DEFINE_COMMAND(kick, 2, MODCMD_REQUIRE_REGCHAN, "template", "op", NULL);
    DEFINE_COMMAND(ban, 2, MODCMD_REQUIRE_REGCHAN, "template", "op", NULL);
    DEFINE_COMMAND(unban, 2, 0, "template", "op", NULL);
    DEFINE_COMMAND(unbanall, 1, 0, "template", "op", NULL);
    DEFINE_COMMAND(unbanme, 1, MODCMD_REQUIRE_CHANUSER, "template", "op", NULL);
    DEFINE_COMMAND(open, 1, MODCMD_REQUIRE_CHANUSER, "template", "op", NULL);
    DEFINE_COMMAND(topic, 1, MODCMD_REQUIRE_REGCHAN, "template", "op", "flags", "+never_csuspend", NULL);
    DEFINE_COMMAND(mode, 1, MODCMD_REQUIRE_REGCHAN, "template", "op", NULL);
    DEFINE_COMMAND(inviteme, 1, MODCMD_REQUIRE_CHANNEL, "access", "peon", NULL);
    DEFINE_COMMAND(invite, 1, MODCMD_REQUIRE_CHANNEL, "access", "master", NULL);
    DEFINE_COMMAND(set, 1, MODCMD_REQUIRE_CHANUSER, "access", "op", NULL);
    DEFINE_COMMAND(setinfo, 1, MODCMD_REQUIRE_CHANUSER, "flags", "+nolog", NULL);
    DEFINE_COMMAND(wipeinfo, 2, MODCMD_REQUIRE_CHANUSER, "access", "master", NULL);

    DEFINE_COMMAND(events, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog", "access", "coowner", NULL);
    DEFINE_COMMAND(addban, 2, MODCMD_REQUIRE_REGCHAN, "access", "master", NULL);
    DEFINE_COMMAND(addtimedban, 3, MODCMD_REQUIRE_REGCHAN, "access", "master", NULL);
    DEFINE_COMMAND(delban, 2, MODCMD_REQUIRE_REGCHAN, "access", "master", NULL);
    DEFINE_COMMAND(uset, 1, MODCMD_REQUIRE_CHANUSER, "access", "peon", NULL);

    DEFINE_COMMAND(bans, 1, MODCMD_REQUIRE_REGCHAN, "access", "peon", "flags", "+nolog", NULL);
    DEFINE_COMMAND(peek, 1, MODCMD_REQUIRE_REGCHAN, "access", "op", "flags", "+nolog", NULL);

    DEFINE_COMMAND(access, 1, 0, "flags", "+nolog,+acceptchan", NULL);
    DEFINE_COMMAND(users, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(wlist, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(clist, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(mlist, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(olist, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(plist, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(info, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(seen, 2, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);
    DEFINE_COMMAND(names, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+nolog,+joinable", NULL);

    DEFINE_COMMAND(note, 1, MODCMD_REQUIRE_REGCHAN, "flags", "+joinable,+acceptchan", NULL);
    DEFINE_COMMAND(delnote, 2, MODCMD_REQUIRE_CHANUSER, NULL);

    DEFINE_COMMAND(netinfo, 1, 0, "flags", "+nolog", NULL);
    DEFINE_COMMAND(ircops, 1, 0, "flags", "+nolog", NULL);
    DEFINE_COMMAND(helpers, 1, 0, "flags", "+nolog", NULL);
    DEFINE_COMMAND(staff, 1, 0, "flags", "+nolog", NULL);
    DEFINE_COMMAND(version, 1, 0, "flags", "+nolog", NULL);

    DEFINE_COMMAND(say, 2, 0, "flags", "+oper,+acceptchan", NULL);
    DEFINE_COMMAND(emote, 2, 0, "flags", "+oper,+acceptchan", NULL);
    DEFINE_COMMAND(expire, 1, 0, "flags", "+oper", NULL);
    DEFINE_COMMAND(search, 3, 0, "flags", "+nolog,+helping", NULL);
    DEFINE_COMMAND(unvisited, 1, 0, "flags", "+nolog,+helping", NULL);

    DEFINE_COMMAND(unf, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);
    DEFINE_COMMAND(ping, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);
    DEFINE_COMMAND(wut, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);
    DEFINE_COMMAND(8ball, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);
    DEFINE_COMMAND(d, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);
    DEFINE_COMMAND(huggle, 1, 0, "flags", "+nolog,+toy,+acceptchan", NULL);

    /* Channel options */
    DEFINE_CHANNEL_OPTION(defaulttopic);
    DEFINE_CHANNEL_OPTION(topicmask);
    DEFINE_CHANNEL_OPTION(greeting);
    DEFINE_CHANNEL_OPTION(usergreeting);
    DEFINE_CHANNEL_OPTION(modes);
    DEFINE_CHANNEL_OPTION(strictop);
    DEFINE_CHANNEL_OPTION(autoop);
    DEFINE_CHANNEL_OPTION(protect);
    DEFINE_CHANNEL_OPTION(enfmodes);
    DEFINE_CHANNEL_OPTION(enftopic);
    DEFINE_CHANNEL_OPTION(pubcmd);
    DEFINE_CHANNEL_OPTION(voice);
    DEFINE_CHANNEL_OPTION(userinfo);
    DEFINE_CHANNEL_OPTION(dynlimit);
    DEFINE_CHANNEL_OPTION(topicsnarf);
    DEFINE_CHANNEL_OPTION(nodelete);
    DEFINE_CHANNEL_OPTION(toys);
    DEFINE_CHANNEL_OPTION(setters);
    DEFINE_CHANNEL_OPTION(topicrefresh);
    DEFINE_CHANNEL_OPTION(ctcpusers);
    DEFINE_CHANNEL_OPTION(ctcpreaction);
    DEFINE_CHANNEL_OPTION(peoninvite);

    /* Alias set topic to set defaulttopic for compatibility. */
    modcmd_register(chanserv_module, "set topic", chan_opt_defaulttopic, 1, 0, NULL);

    /* User options */
    DEFINE_USER_OPTION(noautoop);
    DEFINE_USER_OPTION(autoinvite);

    /* Alias uset autovoice to uset autoop. */
    modcmd_register(chanserv_module, "uset noautovoice", user_opt_noautoop, 1, 0, NULL);

    note_types = dict_new();
    dict_set_free_data(note_types, chanserv_deref_note_type);
    saxdb_register("ChanServ", chanserv_saxdb_read, chanserv_saxdb_write);
    chanserv_service = service_register(chanserv, chanserv_conf.trigger_char);
    reg_chanmsg_func('\001', chanserv, chanserv_ctcp_check);

    if(chanserv_conf.channel_expire_frequency)
    {
	timeq_add(now + chanserv_conf.channel_expire_frequency, expire_channels, NULL);
    }

    if(chanserv_conf.refresh_period)
    {
        time_t next_refresh;
        next_refresh = (now + chanserv_conf.refresh_period - 1) / chanserv_conf.refresh_period * chanserv_conf.refresh_period;
        timeq_add(next_refresh, chanserv_refresh_topics, NULL);
    }
    
    reg_exit_func(chanserv_db_cleanup);
}
